
(define (generator)
  (let ((x 0))
    (lambda ()
      (set! x (+ x 1))
      x)))

(define generate-id (generator))

(define hash (make-hash-table 517))
(define (register id st) (hash-set! hash id st))

(define (convert-hname st hname)
  (cond 
   ((eq? hname 'center)
    (if (eq? st 'point) 'point 'center))
   ((eq? hname 'endpoint-1)
    (cond
     ((eq? st 'quadratic-bezier) 'controlpoint-1)
     ((eq? st 'cubic-bezier) 'controlpoint-1)
     (else 'start)))
   ((eq? hname 'endpoint-2)
    (cond
     ((eq? st 'quadratic-bezier) 'controlpoint-3)
     ((eq? st 'cubic-bezier) 'controlpoint-4)
     (else 'end)))
   (else hname)))

(define (convert-1 src)

  (define (convert-doc head num props tlos)
    (list head 0.6 props (map (lambda (tlo) (apply convert-tlo tlo)) tlos)))

  (define (convert-tlo head props gos)
    (list head props (convert-gos gos '())))

  (define (convert-go id go acc)

    (define (rename-hanger hname gid)
      (if (eq? (car go) 'parametric-point)
	  gid
	  (list (convert-hname (hash-ref hash gid) hname) gid)))

    (define (add-ct x)
      (if (eq? (car go) 'elarc-by-center-2-radii-rotation-2-angles)
	  (cons 'start-dif x)
	  x))
	     
    
    (define (convert-go-part part rest lacc acc)
      (cond 
       ((eq? part 'elarc-by-center-2-radii-rotation-2-angles)
	(convert-go-part (car rest) (cdr rest) (cons 'ellipse-by-center-2-radii-rotation-2-angles lacc) acc))
       ((not (pair? part))
	(convert-go-part (car rest) (cdr rest) (cons part lacc) acc))
       ((eq? (car part) 'coords)
	(convert-go-part (car rest) (cdr rest) (cons part lacc) acc))
       ((symbol? (car part))
	(if (pair? (cadr part))
	    (let ((nid (generate-id)))
	      (let ((nacc (convert-go nid (cadr part) acc)))
	        (convert-go-part (car rest) (cdr rest)
			         (cons (rename-hanger (car part) nid) lacc)
			         nacc)))
	    (convert-go-part (car rest) (cdr rest) 
			     (cons (apply rename-hanger part) lacc) acc)))
       (else
	(if id
	    (begin
	      (register id (car go))
	      (cons (reverse (cons (cons (list 'id id) part) (add-ct lacc))) acc))
	    (let ((oid (assq-ref part 'id)))
	      (if oid (register (car oid) (car go)))
	      (cons (reverse (cons part (add-ct lacc))) acc))))))
    (convert-go-part (car go) (cdr go) '() acc))

  (define (convert-gos src acc)
    (if (null? src)
	(list (list 'top-level-group '() (reverse acc)))
	(convert-gos (cdr src) (convert-go #f (car src) acc)))) 
  (apply convert-doc src))

(define (use fin fout fn)
  (write (fn (read (open-input-file fin)))
	 (open-output-file fout)))

(define (main lst)
  (use (cadr lst) (caddr lst) convert-1)
  (quit))
