; Koch snowflake

; Cretes Koch snowflake at given coordinates (must be given clockwise).
; Every part of this snowflake is hanging on free hangers, so it is not
; possible to move with this entire snowflake by moving with some small
; number of anchors.

; example for: segments, complex numbers, recursion (internal procedure definition)

; internal procedure kochline creates one third of snowflake, starting at *p1*
; and stopping at *p2* (complex coordinates) with depth of *dpt* iterations.

; example: (koch  60 -60 -60 -60 0 60 4)

(define pi (angle -1))

(define (koch x1 y1 x2 y2 x3 y3 dpt)
  (define (kochline p1 p2 dpt)
    (if (= dpt 0)
        (make-segment (coords-c p1) (coords-c p2))
        (let* ((delta (/ (- p2 p1) 3))
	       (pn1 (+ p1 delta))
	       (pn2 (+ pn1 (make-polar (magnitude delta) (+ (angle delta) (/ pi 3)))))
	       (pn3 (+ pn1 delta))
	       (nd (- dpt 1)))
	  (kochline p1 pn1 nd)
	  (kochline pn1 pn2 nd)
	  (kochline pn2 pn3 nd)
	  (kochline pn3 p2 nd))))
  (kochline (make-rectangular x1 y1) (make-rectangular x2 y2) dpt)
  (kochline (make-rectangular x2 y2) (make-rectangular x3 y3) dpt)
  (kochline (make-rectangular x3 y3) (make-rectangular x1 y1) dpt))
