; Sierpinski triangle

; Creates Sierpinski triangle hanging on three hangers h1 h2 and h3 (vertices of
; triangle hangs on these hangers). *iter* gives depth of approximation (number
; of iterations) and *par* is relative positon (0-1) of vertices of internal
; triangles with regard to edges of external triangle (0.5 in classical
; Sierpinski triangle)

; example for: segments, parametric points, recursion (named let)

; named let is used for implementation of internal recursion. Meaning of this
; subprocedure (rec) is : "Do (*i* - 1) divisions of triangle given by
; *(s1, s1f, s1t)*, *(s2, s2f, s2t)* and *(s3, s3f, s3t)*", where
; *(sn, snf, snt)* is triplet specifying edge of triangle - sub-segment
; of segment *sn*, from *snf*  to *snt* (relative coordinates (0-1) at segment
; *sn*. This implementation allows to reuse old segments and prevents
; conglomeration of many overlapping segments (which happens in more
; straighforward implementations).

; example: (sierpinski (coords -60 -60) (coords 60 -60) (coords 0 60) 4 0.5)

(define (split from to par)
  (+ from (* (- to from) par)))

(define (sierpinski h1 h2 h3 iter par)
  (let rec ((s1 (make-segment h1 h2)) (s1f 0) (s1t 1)
            (s2 (make-segment h2 h3)) (s2f 0) (s2t 1)
	    (s3 (make-segment h3 h1)) (s3f 0) (s3t 1)
	    (i iter))
	(if (> i 1)
	  (let ((p1 (split s1f s1t par))
	        (p2 (split s2f s2t par))
	        (p3 (split s3f s3t par))
	        (i2 (- i 1)))
	    (let ((ph1 (h-center (make-parametric-point s1 p1)))
	          (ph2 (h-center (make-parametric-point s2 p2)))
	          (ph3 (h-center (make-parametric-point s3 p3))))
  	      (rec s1 s1f p1
  	           (make-segment ph1 ph3) 0 1
	  	   s3 p3 s3t i2)
	      (rec s1 p1 s1t
		   s2 s2f p2
	           (make-segment ph2 ph1) 0 1 i2)
	      (rec (make-segment ph3 ph2) 0 1
	           s2 p2 s2t
		   s3 s3f p3 i2))))))

