/*
 * VRR FONTLIB main module
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include "font/internal.h"
#include "build/path.h"
#include "lib/garr.h"
#include <fontconfig/fontconfig.h>

#define string kpathsea_string
#include <kpathsea/kpathsea.h>
#undef string


#ifndef LFT_LIB
#error You must define LFT_LIB as the path to freetype shared library.
#endif

/*
 * Bloody hack - we use our own copy of freetype library and because of
 * conflict with GTK library we need to load all symbols manually.
 */

#include <dlfcn.h>

#ifdef __GNUC__
#define TYPEOF(x) __typeof__(x)
TYPEOF(FT_Activate_Size) *font_Activate_Size;
TYPEOF(FTC_CMapCache_Lookup) *font_CMapCache_Lookup;
TYPEOF(FTC_CMapCache_New) *font_CMapCache_New;
TYPEOF(FTC_ImageCache_Lookup) *font_ImageCache_Lookup;
TYPEOF(FTC_ImageCache_New) *font_ImageCache_New;
TYPEOF(FTC_Manager_Done) *font_Manager_Done;
TYPEOF(FTC_Manager_LookupFace) *font_Manager_LookupFace;
TYPEOF(FTC_Manager_LookupSize) *font_Manager_LookupSize;
TYPEOF(FTC_Manager_New) *font_Manager_New;
TYPEOF(FT_Done_Face) *font_Done_Face;
TYPEOF(FT_Done_FreeType) *font_Done_FreeType;
TYPEOF(FT_Done_Glyph) *font_Done_Glyph;
TYPEOF(FT_Get_Glyph) *font_Get_Glyph;
TYPEOF(FT_Get_Glyph_Name) *font_Get_Glyph_Name;
TYPEOF(FT_Get_Char_Index) *font_Get_Char_Index;
TYPEOF(FT_Get_Name_Index) *font_Get_Name_Index;
TYPEOF(FT_Get_Postscript_Name) *font_Get_Postscript_Name;
TYPEOF(FT_Get_PS_Font_Info) *font_Get_PS_Font_Info;
TYPEOF(FT_Get_Sfnt_Name_Count) *font_Get_Sfnt_Name_Count;
TYPEOF(FT_Get_Sfnt_Name) *font_Get_Sfnt_Name;
TYPEOF(FT_Get_Sfnt_Table) *font_Get_Sfnt_Table;
TYPEOF(FT_Glyph_Get_CBox) *font_Glyph_Get_CBox;
TYPEOF(FT_Glyph_To_Bitmap) *font_Glyph_To_Bitmap;
TYPEOF(FT_Glyph_Transform) *font_Glyph_Transform;
TYPEOF(FT_Init_FreeType) *font_Init_FreeType;
TYPEOF(FT_Load_Glyph) *font_Load_Glyph;
TYPEOF(FT_Load_Char) *font_Load_Char;
TYPEOF(FT_Load_Sfnt_Table) *font_Load_Sfnt_Table;
TYPEOF(FT_New_Face) *font_New_Face;
TYPEOF(FT_RoundFix) *font_RoundFix;
TYPEOF(FT_Select_Charmap) *font_Select_Charmap;
TYPEOF(FT_Outline_Decompose) *font_Outline_Decompose;
TYPEOF(FT_Glyph_Copy) *font_Glyph_Copy;

#else

#define TYPEOF(x)
FT_Error (*font_Activate_Size)(FT_Size size);
FT_UInt (*font_CMapCache_Lookup)(FTC_CMapCache, FTC_FaceID, FT_Int, FT_UInt32);
FT_Error (*font_CMapCache_New)(FTC_Manager, FTC_CMapCache *);
FT_Error (*font_ImageCache_Lookup)(FTC_ImageCache, FTC_ImageType, FT_UInt, FT_Glyph *, FTC_Node *);
FT_Error (*font_ImageCache_New)(FTC_Manager, FTC_ImageCache  *);
void (*font_Manager_Done)(FTC_Manager);
FT_Error (*font_Manager_LookupFace)(FTC_Manager, FTC_FaceID, FT_Face *);
FT_Error (*font_Manager_LookupSize)(FTC_Manager, FTC_Scaler, FT_Size *);
FT_Error (*font_Manager_New)(FT_Library library, FT_UInt, FT_UInt, FT_ULong, FTC_Face_Requester, FT_Pointer, FTC_Manager *);
FT_Error (*font_Done_Face)(FT_Face);
FT_Error (*font_Done_FreeType)(FT_Library);
void (*font_Done_Glyph)(FT_Glyph);
FT_Error (*font_Get_Glyph)(FT_GlyphSlot, FT_Glyph *);
FT_Error (*font_Get_Glyph_Name)(FT_Face, FT_UInt, FT_Pointer, FT_UInt);
FT_UInt (*font_Get_Char_Index)(FT_Face);
FT_UInt (*font_Get_Name_Index)(FT_Face, FT_String *);
const char *(*font_Get_Postscript_Name)(FT_Face);
FT_Error (*font_Get_PS_Font_Info)(FT_Face, PS_FontInfoRec *);
FT_UInt (*font_Get_Sfnt_Name_Count)(FT_Face);
FT_Error (*font_Get_Sfnt_Name)(FT_Face, FT_UInt, FT_SfntName *);
void *(*font_Get_Sfnt_Table)(FT_Face, FT_Sfnt_Tag);
void (*font_Glyph_Get_CBox)(FT_Glyph, FT_UInt, FT_BBox *);
FT_Error (*font_Glyph_To_Bitmap)(FT_Glyph *, FT_Render_Mode, FT_Vector *, FT_Bool);
FT_Error (*font_Glyph_Transform)(FT_Glyph, FT_Matrix *, FT_Vector *);
FT_Error (*font_Init_FreeType)(FT_Library *);
FT_Error (*font_Load_Glyph)(FT_Face, FT_UInt, FT_Int32);
FT_Error (*font_Load_Char)(FT_Face, FT_ULong, FT_Int32);
FT_Error (*font_Load_Sfnt_Table)(FT_Face, FT_ULong, FT_Long, FT_Byte *, FT_ULong *);
FT_Error (*font_New_Face)(FT_Library, const char *, FT_Long, FT_Face *);
FT_Fixed (*font_RoundFix)(FT_Fixed);
FT_Error (*font_Select_Charmap)(FT_Face, FT_Encoding);
FT_Error (*font_Outline_Decompose)(FT_Outline *, const FT_Outline_Funcs *, void *);
FT_Error (*font_Glyph_Copy)(FT_Glyph, FT_Glyph *);

#endif

static void *lft_handle;

FT_Library font_lib;
FTC_Manager font_mgr;
FTC_ImageCache font_img_cache;
FTC_CMapCache font_cmap_cache;

struct fntid {
	uns id;
	char fname[1];
};

#define HASH_PREFIX(x) fntid_##x
#define HASH_NODE struct fntid
#define HASH_KEY_ENDSTRING fname
#define HASH_WANT_CLEANUP
#define HASH_WANT_NEW
#define HASH_WANT_FIND
#include "lib/sht.h"

static struct garr *fnttbl;
static uns max_fntid;

#define GARR_PREFIX(x) fnttbl_##x
#define GARR_TYPE char *
#include "lib/garr.h"


static char *xstrncpy(char *dest, const char *src, size_t n)
{
	if(src)
		return strncpy(dest, src, n);
	dest[0] = 0;
	return NULL;
}

static FT_Error face_request(FTC_FaceID face_id, FT_Library library, FT_Pointer request_data UNUSED, FT_Face *aface)
{
	if (!(addr_int_t)face_id)
		return 1;
	char *fname = *fnttbl_get(fnttbl, (addr_int_t)face_id);
	FT_Error r;

	r = font_New_Face(library, fname, 0, aface);
	return r;
}

static void *symload(const char *sym)
{
	void *r = dlsym(lft_handle, sym);

	if (!r)
		die("Unable to load symbol '%s'\n", sym);
	return r;
}

void font_init(const char* vrr_name)
{
	lft_handle = dlopen(LFT_LIB, RTLD_LOCAL | RTLD_NOW);
	if (!lft_handle)
		die("Error loading shared library '%s'\n", LFT_LIB);

	font_Activate_Size = (TYPEOF(font_Activate_Size))symload("FT_Activate_Size");
	font_CMapCache_Lookup = (TYPEOF(font_CMapCache_Lookup))symload("FTC_CMapCache_Lookup");
	font_CMapCache_New = (TYPEOF(font_CMapCache_New))symload("FTC_CMapCache_New");
	font_Done_Face = (TYPEOF(font_Done_Face))symload("FT_Done_Face");
	font_Done_FreeType = (TYPEOF(font_Done_FreeType))symload("FT_Done_FreeType");
	font_Done_Glyph = (TYPEOF(font_Done_Glyph))symload("FT_Done_Glyph");
	font_Get_Glyph_Name = (TYPEOF(font_Get_Glyph_Name))symload("FT_Get_Glyph_Name");
	font_Get_Glyph = (TYPEOF(font_Get_Glyph))symload("FT_Get_Glyph");
	font_Get_Char_Index = (TYPEOF(font_Get_Char_Index))symload("FT_Get_Char_Index");
	font_Get_Name_Index = (TYPEOF(font_Get_Name_Index))symload("FT_Get_Name_Index");
	font_Get_Postscript_Name = (TYPEOF(font_Get_Postscript_Name))symload("FT_Get_Postscript_Name");
	font_Get_PS_Font_Info = (TYPEOF(font_Get_PS_Font_Info))symload("FT_Get_PS_Font_Info");
	font_Get_Sfnt_Name_Count = (TYPEOF(font_Get_Sfnt_Name_Count))symload("FT_Get_Sfnt_Name_Count");
	font_Get_Sfnt_Name = (TYPEOF(font_Get_Sfnt_Name))symload("FT_Get_Sfnt_Name");
	font_Get_Sfnt_Table = (TYPEOF(font_Get_Sfnt_Table))symload("FT_Get_Sfnt_Table");
	font_Glyph_Get_CBox = (TYPEOF(font_Glyph_Get_CBox))symload("FT_Glyph_Get_CBox");
	font_Glyph_To_Bitmap = (TYPEOF(font_Glyph_To_Bitmap))symload("FT_Glyph_To_Bitmap");
	font_Glyph_Transform = (TYPEOF(font_Glyph_Transform))symload("FT_Glyph_Transform");
	font_ImageCache_Lookup = (TYPEOF(font_ImageCache_Lookup))symload("FTC_ImageCache_Lookup");
	font_ImageCache_New = (TYPEOF(font_ImageCache_New))symload("FTC_ImageCache_New");
	font_Init_FreeType = (TYPEOF(font_Init_FreeType))symload("FT_Init_FreeType");
	font_Load_Glyph = (TYPEOF(font_Load_Glyph))symload("FT_Load_Glyph");
	font_Load_Char = (TYPEOF(font_Load_Char))symload("FT_Load_Char");
	font_Load_Sfnt_Table = (TYPEOF(font_Load_Sfnt_Table))symload("FT_Load_Sfnt_Table");
	font_Manager_Done = (TYPEOF(font_Manager_Done))symload("FTC_Manager_Done");
	font_Manager_LookupFace = (TYPEOF(font_Manager_LookupFace))symload("FTC_Manager_LookupFace");
	font_Manager_LookupSize = (TYPEOF(font_Manager_LookupSize))symload("FTC_Manager_LookupSize");
	font_Manager_New = (TYPEOF(font_Manager_New))symload("FTC_Manager_New");
	font_New_Face = (TYPEOF(font_New_Face))symload("FT_New_Face");
	font_RoundFix = (TYPEOF(font_RoundFix))symload("FT_RoundFix");
	font_Select_Charmap = (TYPEOF(font_Select_Charmap))symload("FT_Select_Charmap");
	font_Outline_Decompose = (TYPEOF(font_Outline_Decompose))symload("FT_Outline_Decompose");
	font_Glyph_Copy = (TYPEOF(font_Glyph_Copy))symload("FT_Glyph_Copy");

	if (font_Init_FreeType(&font_lib))
		die("FreeType library initialization error\n");
	if (font_Manager_New(font_lib, 0, 0, 0, face_request, NULL, &font_mgr))
		die("FreeType font caching subsystem initialization error\n");
	if (FcInit() != FcTrue)
		die("Fontconfig initialization error\n");

	fntid_init();
	fnttbl = fnttbl_new(32);
	kpse_set_program_name(vrr_name, NULL);
}

void font_finish(void)
{
	fnttbl_destroy(fnttbl);
	fntid_cleanup();
	font_Manager_Done(font_mgr);
	font_Done_FreeType(font_lib);
	dlclose(lft_handle);
}

int font_id_valid(int fnt_id)
{
	FT_Face face;

	return fnt_id>0 && !font_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face);
}

const char *font_filename(int fnt_id)
{
	if (!fnt_id)
		return NULL;
	return *fnttbl_get(fnttbl, fnt_id);
}

int font_load_file(const char *fname)
{
	FT_Face face;

	struct fntid *f = fntid_find((char*)fname);

	if (!f) {
		/* Try to load the font to check if it is usable by Freetype */
		if (font_New_Face(font_lib, fname, 0, &face))
			return 0;	// it isn't...
		font_Done_Face(face);
			
		f = fntid_new((char*)fname);
		f->id = ++max_fntid;
		*fnttbl_get(fnttbl, f->id) = f->fname;
	}
	return f->id;
}

#define POINT2MM (25.4/72.0)

int font_get_glyph_advances(int font_id, struct garr *width)
{
	FT_Face face;
	int i;
	int w;

	if (font_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)font_id, &face))
		return -1;
	
	DBG("Scanning widths of font %s\n", font_filename(font_id));
	for (i=0; i < face->num_glyphs; i++) {
		if (font_Load_Glyph(face, i, FT_LOAD_NO_SCALE)) {
			DBG("Error loading glyph %d!\n", i);
			w = 0;
		}
		else {
			w = face->glyph->metrics.width;
			DBG("glyph %d metrics: width=%d hbx=%d hby=%d ha=%d lha=%d ax=%d\n", i, (int)face->glyph->metrics.width,
					(int)face->glyph->metrics.horiBearingX, (int)face->glyph->metrics.horiBearingY,
					(int)face->glyph->metrics.horiAdvance, (int)face->glyph->linearHoriAdvance,
					(int)face->glyph->advance.x);
		}
		*(real *)garr_get(width, i) = POINT2MM * w;
	}
	return face->num_glyphs;
}

void font_info_dump(int fnt_id)
{
	FT_Face face;
	uns flg;
	char buf[200];
	int i;
	struct font_t1_info fi;

	if (font_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face)) {
		msg("Font %d not found\n", fnt_id);
		return;
	}

	msg("Font %d, family '%s', style '%s':\n", fnt_id,
		face->family_name ? face->family_name : "", face->style_name ? face->style_name : "");
	msg("%d glyphs, %d charmaps, %d fixed sizes\n", (int)face->num_glyphs, face->num_charmaps, face->num_fixed_sizes);
	const char *name = font_Get_Postscript_Name(face);
	if (name)
		msg("PS name: '%s'\n", name);
	if (font_t1_info(fnt_id, &fi)) {
		msg("T1 info:\n");
		msg("family '%s'\n", fi.family);
		msg("style '%s'\n", fi.style);
		msg("full_name '%s'\n", fi.full_name);
		msg("family_name '%s'\n", fi.family_name);
	}

	flg = face->face_flags;
	msg("Flags: %s%s%s%s%s%s%s%s%s\n",
		flg & FT_FACE_FLAG_SCALABLE ? "scalable " : "",
		flg & FT_FACE_FLAG_FIXED_SIZES ? "fixed_sizes " : "",
		flg & FT_FACE_FLAG_FIXED_WIDTH ? "fixed_width " : "",
		flg & FT_FACE_FLAG_SFNT ? "sfnt " : "",
		flg & FT_FACE_FLAG_HORIZONTAL ? "horizontal " : "",
		flg & FT_FACE_FLAG_VERTICAL ? "vertical " : "",
		flg & FT_FACE_FLAG_KERNING ? "kerning " : "",
		flg & FT_FACE_FLAG_MULTIPLE_MASTERS ? "multimasters " : "",
		flg & FT_FACE_FLAG_GLYPH_NAMES ? "glyph_names " : "");

	msg("Style: %s%s\n",
		face->style_flags & FT_STYLE_FLAG_ITALIC ? "italic " : "",
		face->style_flags & FT_STYLE_FLAG_BOLD ? "bold " : "");
	for (i=0; i < face->num_charmaps; i++) {
		u32 e = face->charmaps[i]->encoding;
		msg("Charmap %d: encoding=%c%c%c%c, platform_id=%d encoding_id=%d\n", i, e >> 24, (e >> 16) & 255, (e >> 8) & 255, e & 255,
				face->charmaps[i]->platform_id, face->charmaps[i]->encoding_id);
	}

	if (FT_HAS_GLYPH_NAMES(face)) {
		msg("Glyph names:\n");
		for (i=0; i < face->num_glyphs; i++) {
			font_Get_Glyph_Name(face, i, buf, sizeof(buf));
			msg("%d: %s\n", i, buf);
		}
	}
}

static int detect_type(FT_Face face)
{
	PS_FontInfoRec psfi;

	if (!font_Get_PS_Font_Info(face, &psfi))
		return FONT_TYPE_TYPE1;

	if (font_Get_Sfnt_Table(face, ft_sfnt_head))
		return FONT_TYPE_TRUETYPE;

	return FONT_TYPE_UNKNOWN;
}

int font_get_type(int fnt_id)
{
	FT_Face face;

	if (font_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face))
		return FONT_TYPE_INVALID;
	
	return detect_type(face);
}

static struct font_info *fnt_info(FT_Face face, struct font_info *fnt_info)
{
	struct font_info fi;

	if (!(face->face_flags & FT_FACE_FLAG_SCALABLE))
		return NULL;
	fi.type = detect_type(face);
	fi.italic = face->style_flags & FT_STYLE_FLAG_ITALIC;
	fi.bold = face->style_flags & FT_STYLE_FLAG_BOLD;
	fi.fixed_width = face->face_flags & FT_FACE_FLAG_FIXED_WIDTH;
	fi.glyphs = face->num_glyphs;
	fi.charmaps = face->num_charmaps;

	if (face->family_name) {
		strncpy(fi.family, face->family_name, sizeof(fi.family)-1);
		fi.family[sizeof(fi.family)-1] = 0;
	} else
		fi.family[0] = 0;

	if (face->style_name) {
		strncpy(fi.style, face->style_name, sizeof(fi.style)-1);
		fi.style[sizeof(fi.style)-1] = 0;
	} else
		fi.style[0] = 0;

	fi.fontid = -1;

	const char *name = font_Get_Postscript_Name(face);
	if (!name)
		goto end;
	strncpy(fi.name, name, sizeof(fi.name)-1);
	fi.name[sizeof(fi.name)-1] = 0;

	fi.bbox = (struct geom_rectangle){ { face->bbox.xMin, face->bbox.yMin }, { face->bbox.xMax, face->bbox.yMax } };
	fi.ascender = face->ascender;
	fi.descender = face->descender;
	
	*fnt_info = fi;
	return fnt_info;
end:
	return NULL;
}

struct font_info *font_info_temporary(const char *fontpath, struct font_info *fi)
{
	FT_Face face;

	ASSERT(fi);

	if (font_New_Face(font_lib, fontpath, 0, &face))
		return NULL;
	fnt_info(face, fi);
	font_Done_Face(face);
	return fi;
}

struct font_info *font_info(int fntid, struct font_info *fi)
{
	FT_Face face;

	ASSERT(fi);

	if (font_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fntid, &face))
		return NULL;
	return fnt_info(face, fi);
}

const char *font_get_ps_name(int fnt_id)
{
	FT_Face face;

	if (font_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face))
		return NULL;

	return font_Get_Postscript_Name(face);
}

struct font_t1_info *font_t1_info(int fnt_id, struct font_t1_info *fi)
{
	FT_Face face;
	PS_FontInfoRec psfi;

	ASSERT(fi);

	if (font_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face) || font_Get_PS_Font_Info(face, &psfi))
		return NULL;

	if (!(face->face_flags & FT_FACE_FLAG_SCALABLE))
		return NULL;
	fi->italic = face->style_flags & FT_STYLE_FLAG_ITALIC;
	fi->bold = face->style_flags & FT_STYLE_FLAG_BOLD;
	fi->glyphs = face->num_glyphs;
	fi->charmaps = face->num_charmaps;

	if (face->family_name) {
		strncpy(fi->family, face->family_name, sizeof(fi->family)-1);
		fi->family[sizeof(fi->family)-1] = 0;
	} else
		fi->family[0] = 0;

	if (face->style_name) {
		strncpy(fi->style, face->style_name, sizeof(fi->style)-1);
		fi->style[sizeof(fi->style)-1] = 0;
	} else
		fi->style[0] = 0;
	
	/* FIXME: check whether the bbox is 16/16 or 24/8 format */
	fi->bbox.ll.x = face->bbox.xMin / (double) 0x10000;
	fi->bbox.ll.y = face->bbox.yMin / (double) 0x10000;
	fi->bbox.ll.x = face->bbox.xMax / (double) 0x10000;
	fi->bbox.ll.y = face->bbox.xMax / (double) 0x10000;

	fi->units_per_em = face->units_per_EM;
	fi->ascender = face->ascender;
	fi->descender = face->descender;

	xstrncpy(fi->full_name, psfi.full_name, sizeof(fi->full_name)-1);
	xstrncpy(fi->family_name, psfi.family_name, sizeof(fi->family_name)-1);
	xstrncpy(fi->version, psfi.version, sizeof(fi->version)-1);
	xstrncpy(fi->notice, psfi.notice, sizeof(fi->notice)-1);
	xstrncpy(fi->weight, psfi.weight, sizeof(fi->weight)-1);
	fi->italic_angle = psfi.italic_angle;
	fi->is_fixed_pitch = psfi.is_fixed_pitch;
	fi->underline_position = psfi.underline_position;
	fi->underline_thickness = psfi.underline_thickness;

	return fi;
}

struct font_truetype_info *font_truetype_info(int fnt_id, struct font_truetype_info *tt)
{
	FT_Face face;
	FT_Sfnt_Tag tag = ft_sfnt_head;

	ASSERT(tt);

	if (font_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face) || !font_Get_Sfnt_Table(face, tag))
		return NULL;

	return tt;
}

uns font_get_max_fontid(void)
{
	return max_fntid;
}

char *font_search(const char *family, const char *style, char *outbuf, uns bufsize)
{
	FcPattern *patt, *fnt;
	FcResult res;
	FcChar8 *rv = NULL;
	
	patt = FcPatternCreate();
	if (!patt)
		return NULL;

	FcPatternAddBool(patt, FC_SCALABLE, FcTrue);
	if (family)
		FcPatternAddString(patt, FC_FAMILY, family);
	if (style)
		FcPatternAddString(patt, FC_STYLE, style);

	FcConfigSubstitute(NULL, patt, FcMatchFont);
	FcDefaultSubstitute(patt);
	fnt = FcFontMatch(NULL, patt, &res);
	/* It seems that fontconfig ignores res. How am I supposed to test the result? */

	FcPatternGetString(fnt, FC_FILE, 0, &rv);
	strncpy(outbuf, rv, bufsize-1);
	uns l = strlen(rv);
	if (l >= bufsize)
		outbuf[l-1] = 0;
	rv = outbuf;

	FcPatternDestroy(patt);
	FcPatternDestroy(fnt);
	return rv;
}

