#ifndef _VRR_GEOMLIB_ITEM_H_
#define _VRR_GEOMLIB_ITEM_H_

/** @file
 *
 * GEOMLIB - Base class for "items" in GEOMLIB.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "geomlib/vector.h"
#include "geomlib/cache.h"


#define GEOM_DUMP_CLASS     0x1
#define GEOM_DUMP_OWNER     0x2
#define GEOM_DUMP_ITEMS     0x4
#define GEOM_DUMP_RECURSIVE 0x8
#define GEOM_DUMP_ALL       0xffffffff


/* Item/group flags in instance header */

#define GEOM_ITEM_FLAGS(i) GEOM_INSTANCE_FLAGS2(i)

// Item is closed curve/path.
#define GEOM_ITEM_CLOSED     0x01

// Item is derived from group (fast group detection).
#define GEOM_ITEM_GROUP      0x02

// Item is derived from path (fast path detection).
#define GEOM_ITEM_PATH       0x04

// Zero flag. Can be used locally by algorithms, but must be cleared again.
#define GEOM_ITEM_ZERO_FLAG  0x80

#define GEOM_ITEM_EXPANSION_VALID 0x08


/* Class geom_item declaration. */

#define geom_item_VMT geom_o_VMT \
	\
	void (*create)(void *); \
	const void *(*assert)(const void *, GEOM_ASSERT_DEF); \
	int (*valid)(const void *); \
	void (*subdump)(const void *, uns); \
	void (*copy)(void *, void *); \
	void (*after_change)(void *); \
	\
	real (*tlength)(void *); \
	real (*alength)(void *); \
	int (*blength)(void *); \
	\
	real (*time_to_atime)(void *, real); \
	int (*times_to_atimes)(void *, uns, real *, real *); \
	real (*atime_to_time)(void *, real); \
	int (*atimes_to_times)(void *, uns, real *, real *); \
	\
	real (*time_to_ratime)(void *, real); \
	int (*times_to_ratimes)(void *, uns, real *, real *); \
	real (*ratime_to_time)(void *, real); \
	int (*ratimes_to_times)(void *, uns, real *, real *); \
	\
	int time_dif_btime; \
	real (*time_to_btime)(void *, struct geom_expansion *, real); \
	int (*times_to_btimes)(void *, struct geom_expansion *, uns, real *, real *); \
	real (*btime_to_time)(void *, struct geom_expansion *, real); \
	int (*btimes_to_times)(void *, struct geom_expansion *, uns, real *, real *); \
	\
	int (*expansion_append)(void *, struct geom_fpath *); \
	int (*to_bezier)(void *, struct geom_bezier *); \
	int (*conversion_append)(void *, uns, struct geom_fpath *); \
	\
	int (*split)(void *, real, void *, void *); \
	int (*split_middle)(void *, void *, void *); \
	int (*reverse)(void *, void *); \
	int (*part)(void *, real, real, void *); \
	\
	int (*transform)(void *, struct geom_transform *, void *); \
	\
	int (*point_start)(void *, struct geom_point *); \
	int (*point_end)(void *, struct geom_point *); \
	int (*point_at_time)(void *, real, struct geom_point *); \
	\
	int (*derivation_start)(void *, struct geom_vector *); \
	int (*derivation_end)(void *, struct geom_vector *); \
	int (*derivation_at_time)(void *, real, struct geom_vector *); \
	\
	int (*bbox)(void *, struct geom_rectangle *); \
	int (*nearest_to_point)(void *, struct geom_point *, uns, struct geom_nearest *); \
	int (*direction_times)(void *, struct geom_vector *, struct garr *); \
	int (*distance_times)(void *, struct geom_point *, real, struct garr *); \

#define geom_item_INIT geom_o_INIT \
	\
	.create = (void (*)(void *)) &geom_item_create, \
	.destroy = (void (*)(void *)) &geom_item_destroy, \
	.assert = (const void *(*)(const void *, GEOM_ASSERT_DEF)) &geom_item_assert, \
	.valid = (int (*)(const void *)) &geom_item_valid__default, \
	.subdump = (void (*)(const void *, uns)) &geom_item_subdump, \
	.copy = (void (*)(void *, void *)) &geom_item_copy__default, \
	.after_change = (void (*)(void *)) &geom_item_after_change__empty, \
	\
	.tlength = &geom_item_tlength__invalid, \
	.alength = &geom_item_alength__invalid, \
	.blength = &geom_item_blength__invalid, \
	\
	.time_to_atime = &geom_item_time_to_atime__invalid, \
	.times_to_atimes = &geom_item_times_to_atimes__invalid, \
	.atime_to_time = &geom_item_atime_to_time__invalid, \
	.atimes_to_times = &geom_item_atimes_to_times__invalid, \
	\
	.time_to_ratime = &geom_item_time_to_ratime__invalid, \
	.times_to_ratimes = &geom_item_times_to_ratimes__invalid, \
	.ratime_to_time = &geom_item_ratime_to_time__invalid, \
	.ratimes_to_times = &geom_item_ratimes_to_times__invalid, \
	\
	.time_to_btime = &geom_item_time_to_btime__invalid, \
	.times_to_btimes = &geom_item_times_to_btimes__invalid, \
	.btime_to_time = &geom_item_btime_to_time__invalid, \
	.btimes_to_times = &geom_item_btimes_to_times__invalid, \
	\
	.expansion_append = &geom_item_expansion_append__invalid, \
	.to_bezier = &geom_item_to_bezier__invalid, \
	.conversion_append = &geom_item_conversion_append__invalid, \
	\
	.transform = &geom_item_transform__invalid, \
	\
	.split = &geom_item_split__invalid, \
	.split_middle = &geom_item_split_middle__invalid, \
	.reverse = &geom_item_reverse__invalid, \
	.part = &geom_item_part__invalid, \
	\
	.point_start = &geom_item_point_start__invalid, \
	.point_end = &geom_item_point_end__invalid, \
	.point_at_time = &geom_item_point_at_time__invalid, \
	\
	.derivation_start = &geom_item_derivation_start__invalid, \
	.derivation_end = &geom_item_derivation_end__invalid, \
	.derivation_at_time = &geom_item_derivation_at_time__invalid, \
	\
	.bbox = &geom_item_bbox__invalid, \
	.nearest_to_point = &geom_item_nearest_to_point__invalid, \
	.direction_times = &geom_item_direction_times__invalid, \
	.distance_times = &geom_item_distance_times__invalid, \

GEOM_CLASS_HEAD(item, o);


/* Useful macros to define trivial methods. */
#define GEOM_ITEM_CALL(method) GEOM_INSTANCE_VMT(item, item, method)(item)
#define GEOM_ITEM_CALL2(method,...) GEOM_INSTANCE_VMT(item, item, method)(item, __VA_ARGS__)
#define GEOM_ITEM_CALL_RET(method) return GEOM_INSTANCE_VMT(item, item, method)(item)
#define GEOM_ITEM_CALL2_RET(method,...) return GEOM_INSTANCE_VMT(item, item, method)(item, __VA_ARGS__)

#define GEOM_ITEM_EDEF(method) \
	static inline int geom_item_##method##__invalid(void UNUSED *item) { GEOM_FUNC; GEOM_ERET(GEOM_ERR_ITEM_INVALID); }
#define GEOM_ITEM_EDEF2(method,...) \
	static inline int geom_item_##method##__invalid(void UNUSED *item, __VA_ARGS__) { GEOM_FUNC; GEOM_ERET(GEOM_ERR_ITEM_INVALID); }
#define GEOM_ITEM_RDEF(method) \
	static inline real geom_item_##method##__invalid(void UNUSED *item) { GEOM_FUNC; geom_errno = GEOM_ERR_ITEM_INVALID; GEOM_RRET(NAN); }
#define GEOM_ITEM_RDEF2(method,...) \
	static inline real geom_item_##method##__invalid(void UNUSED *item, __VA_ARGS__) { GEOM_FUNC; geom_errno = GEOM_ERR_ITEM_INVALID; GEOM_RRET(NAN); }


/** Item constructor. */
static inline void geom_item_create(struct geom_item UNUSED *item)
{ GEOM_INSTANCE_IS_ASSERT(item, item); }


/** Item destructor. */
static inline void geom_item_destroy(struct geom_item UNUSED *item)
{ GEOM_INSTANCE_IS_ASSERT(item, item); GEOM_BASSERT(item->owner == NULL); }


/*
 * Make assertions to test validity of a given item.
 *
 * @ai  item  Item pointer.
 * @ai  ...   Assertion parameters.
 */
const void *geom_item_assert(const void *item, GEOM_ASSERT_DEF);
const void *geom_item_call_assert(GEOM_ASSERT_DEF, const void *item, struct geom_class *pclass);
#define GEOM_ITEM_ASSERT(nclass,i) GEOM_ASSERT2(item_call,(void *)i,GEOM_CLASS(nclass))
#define GEOM_ITEM_ASSERTX(nclass,i) GEOM_ASSERTX2(item_call,(void *)i,GEOM_CLASS(nclass))


/**
 * Item validation.
 *
 * @ai  item  Item pointer.
 *
 * @rv  nonzero  Item is valid.
 * @rv  zero     Item is invalid.
 */
static inline int geom_item_valid(const void *item) { return GEOM_ITEM_CALL(valid); }
static inline int geom_item_valid__default(const void UNUSED *item) { return 1; }


/**
 * Dump a given item.
 *
 * @ai  item   Item pointer.
 * @ai  flags  Dump options (see GEOM_DUMP_x definitions).
 */
void geom_item_dump(const void *item, uns flags);
static inline void geom_item_subdump(const struct geom_item UNUSED *item, uns UNUSED flags) {}


/**
 * Makes an exact copy of another item.
 * Owner of destination item is not changed.
 * Can be used as item constructor.
 *
 * @ai  item  Destination item.
 * @ai  src   Source item.
 */
static inline void geom_item_copy(void *item, void *src) { GEOM_ITEM_CALL2(copy, src); }
void geom_item_copy__default(void *item, void *src);
void geom_item_copy__fail(void *item, void *src);


/**
 * Allocate a new item of same class type and copy data.
 * Resulting item should be destroyed by GEOM_INSTANCE_FREE or equivalent routine.
 *
 * @ai  item  Item pointer.
 *
 * @r  Duplicated item.
 */
void *geom_item_duplicate(void *item);


/**
 * Should be called after any manual change to item's data.
 * Invalidates cache of item and derived groups.
 *
 * @ai  item  Item pointer.
 */
static inline void geom_item_after_change(void *item) { GEOM_ITEM_CALL(after_change); }
static inline void geom_item_after_change__empty(void UNUSED *item) {}


/*----------------------------------------------*
 *                                              *
 *              PARAMETRIZATIONS                *
 *                                              *
 *----------------------------------------------*/


/**
 * Returns TIME parametrization length (1 for elementary curves).
 *
 * @ai  item  Item pointer.
 *
 * @rv  finite      TIME parametrization length.
 * @rv  otherwise   Failed.
 */
static inline real geom_item_tlength(void *item)
{ GEOM_ITEM_CALL_RET(tlength); }
GEOM_ITEM_RDEF(tlength)


/**
 * Returns Euclidean arc length (ATIME parametrization).
 *
 * @ai  item  Item pointer.
 *
 * @rv  finite      Euclidean arc length.
 * @rv  otherwise   Failed.
 */
static inline real geom_item_alength(void *item)
{ GEOM_ITEM_CALL_RET(alength); }
GEOM_ITEM_RDEF(alength)


/**
 * Returns BTIME parametrization length.
 *
 * @ai  item  Item pointer.
 *
 * @rv  non-negative  Number of Bezier curves in expansion.
 * @rv  negative      Failed.
 */
static inline int geom_item_blength(void *item)
{ GEOM_ITEM_CALL_RET(blength); }
GEOM_ITEM_EDEF(blength)


/* TIME <-> ATIME conversions */


/**
 * Conversion from TIME to ATIME parametrization.
 *
 * @ai  item  Item pointer.
 * @ai  time  Source parametrization value.
 *
 * @rv  finite     Destination parametrization value.
 * @rv  otherwise  Failed.
 */
static inline real geom_item_time_to_atime(void *item, real time)
{ GEOM_ITEM_CALL2_RET(time_to_atime, time); }
GEOM_ITEM_RDEF2(time_to_atime, real UNUSED time)


/**
 * Converstion from TIME to ATIME parametrization.
 *
 * @ai  item    Item pointer.
 * @ai  count   Number of parameters to convert.
 * @ai  times   Sorted array of source parametrization values.
 * @ao  atimes  Buffer for sorted destination parametrization values.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 */
static inline int geom_item_times_to_atimes(void *item, uns count, real *times, real *atimes)
{ GEOM_ITEM_CALL2_RET(times_to_atimes, count, times, atimes); }
GEOM_ITEM_EDEF2(times_to_atimes, uns UNUSED count, real UNUSED *times, real UNUSED *atimes)


/**
 * Conversion from ATIME to TIME parametrization.
 *
 * @ai  item   Item pointer.
 * @ai  atime  Source parametrization value.
 *
 * @rv  finite     Destination parametrization value.
 * @rv  otherwise  Failed.
 */
static inline real geom_item_atime_to_time(void *item, real atime)
{ GEOM_ITEM_CALL2_RET(atime_to_time, atime); }
GEOM_ITEM_RDEF2(atime_to_time, real UNUSED atime)


/**
 * Converstion from ATIME to TIME parametrization.
 *
 * @ai  item    Item pointer.
 * @ai  count   Number of parameters to convert.
 * @ai  atimes  Sorted array of source parametrization values.
 * @ao  times   Buffer for sorted destination parametrization values.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 */
static inline int geom_item_atimes_to_times(void *item, uns count, real *atimes, real *times)
{ GEOM_ITEM_CALL2_RET(atimes_to_times, count, atimes, times); }
GEOM_ITEM_EDEF2(atimes_to_times, uns UNUSED count, real UNUSED *atimes, real UNUSED *times)


/* TIME <-> RATIME conversions */


/**
 * Conversion from TIME to RATIME parametrization.
 *
 * @ai  item  Item pointer.
 * @ai  time  Source parametrization value.
 *
 * @rv  finite     Destination parametrization value.
 * @rv  otherwise  Failed.
 */
static inline real geom_item_time_to_ratime(void *item, real time)
{ GEOM_ITEM_CALL2_RET(time_to_ratime, time); }
GEOM_ITEM_RDEF2(time_to_ratime, real UNUSED time)


/**
 * Converstion from TIME to RATIME parametrization.
 *
 * @ai  item     Item pointer.
 * @ai  count    Number of parameters to convert.
 * @ai  times    Sorted array of source parametrization values.
 * @ao  ratimes  Buffer for sorted destination parametrization values.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 */
static inline int geom_item_times_to_ratimes(void *item, uns count, real *times, real *ratimes)
{ GEOM_ITEM_CALL2_RET(times_to_ratimes, count, times, ratimes); }
GEOM_ITEM_EDEF2(times_to_ratimes, uns UNUSED count, real UNUSED *times, real UNUSED *ratimes)


/**
 * Conversion from RATIME to TIME parametrization.
 *
 * @ai  item    Item pointer.
 * @ai  ratime  Source parametrization value.
 *
 * @rv  finite     Destination parametrization value.
 * @rv  otherwise  Failed.
 */
static inline real geom_item_ratime_to_time(void *item, real ratime)
{ GEOM_ITEM_CALL2_RET(ratime_to_time, ratime); }
GEOM_ITEM_RDEF2(ratime_to_time, real UNUSED ratime)


/**
 * Converstion from RATIME to TIME parametrization.
 *
 * @ai  item     Item pointer.
 * @ai  count    Number of parameters to convert.
 * @ai  ratimes  Sorted array of source parametrization values.
 * @ao  times    Buffer for sorted destination parametrization values.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 */
static inline int geom_item_ratimes_to_times(void *item, uns count, real *ratimes, real *times)
{ GEOM_ITEM_CALL2_RET(ratimes_to_times, count, ratimes, times); }
GEOM_ITEM_EDEF2(ratimes_to_times, uns UNUSED count, real UNUSED *times, real UNUSED *ratimes)


/* TIME <-> BTIME conversions */


/**
 * Returns whether TIME and BTIME are different.
 *
 * @ai  item  Item pointer.
 *
 * @rv  nonzero  Parametrizations are different.
 * @rv  zero     Parametrizations are the same.
 */
static inline int geom_item_time_dif_btime(void *item)
{ return GEOM_INSTANCE_VMT(item, item, time_dif_btime); }


/**
 * Conversion from TIME to BTIME parametrization.
 *
 * @ai  item  Item pointer.
 * @ai  time  Source parametrization value.
 *
 * @rv  finite     Destination parametrization value.
 * @rv  otherwise  Failed.
 */
static inline real geom_item_time_to_btime(void *item, struct geom_expansion *expansion, real time)
{ GEOM_ITEM_CALL2_RET(time_to_btime, expansion, time); }
GEOM_ITEM_RDEF2(time_to_btime, struct geom_expansion UNUSED *expansion, real UNUSED time)


/**
 * Converstion from TIME to BTIME parametrization.
 *
 * @ai  item    Item pointer.
 * @ai  count   Number of parameters to convert.
 * @ai  times   Sorted array of source parametrization values.
 * @ao  btimes  Buffer for sorted destination parametrization values.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 */
static inline int geom_item_times_to_btimes(void *item, struct geom_expansion *expansion, uns count, real *times, real *btimes)
{ GEOM_ITEM_CALL2_RET(times_to_btimes, expansion, count, times, btimes); }
GEOM_ITEM_EDEF2(times_to_btimes, struct geom_expansion UNUSED *expansion, uns UNUSED count, real UNUSED *times, real UNUSED *btimes)


/**
 * Conversion from BTIME to TIME parametrization.
 *
 * @ai  item   Item pointer.
 * @ai  btime  Source parametrization value.
 *
 * @rv  finite     Destination parametrization value.
 * @rv  otherwise  Failed.
 */
static inline real geom_item_btime_to_time(void *item, struct geom_expansion *expansion, real btime)
{ GEOM_ITEM_CALL2_RET(btime_to_time, expansion, btime); }
GEOM_ITEM_RDEF2(btime_to_time, struct geom_expansion UNUSED *expansion, real UNUSED btime)


/**
 * Converstion from BTIME to TIME parametrization.
 *
 * @ai  item    Item pointer.
 * @ai  count   Number of parameters to convert.
 * @ai  btimes  Sorted array of source parametrization values.
 * @ao  times   Buffer for sorted destination parametrization values.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 */
static inline int geom_item_btimes_to_times(void *item, struct geom_expansion *expansion, uns count, real *btimes, real *times)
{ GEOM_ITEM_CALL2_RET(btimes_to_times, expansion, count, btimes, times); }
GEOM_ITEM_EDEF2(btimes_to_times, struct geom_expansion UNUSED *expansion, uns UNUSED count, real UNUSED *btimes, real UNUSED *times)


/*----------------------------------------------*
 *                                              *
 *     CONVERSION TO OTHER TYPES OF CURVES      *
 *                                              *
 *----------------------------------------------*/


 static inline const struct geom_expansion *geom_expansion_assert(GEOM_ASSERT_DEF, const struct geom_expansion *expansion)
{ geom_p_assert(GEOM_ASSERT_COPY, expansion); return expansion; }


/**
 * Expand curve to tail of expansion path.
 * If curve is closed, input path should be empty and function should close resulting path.
 *
 * @ai   item   Item pointer.
 * @aio  bpath  Bezier path, where to append Bezier curves.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 */
static inline int geom_item_expansion_append(void *item, struct geom_fpath *expansion)
{ GEOM_ITEM_CALL2_RET(expansion_append, expansion); }
GEOM_ITEM_EDEF2(expansion_append, struct geom_fpath UNUSED *expansion)


/**
 * Makes one reference to item expansion.
 * If item cannot be expanded to Bezier path, result is NULL and no reference is made.
 *
 * @rv  NULL       Cannot expand. Error code is stored in geom_errno.
 * @rv  otherwise  Pointer to referenced expansion.
 */
struct geom_expansion *geom_item_expand(void *item);


/**
 * Unref one reference to expansion structure.
 * If number of references decreased to zero, expansion can be destroyed anytime.
 */
void geom_expansion_unref(struct geom_expansion *ref);


/**
 * Convert curve to single Bezier curve, if possible.
 *
 * @ai  item    Item pointer.
 * @ao  bezier  Resulting bezier curve (initialized).
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 * @rv  ...
 */
static inline int geom_item_to_bezier(void *item, struct geom_bezier *bezier)
{ GEOM_ITEM_CALL2_RET(to_bezier, bezier); }
GEOM_ITEM_EDEF2(to_bezier, struct geom_bezier UNUSED *bezier)


static inline int geom_item_conversion_append(void *item, uns flags, struct geom_fpath *result)
{ GEOM_ITEM_CALL2_RET(conversion_append, flags, result); }
GEOM_ITEM_EDEF2(conversion_append, uns UNUSED flags, struct geom_fpath UNUSED *result)

int geom_item_convert(void *item, uns flags, struct geom_fpath *result);


/* FIXME: stary interface pro jednoduchy prechod k novym cestam, pomale! */
int geom_old_expand(void *item, struct geom_path *path);



/*----------------------------------------------*
 *                                              *
 *                    SPLITS                    *
 *                                              *
 *----------------------------------------------*/


static inline int geom_item_split(void *item, real time, void *left, void *right)
{ GEOM_ITEM_CALL2_RET(split, time, left, right); }
GEOM_ITEM_EDEF2(split, real UNUSED time, void UNUSED *left, void UNUSED *right)

static inline int geom_item_split_middle(void *item, void *left, void *right)
{ GEOM_ITEM_CALL2_RET(split_middle, left, right); }
GEOM_ITEM_EDEF2(split_middle, void UNUSED *left, void UNUSED *right)

static inline int geom_item_reverse(void *item, void UNUSED *result)
{ GEOM_ITEM_CALL2_RET(reverse, result); }
GEOM_ITEM_EDEF2(reverse, void UNUSED *result)

static inline int geom_item_part(void *item, real left_time, real right_time, void UNUSED *result)
{ GEOM_ITEM_CALL2_RET(part, left_time, right_time, result); }
GEOM_ITEM_EDEF2(part, real UNUSED left_time, real UNUSED right_time, void UNUSED *result)


/*----------------------------------------------*
 *                                              *
 *                TRANSFORMATIONS               *
 *                                              *
 *----------------------------------------------*/


/**
 * Apply affine transformation to a given curve.
 *
 * @ai  item       Item pointer.
 * @ai  transform  Transformation matrix.
 * @ao  dest       Item of same type (possibly item=dest).
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 */
static inline int geom_item_transform(void *item, struct geom_transform *transform, void *dest)
{ return GEOM_ITEM_CALL2(transform, transform, dest); }
GEOM_ITEM_EDEF2(transform, struct geom_transform UNUSED *transform, void UNUSED *dest)


/*----------------------------------------------*
 *                                              *
 *             GEOMETRIC PROPERTIES             *
 *                                              *
 *----------------------------------------------*/


/**
 * Returns whether item is a closed curve.
 *
 * @ai  item  Item pointer.
 *
 * @rv  1  Item is a closed curve.
 * @rv  0  Item is not a closed curve.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 * @rv  GEOM_ERR_ITEM_NOT_IMPLEMENTED  Function not implemented, yet.
 */
static inline int geom_item_is_closed(void *item)
{  return GEOM_ITEM_FLAGS(item) & GEOM_ITEM_CLOSED; }


/**
 * Returns starting point of given curve.
 *
 * @ai  item    Item pointer.
 * @ao  result  Resulting point.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 * @rv  GEOM_ERR_PATH_EMPTY            Path is empty.
 */
static inline int geom_item_point_start(void *item, struct geom_point *result)
{ return GEOM_ITEM_CALL2(point_start, result); }
GEOM_ITEM_EDEF2(point_start, struct geom_point UNUSED *result)


/**
 * Returns ending point of given curve.
 *
 * @ai  item    Item pointer.
 * @ao  result  Resulting point.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 * @rv  GEOM_ERR_PATH_EMPTY            Path is empty.
 */
static inline int geom_item_point_end(void *item, struct geom_point *result)
{ return GEOM_ITEM_CALL2(point_end, result); }
GEOM_ITEM_EDEF2(point_end, struct geom_point UNUSED *result)


/**
 * Returns curve point at given parameter.
 *
 * @ai  item    Item pointer.
 * @ai  time    TIME parameter.
 * @ao  result  Resulting point.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 * @rv  GEOM_ERR_PATH_EMPTY            Path is empty.
 */
static inline int geom_item_point_at_time(void *item, real time, struct geom_point *result)
{ return GEOM_ITEM_CALL2(point_at_time, time, result); }
GEOM_ITEM_EDEF2(point_at_time, real UNUSED time, struct geom_point UNUSED *result)


/**
 * Returns first derivation in curve starting point.
 *
 * @ai  item    Item pointer.
 * @ao  result  Resulting vector.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 * @rv  GEOM_ERR_PATH_EMPTY            Path is empty.
 */
static inline int geom_item_derivation_start(void *item, struct geom_vector *result)
{ return GEOM_ITEM_CALL2(derivation_start, result); }
GEOM_ITEM_EDEF2(derivation_start, struct geom_vector UNUSED *result)


/**
 * Returns first derivation in curve ending point.
 *
 * @ai  item    Item pointer.
 * @ao  result  Resulting vector.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 * @rv  GEOM_ERR_PATH_EMPTY            Path is empty.
 */
static inline int geom_item_derivation_end(void *item, struct geom_vector *result)
{ return GEOM_ITEM_CALL2(derivation_end, result); }
GEOM_ITEM_EDEF2(derivation_end, struct geom_vector UNUSED *result)


/**
 * Returns first derivation in a given curve point.
 *
 * @ai  item    Item pointer.
 * @ai  time    TIME parameter.
 * @ao  result  Resulting vector.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 * @rv  GEOM_ERR_PATH_EMPTY            Path is empty.
 */
static inline int geom_item_derivation_at_time(void *item, real time, struct geom_vector *result)
{ return GEOM_ITEM_CALL2(derivation_at_time, time, result); }
GEOM_ITEM_EDEF2(derivation_at_time, real UNUSED time, struct geom_vector UNUSED *result)


/**
 * Returns bounding box of a given curve.
 *
 * @ai  item  Item pointer.
 * @ai  bbox  Resulting bounding box.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 * @rv  GEOM_ERR_PATH_EMPTY            Path is empty.
 */
static inline int geom_item_bbox(void *item, struct geom_rectangle *bbox)
{ return GEOM_ITEM_CALL2(bbox, bbox); }
GEOM_ITEM_EDEF2(bbox, struct geom_rectangle UNUSED *bbox)


/**
 * Finds nearest point on curve to a given point.
 *
 * @ai  item     Item pointer.
 * @ai  flags    Wanted entries in resulting structure (see GEOM_NEAREST_x flags).
 * @ao  nearest  Resulting structure.
 *
 * @rv  zero                           Success.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 * @rv  GEOM_ERR_PATH_EMPTY            Path is empty.
 */
static inline int geom_item_nearest_to_point(void *item, struct geom_point *point, uns flags, struct geom_nearest *nearest)
{ return GEOM_ITEM_CALL2(nearest_to_point, point, flags, nearest); }
GEOM_ITEM_EDEF2(nearest_to_point, struct geom_point UNUSED *point, uns UNUSED flags, struct geom_nearest UNUSED *nearest)


/**
 * Find TIME parameters, with derivation vector parallel to a given vector.
 * Parameters (reals) are added to current garr position.
 *
 * @ai  item    Item pointer.
 * @ai  dir     Direction vector.
 * @ai  result  Growing array, where to store parameters.
 *
 * @rv  >=0                            Success - number of found parameters.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 */
static inline int geom_item_direction_times(void *item, struct geom_vector *dir, struct garr *result)
{ return GEOM_ITEM_CALL2(direction_times, dir, result); }
GEOM_ITEM_EDEF2(direction_times, struct geom_vector UNUSED *dir, struct garr UNUSED *result)


/**
 * Find TIME parameters, with derivation vector parallel to X-axis.
 * Parameters (reals) are added to current garr position.
 *
 * @ai  item    Item pointer.
 * @ai  result  Growing array, where to store parameters.
 *
 * @rv  >=0                            Success - number of found parameters.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 */
static inline int geom_item_horizontal_times(void *item, struct garr *result)
{ return geom_item_direction_times(item, &geom_vector_right, result); }


static inline int geom_item_distance_times(void *item, struct geom_point *point, real dist, struct garr *result)
{ return GEOM_ITEM_CALL2(distance_times, point, dist, result); }
GEOM_ITEM_EDEF2(distance_times, struct geom_point UNUSED *point, real UNUSED dist, struct garr UNUSED *result)


/**
 * Finds intersections of two given curves.
 * Intersections (geom_intersection structures) are added to current garr position.
 *
 * @ai  item    First item pointer.
 * @ai  item2   Second item pointer.
 * @ai  flags   Wanted entries in resulting structure (see GEOM_INTERSECTION_x flags).
 * @ao  result  Growing array, where to store intersections.
 *
 * @rv  >=0                            Success - number of found intersections.
 * @rv  GEOM_ERR_NUMERIC               Numeric error.
 * @rv  GEOM_ERR_ITEM_INVALID          Invalid type of item for this function.
 */
int geom_items_intersections(void *item, void *item2, uns flags, struct garr *result);




/* Prehlednejsi interface navenek... */

#define GEOM_ITEM_INIT(item, nclass) GEOM_INSTANCE_INIT(item, nclass)
#define GEOM_ITEM_ZERO_INIT(item, nclass) GEOM_INSTANCE_ZERO_INIT(item, nclass)
#define GEOM_ITEM_CREATE(item, nclass) do{GEOM_INSTANCE_CREATE(item, nclass, geom_##nclass##_create);}while(0)
#define GEOM_ITEM_ZERO_CREATE(item, nclass) do{GEOM_INSTANCE_ZERO_CREATE(item, nclass, geom_##nclass##_create);}while(0)
#define GEOM_ITEM_DESTROY(item) GEOM_INSTANCE_DESTROY(item)


/* GEOM_SOURCE shortcuts */
#ifdef GEOM_SOURCE
#define ITEM_ASSERT(nclass,item) GEOM_ITEM_ASSERT(nclass,item)
#define ITEM_ASSERTX(nclass,item) GEOM_ITEM_ASSERTX(nclass,item)
#define ITEM_FLAGS(item) GEOM_ITEM_FLAGS(item)
#endif


#endif /* _VRR_GEOMLIB_ITEM_H_ */


