/*
 * GEOMLIB - Auxilliary numeric routines.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/base.h"


/* geom_reals_valid */

int geom_reals_valid(uns n, real reals[])
{
	if (!PTEST(reals)) return 0;
	for (uns i = 0; i < n; i++) if (!RTEST(reals[i])) return 0;
	return 1;
}


/* geom_reals_msg */

void geom_reals_msg(uns n, real reals[])
{
	if (reals) {
		if (!geom_reals_valid(n, reals)) msg("!");
		msg("(");
		if (n) {
			msg("%f", reals[0]);
			for (uns i = 1; i < n; i++)
				msg(",%f", reals[i]);
		}
		msg(")");
	}
	else
		msg("NULL");
}



/* geom_factorial */

int geom_factorial(int n)
{
	static int table[] = { 1, 1, 2, 6, 24, 120, 720, 5040,
		40320, 362880, 3628800, 39916800, 479001600
	};

	if (n < 0 || n >= (int)ARRAY_SIZE(table))
		return -1;

	else
		return table[n];
}

/* geom_binomial
  TODO:
  - rychlejc pres tabulku, vetsi rozsah
*/

int geom_binomial(int n, int k)
{
	int i, r;
	if (n > 17)
		return -1;
	if (n < 0 || k < 0 || k > n)
		return 1;
	if (k > n - k)
		k = n - k;
	r = 1;
	for (i = n - k + 1; i <= n; i++)
		r *= i;
	for (i = 2; i <= k; i++)
		r /= i;
	return r;
}

/* geom_doubles_sort
   TODO:
   - nedelat pres qsort, zbytecne pomale
*/

static int geom_doubles_sort_cmp(const void *a, const void *b)
{
	if (*(double *) a < *(double *) b)
		return -1;
	else if (*(double *) a > *(double *) b)
		return 1;
	else
		return 0;
}

int geom_doubles_sort(int n, double seq[])
{
	qsort(seq, n, sizeof(double), geom_doubles_sort_cmp);
	return 0;
}

static int geom_reals_sort_cmp(const void *a, const void *b)
{
	if (*(real *) a < *(real *) b)
		return -1;
	else if (*(real *) a > *(real *) b)
		return 1;
	else
		return 0;
}

int geom_reals_sort(int n, real seq[])
{
	qsort(seq, n, sizeof(real), geom_reals_sort_cmp);
	return 0;
}

/* geom_doubles_merge */

int geom_doubles_merge(int n1, double seq1[], int n2, double seq2[], double result[])
{
	int i, j, k;

	i = j = k = 0;
	if (n1 && n2)
		while (1) {
			if (seq1[i] <= seq2[j]) {
				result[k++] = seq1[i++];
				if (i >= n1)
					break;
			} else {
				result[k++] = seq2[j++];
				if (j >= n2)
					break;
			}
		}

	while (i < n1)
		result[k++] = seq1[i++];
	while (j < n2)
		result[k++] = seq2[j++];

	return k;
}

/* geom_permutation_generate_init */

int geom_permutation_generate_init(int n, int p[] UNUSED, int x[])
{
	int i;

	BASSERT(n > 0 && p && x);

	for (i = 0; i < n - 1; i++)
		x[i] = 0;
	x[n - 1] = -1;

	return 0;
}

/* geom_permutation_generate */

int geom_permutation_generate(int n, int p[], int x[])
{
	int i;

	BASSERT(n > 0 && p && x);

	if (n > 1) {
		for (i = n - 1; i > 0; i--) {
			if (x[i] != i)
				break;
			x[i] = 0;
		}

		if (i == 0)
			return 0;

		x[i]++;
		p[0] = 1;
		for (i = 0; i < n; i++) {
			p[i] = p[i - x[i]];
			p[i - x[i]] = i + 1;
		}

		return 1;
	};
	p[0] = 1;
	return (x[0]++) != 0;
}

/* geom_permutation_sign */

int geom_permutation_sign(int n, int p[])
{
	int i, k;

	BASSERT(n > 0 && p);

	k = n;

	for (i = 0; i < n; i++)
		if (p[i] > 0) {
			k--;
			while (p[i] > 0) {
				p[i] = -p[i];
				i = -p[i] - 1;
			}
		}

	for (i = 0; i < n; i++)
		p[i] = -p[i];

	return k & 1 ? -1 : 1;
}
