#ifndef __VRR_CMD_MGR_H__
#define __VRR_CMD_MGR_H__

/** @file
 *  GUI - The command manager.
 *
 *  Structures, types, enums, functions etc. for commads and menu items
 *  management.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "gui/main.h"

#define CS_VISIBLE	0x1
#define CS_ENABLED	0x2
#define CS_VIS_EN	( CS_VISIBLE | CS_ENABLED )
#define CS_CHECKED	0x4
#define CS_UNEXPANDED	0x8

#define CL_MAIN_TBAR	0x10
#define CL_MAIN_MENU	0x20
#define CL_VIEW_MENU	0x40
#define CL_UB_MENU	0x80
#define CL_VIEW_TBAR	0x100
#define CL_VIEW_TBAR2	0x200
#define CL_WUH_TBAR	0x400
#define CL_LWUH_TBAR	0x800
#define CL_VIEW		( CL_VIEW_MENU | CL_VIEW_TBAR | CL_VIEW_TBAR2 )
#define CL_LOCAL_MENUS	( CL_VIEW_MENU | CL_UB_MENU )
#define CL_ALL_MENUS	( CL_MAIN_MENU | CL_VIEW_MENU | CL_UB_MENU )
#define CL_ALL_TBARS	( CL_MAIN_TBAR | CL_VIEW_TBAR | CL_WUH_TBAR | CL_LWUH_TBAR | CL_VIEW_TBAR2 )
#define CL_EVERYWHERE	( CL_ALL_TBARS | CL_ALL_MENUS )

#define CC_CHECKABLE	0x1000
#define CC_STOCK	0x2000
#define CC_SAFE		0x8000

/** Object type bit masks. */
enum obj_vtm
{
	VTM_DOCUMENT		= 1 << OT_DOCUMENT,
	VTM_TLO			= 1 << OT_TLO,
	VTM_OBJ			= ( VTM_DOCUMENT | VTM_TLO ),

	VTM_GROUP		= 1 << (GOT_GROUP + OT_MAX),
	VTM_POINT		= 1 << (GOT_POINT + OT_MAX),
	VTM_SEGMENT		= 1 << (GOT_SEGMENT + OT_MAX),
	VTM_BEZIER		= 1 << (GOT_BEZIER + OT_MAX),
	VTM_ELARC		= 1 << (GOT_ELARC + OT_MAX),
	VTM_PARPOINT		= 1 << (GOT_PARAMETRIC_POINT + OT_MAX),
	VTM_ARROW		= 1 << (GOT_ARROW + OT_MAX),
	VTM_INTERSECTION	= 1 << (GOT_INTERSECTION + OT_MAX),
	VTM_TEXT		= 1 << (GOT_TEXT + OT_MAX),
	VTM_TEX_TEXT		= 1 << (GOT_TEX_TEXT + OT_MAX),
	VTM_DECPOINT		= 1 << (GOT_DECORATOR_POINT + OT_MAX),
	VTM_GO			= ( VTM_GROUP | VTM_POINT | VTM_SEGMENT | VTM_BEZIER | VTM_ELARC | VTM_PARPOINT |
					VTM_INTERSECTION | VTM_TEXT | VTM_TEX_TEXT | VTM_DECPOINT | VTM_ARROW),

	VTM_ANYTHING		= ( VTM_OBJ | VTM_GO ),
	VTM_VIEW		= 1 << ( OT_MAX + GOT_MAX + WT_VIEW ),
	VTM_UNDO_HISTORY	= 1 << ( OT_MAX + GOT_MAX + WT_UNDO_HISTORY ),
	VTM_UNIVERSE_BROWSER	= 1 << ( OT_MAX + GOT_MAX + WT_UNIVERSE_BROWSER ),

	VTM_CT_DOC		= 1 << ( OT_MAX + GOT_MAX + WT_UNIVERSE_BROWSER + 1 ),
	VTM_CT_TLO		= 1 << ( OT_MAX + GOT_MAX + WT_UNIVERSE_BROWSER + 2 ),
	VTM_CT_GRP		= 1 << ( OT_MAX + GOT_MAX + WT_UNIVERSE_BROWSER + 3 ),
};

/** Convert object type to bitmask */
#define OT_TO_VTM(x)	(uns)(1 << (x))
/** Convert GO type to bitmask */
#define GOT_TO_VTM(x)	(uns)(1 << (OT_MAX + (x)))
/** Convert window type to bitmask */
#define WT_TO_VTM(x)	(uns)(1 << (OT_MAX + GOT_MAX + (x)))

#define COMBINE_FLAGS(c,m) if (c) cmd->flags |= (m); else cmd->flags &= ~(m)
#define FILTER_FLAGS(c,m) if (!(c)) cmd->flags &= ~(m)

struct cmd;
struct cmd_category;
/** A structure for data that are attached to each toolbar button */
struct cmddata {
	struct window * window;
	struct go_group * group;
};
/** The type of a command item */
enum cmd_type { CT_FUNC, CT_FACTORY_OP, CT_CATEGORY, CT_SEPARATOR };
/** The type of the command item request */
enum req_type {
	RT_META,	//!< Meta selection or context
	RT_GROUP,	//!< Nonmeta (group) selection
	RT_ANY		//!< Both meta and nonmeta selection together
};
/** Command item structure */
struct cmd
{
	char * title;
	char * description;
	uns type;		//!< The item type. One of the enum cmd_type values
	uns flags;
	char * icon_path;	//!< The icon path or the GTK stock name
	GdkPixbuf * icon_pixbuf;	//!< The pixbuf created during icon load
	int accel;			//!< The command keyboard shortcut
	GdkModifierType modifier;	//!< The command keyboard modifiers
	union
	{
		struct
		{
			uns request_mask;	//!< The bitmask of requested objects. All possible objects the command can be applied on.
			int request_cnt;	//!< The requested object count. If 0, any number is suitable, if -1, any nonzero number is suitable.
			uns request_type;	//!< One of the enum req_type values.
			void (*modify_state_f)(struct cmd *, GtkWidget * widget);	//!< A function to modify the state (selected, ...) of the command in addition to the standard request.
			void (*action_f)(void *);	//!< The command action function.
		} func;
		struct
		{
			struct of_state * of_state;	//!< The factory state which to start an operation with.
		} op;
		struct
		{
			uns flags;
			struct cmd * first;	//!< The first command of the category.
			struct cmd * selected;	//!< The selected command of the category.
		} ctg;
	} spec;
	struct cmd * parent;	//!< The parent category
	struct cmd * next;	//!< The next command in the same category.
};

/** The global main category, the root category of all. */
extern struct cmd main_ctg;
extern struct cmd create_ctg;
extern struct cmd plugin_ctg;

/** Initialize the command manager.  */
void command_manager_init(void);
/** Finalize the command manager. */
void command_manager_finish(void);

/** Creates the main menu.
 * \param A bit mask for location, e.g. VCL_MAIN_MENU.
 * \returns A pointer to the main menubar widget */
GtkWidget * create_menu( uns location, uns popup, struct cmddata cd );

/** Sets the state of the given menu to the current state
 *  of the appropriate commands. Goes through the contents of the menu
 *  and checks if axisting items are the same as items that should exist.
 *  If so, updates their state and returns 1. If not, returns 0 (indicating
 *  that the menu should be recreated).
 * \param menu A pointer to the widget to be set.
 * \param location The bit mask for menu location.
 */
int sync_menu( GtkWidget * menu, uns location );

/** Creates the toolbar for the given location.
 *  \returns Pointer to the newly created toolbar.
 */
GtkWidget * create_toolbar( uns location, uns vertical, struct cmddata cd );

/** Sets the state of the given toolbar to the current state
 *  of the appropriate commands. Creates new subwidgets or removes
 *  some if needed.
 * \param menu A pointer to the widget to be set.
 */
int sync_toolbar( GtkWidget * toolbar, uns location );

void execute_command_item( struct cmd * cmd );

/** A pointer to the global accel group.
 */
extern GtkAccelGroup * accel_group;

void on_context_dump( void * subj UNUSED );

/* DYNAMIC COMMAND ADDING AND REMOVING */
/** Add the command to the start of the given category */
void add_new_command_into( struct cmd * cmd, struct cmd * category );
/** Add the command cmd to after the command after */
void add_new_command_after( struct cmd * cmd, struct cmd * after );
/** Remove the command cmd from the category from */
void remove_command( struct cmd * cmd, struct cmd * from );

/* PLUGINS */
/** A GUI plugin menu record */
struct plugin_menu_rec
{
	struct cmd category;	//!< The menu category of the plugin
	slist cmd_list;		//!< The command list of the category
};
/** A GUI plugin command menu record */
struct plugin_menu_cmd_rec
{
	snode n;		//!< An internal list node
	struct cmd cmd;		//!< The menu command for the plugin function
};
/** Register the plugin with the given description. Create a menu category
 * where the plugin commands will be placed.
 * @rv The plugin menu id
 */
int plugin_menu_register( char * desc );
/** Unregister the plugin by id, destroy all menu commands of the plugin.
 * @rv zero if succesful
 */
int plugin_menu_unregister( int plugin_id );
/** Register a plugin function and place the given command in the plugin's
 * menu category.
 * @rv zero if succesful
 */
int plugin_menu_command_register( int plugin_id, struct cmd * cmd );
/** Register a plugin function and create a menu command for it in the
 * plugin's menu category. The command will be destroyed when unregistering
 * the plugin.
 * @rv zero if succesful
 */
int plugin_menu_command_register_conv( int plugin_id, char * title, char * desc, uns request_mask,
					int request_cnt, uns request_type, void (*action_f)(void *) );
/** Print out the list of registered plugins and plugin functions. */
void plugin_menu_dump(void * v);
#endif
