#ifndef __VRR_DIALOGS_H__
#define __VRR_DIALOGS_H__

/** @file
 * GUI - Macros for open and save dialogs
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
/** The least recently used folder to open a file from */
extern char last_open_folder[PATH_MAX];
/** The least recently used folder to save a file to */
extern char last_save_folder[PATH_MAX];
/** Extract the folder path and remember it in the dest */
void save_folder( const char * filename, char * dest );
/** Show the question dialog and ask if to overwrite the file
 * @rv 1 if overwrite
 */
int prompt_overwrite_file( const char * filename );

/** Show the file open dialog, let the user select multiple files
 * and cycle through their list (accesible as filenames[i]).
 * Must be followed by a matching OPEN_DLG_END.
 */
#define OPEN_DLG_START(_title, _extra, _fname) \
	GtkFileSelection * dlg = GTK_FILE_SELECTION( gtk_file_selection_new(_title) );			\
													\
	if (_fname) gtk_file_selection_set_filename( dlg, _fname );					\
	else { if (last_open_folder[0]) gtk_file_selection_set_filename( dlg, last_open_folder ); }	\
	gtk_file_selection_set_select_multiple( dlg, TRUE );						\
													\
	if (_extra) gtk_box_pack_start( GTK_BOX(dlg->action_area), _extra, GTK_FILL, GTK_FILL, 0 );	\
													\
	if ( gtk_dialog_run( GTK_DIALOG(dlg) ) == GTK_RESPONSE_OK )					\
	{												\
		char ** filenames = gtk_file_selection_get_selections(dlg);				\
		save_folder( filenames[0], last_open_folder );						\
		for ( volatile int i = 0; filenames[i]; i++ )

/** The enclosing macro for OPEN_DLG_START.
 */
#define OPEN_DLG_END \
		g_strfreev(filenames);			\
	}						\
	gtk_widget_destroy( GTK_WIDGET(dlg) )

/** Show the file save dialog, let the user select one file
 * and prompt for overwriting if the file already exists.
 * The name of the file can be then found in the filename variable.
 * Must be followed with a matching SAVE_DLG_END.
 */
#define SAVE_DLG_START(_title, _extra, _fname) \
	GtkFileSelection * dlg = GTK_FILE_SELECTION( gtk_file_selection_new(_title) );			\
													\
	if (_fname) gtk_file_selection_set_filename( dlg, _fname );					\
	else { if (last_save_folder[0]) gtk_file_selection_set_filename( dlg, last_save_folder ); }	\
	gtk_file_selection_set_select_multiple( dlg, FALSE );						\
													\
	if (_extra) gtk_box_pack_start( GTK_BOX(dlg->action_area), _extra, GTK_FILL, GTK_FILL, 0 );	\
													\
	volatile int exists = 1;									\
	struct stat st;											\
	while ( exists && gtk_dialog_run( GTK_DIALOG(dlg) ) == GTK_RESPONSE_OK )			\
	{												\
		const char * filename = gtk_file_selection_get_filename(dlg);				\
		if (!stat(filename, &st) && !prompt_overwrite_file(filename)) continue;			\
		else exists = 0;									\
		save_folder( filename, last_save_folder );

/** The enclosing macro for SAVE_DLG_START */
#define SAVE_DLG_END \
	}					\
	gtk_widget_destroy( GTK_WIDGET(dlg) )	\

#endif

/** Suggest the filename according the the name of the object _o.
 * @ai ext the file extension
 * @ao _sugname the resulting name
 */
#define SUGGEST_FILENAME(_o, _ext, _sugname) 							\
	const char * on = o_lookup_name((struct o*)_o);						\
	char _sugname[strlen(on) + strlen(last_save_folder) + strlen(_ext) + 2];		\
	sprintf( _sugname, "%s%s.%s", last_save_folder, on, _ext )
