/** @file
 * GUI - Implementation of the Main window.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include <gtk/gtk.h>
#include "gui/main.h"
#include "gui/cmdmgr.h"

struct wnd_main wnd_main;

// callbacks
static void main_window_destroy( GtkWidget *widget UNUSED, gpointer data UNUSED)
{
	win_o_finish( (struct window*)&wnd_main );
	wnd_main.window = NULL;
}

static gboolean delete_event( GtkWidget * widget UNUSED, GdkEvent * event UNUSED, gpointer data UNUSED)
{
	main_quit();
	return TRUE;
}

void main_window_show(void){ gtk_widget_show( wnd_main.window ); }

void main_window_refresh_menu(void)
{
	if ( !wnd_main.menubar || !sync_menu( wnd_main.menubar, CL_MAIN_MENU ) ){
		struct cmddata cd = { NULL, NULL };
		DBG("Main menu needs rebuilding!\n");
		if ( wnd_main.menubar )
			gtk_container_remove( GTK_CONTAINER(wnd_main.menubox), wnd_main.menubar );
		gtk_widget_show( wnd_main.menubar = create_menu( CL_MAIN_MENU, 0, cd ) );// create menubar
		gtk_box_pack_start( GTK_BOX(wnd_main.menubox), wnd_main.menubar, FALSE, FALSE, 0 );
		DBG("Rebuild OK\n");
	}
}
void main_window_refresh_toolbar(void)
{
	if ( !wnd_main.toolbar || !sync_toolbar( wnd_main.toolbar, CL_MAIN_TBAR ) ){
		struct cmddata cd = { NULL, NULL };
		DBG("Toolbar needs rebuilding!\n");
		if ( wnd_main.toolbar )
			gtk_container_remove( GTK_CONTAINER(wnd_main.toolbox), wnd_main.toolbar );
		gtk_widget_show( wnd_main.toolbar = create_toolbar( CL_MAIN_TBAR, 0, cd ) );// create toolbar
		gtk_box_pack_start( GTK_BOX( wnd_main.toolbox), wnd_main.toolbar, FALSE, FALSE, 0 );
		DBG("Rebuild OK\n");
	}
}
void main_window_create(void)
{
	// new window
	main_window = wnd_main.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	win_o_init( (struct window*)&wnd_main, WT_MAIN_WINDOW );
	
	g_signal_connect (G_OBJECT (wnd_main.window), "delete_event", G_CALLBACK (delete_event), NULL);
	g_signal_connect (G_OBJECT (wnd_main.window), "destroy", G_CALLBACK (main_window_destroy), NULL);
	gtk_window_set_title(GTK_WINDOW(wnd_main.window), "Vrr!");

	GtkWidget * main_vbox = gtk_vbox_new(FALSE, 0);
	gtk_widget_show( main_vbox );
	gtk_widget_show( wnd_main.menubox = gtk_vbox_new(FALSE, 0) );
	gtk_widget_show( wnd_main.toolbox = gtk_vbox_new(FALSE, 0) );
	gtk_widget_show( wnd_main.statusbar );

	gtk_window_add_accel_group( GTK_WINDOW( wnd_main.window ), accel_group);
	gtk_box_pack_start(GTK_BOX(main_vbox), wnd_main.menubox, FALSE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(main_vbox), wnd_main.toolbox, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(main_vbox), wnd_main.statusbar, FALSE, TRUE, 0);
	gtk_container_add(GTK_CONTAINER (wnd_main.window), main_vbox);

	main_window_refresh_menu();
	main_window_refresh_toolbar();

	gtk_widget_show( wnd_main.window );
}
