#ifndef __VRR_SELECT_COLOR_H__
#define __VRR_SELECT_COLOR_H__

/** @file
 * GUI - Color selection window widget.
 * It is derived from the GtkBin widget and its widgetness means nothing
 * special (but, of course, it can be packed into other widgets or so).
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include <gtk/gtk.h>
#define COLOR_MULT_COEF 255.0 /**< The coefficient by which the outer color range is greater than the inner (of the window).*/
#define COLOR_MAX_VALUE 255 /**< The maximum value that can be passed in set and get functions */
#define COLOR_MIN_VALUE 0 /**< The minimum value that can be passed in set and get functions */
#define GDK_COLOR_MULT_COEF 65535.0

#define INNER_TO_OUTER(x) ((int)( COLOR_MULT_COEF * (x) + 0.5))
#define OUTER_TO_GDK(x) ((int)( (x) / COLOR_MULT_COEF * GDK_COLOR_MULT_COEF + 0.5))
#define GDK_TO_INNER(x) ( (x) / GDK_COLOR_MULT_COEF )
#define GDK_TO_OUTER(x) ( (x) / 256 )
#define OUTER_TO_INNER(x) ( (x) / COLOR_MULT_COEF )
#define EPSILON 0.001
#define ALMOST_ZERO(x) ( (x) <= EPSILON && (x) >= -EPSILON )

/** \brief Color names enumeration
 */
enum colors{
	COLOR_RED,
	COLOR_GREEN,
	COLOR_BLUE,
	COLOR_HUE,
	COLOR_SATURATION,
	COLOR_VALUE,
	COLOR_CYAN,
	COLOR_MAGENTA,
	COLOR_YELLOW,
	COLOR_BLACK,
	COLOR_ALPHA,
	COLOR_RESULT,
	COLOR_MAX
};


#define GTK_TYPE_VRR_SELECT_COLOR                  (wnd_select_color_get_type ())
#define GTK_VRR_SELECT_COLOR(obj)                  (GTK_CHECK_CAST ((obj), GTK_TYPE_VRR_SELECT_COLOR, WndSelectColor))
#define GTK_VRR_SELECT_COLOR_CLASS(klass)          (GTK_CHECK_CLASS_CAST ((klass), GTK_TYPE_VRR_SELECT_COLOR, WndSelectColorClass))
#define GTK_IS_VRR_SELECT_COLOR(obj)               (GTK_CHECK_TYPE ((obj), GTK_TYPE_VRR_SELECT_COLOR))
#define GTK_IS_VRR_SELECT_COLOR_CLASS(klass)       (GTK_CHECK_CLASS_TYPE ((klass), GTK_TYPE_VRR_SELECT_COLOR))
#define GTK_VRR_SELECT_COLOR_GET_CLASS(obj)        (GTK_CHECK_GET_CLASS ((obj), GTK_TYPE_VRR_SELECT_COLOR, WndSelectColorClass))

typedef struct WndSelectColor       WndSelectColor;
typedef struct WndSelectColorClass  WndSelectColorClass;

/** \brief The main window structure.
 * Its first member is the ancestor widget GtkBin. Other members are
 * my private members.
 */
struct WndSelectColor
{
	GtkBin ancestor; /**< The ancestor widget */
	GtkNotebook * notebook; /**< The tab widget */
	GtkAdjustment * adj[COLOR_MAX]; /**< The adjustments whose values express the current setting. Values are in [0,1].*/
	GtkWidget * drawing_area[COLOR_MAX]; /**< Drawing areas */
	GtkWidget * dropper; /**< The invisible widget for dropper (used to grab mouse and keyboard events) */
	GtkWidget * dropper_button; /**< The button for dropper */
	GtkEntry * hexet_entry; /**< The RGB hexet entry */
	guchar * buf[COLOR_MAX]; /**< Buffers that contain rgb color contents of drawing areas */
	int buf_width[COLOR_MAX]; /**< Buffer allocated widths */
	int buf_height[COLOR_MAX]; /**< Buffer allocated heights */
	int change_in_progress; /**< Whether widgets are being changed and change events should be ignored */
};

/** \brief The WndSelectColorClass used by GTK object system.
 */
struct WndSelectColorClass
{
	GtkBinClass parent_class;
};

/** \brief A function used by GTK */
GtkType wnd_select_color_get_type(void) G_GNUC_CONST;
/** \brief The function that creates the window and returns a pointer
 *         to its GtkWidget existence.
 */
GtkWidget* wnd_select_color_new( void );

/** \brief Gets the current color value. Values are in the range [0,255].
 * \param red The red component to be filled in or NULL
 * \param green The green component to be filled in or NULL
 * \param blue The blue component to be filled in or NULL
 */
void wnd_select_color_get_rgb( WndSelectColor * wnd, int * red, int * green, int * blue );
/** \brief Gets the alpha value in range [0,255].
 * \param alpha The alpha to be filled in.
 */
void wnd_select_color_get_alpha( WndSelectColor * wnd, int * alpha );
/** \brief Sets the current color value. Values are supposed
 *         to be in the range [0,255].
 * \param red The red component to be set
 * \param green The green component to be set
 * \param blue The blue component to be set
 */
void wnd_select_color_set_rgb( WndSelectColor * wnd, int red, int green, int blue );
/** \brief Sets the alpha value.
 * \param alpha The alpha to be set (in [0,255].
 */
void wnd_select_color_set_alpha( WndSelectColor * wnd, int alpha );

#endif
