#undef LOCAL_DEBUG

/** \file gui/univbrowser.c
 The Universe browser winwdow and the interface between the GTK widget GtkTreeView and our kernel structures via GtkTreeModel.
*/

/* TODO
   Pozor na hooky, ktere mi nedopatrenim prijdou od tlg (neni ve stromu)!!!
*/

#include "lib/lib.h"
#include <stdio.h>

#include "kernel/kernel.h"

#include <gtk/gtk.h>
#include "gui/main.h"
#include "gui/cmdmgr.h"
#include "scheme/glt_gui.h"

typedef struct VrrKernelStore VrrKernelStore;
typedef struct VrrKernelStoreClass	VrrKernelStoreClass;

#define GTK_TYPE_VRR_STORE		(vrr_kernel_store_get_type())
#define VRR_KERNEL_STORE(obj)		(G_TYPE_CHECK_INSTANCE_CAST ((obj), GTK_TYPE_VRR_STORE, VrrKernelStore))
#define VRR_KERNEL_STORE_CLASS(klass)	(G_TYPE_CHECK_CLASS_CAST((klass), GTK_TYPE_VRR_STORE, VrrKernelStoreClass))
#define VRR_IS_KERNEL_STORE(obj)		(G_TYPE_CHECK_INSTANCE_TYPE((obj), GTK_TYPE_VRR_STORE))
#define VRR_IS_KERNEL_STORE_CLASS(klass)	(G_TYPE_CHECK_CLASS_TYPE((klass), GTK_TYPE_VRR_STORE))
#define VRR_KERNEL_STORE_GET_CLASS(obj)	(G_TYPE_INSTANCE_GET_CLASS((obj), GTK_TYPE_VRR_STORE, VrrKernelStoreClass))

enum vrr_kernel_store_columns{
	VSC_TITLE,
	VSC_TYPE,
	VSC_ADDRESS,
	VSC_BG_COLOR,
	VSC_MODIFIED,
	VSC_IS_CONTEXT,
	VSC_FILENAME,
	VSC_COUNT
};
static GType column_types[] = { G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING,
	G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_STRING };

/** The one and the only instance of the universe browser window. */
struct wnd_univ_browser wnd_univ_browser;

/** The structure providing the mediator between GtkTreeView and our internal kernel structures.*/
struct VrrKernelStore
{
	GObject parent; /**< The parent GTK object. */
	gint stamp; /**< A random naumber used to identify the TreeIter data. */
};

/** A structure needed by the GTK object system. */
struct VrrKernelStoreClass
{
	GObjectClass parent_class;
};
/** A function needed by the GTK object system. */
static GType vrr_kernel_store_get_type(void);

/** Creates a new VrrKernelStore. */
static VrrKernelStore * vrr_kernel_store_new(void);
/** Emits a GTK signal to let other widgets know that the patricular object changed. */
static void vrr_kernel_store_object_changed( VrrKernelStore * store, struct o * o );
/** Emits a GTK signal to let other widgets know that the patricular object was deleted. */
static void vrr_kernel_store_object_deleted( VrrKernelStore * store, struct o * father, int son_number );
/** Emits a GTK signal to let other widgets know that the patricular object was added. */
static void vrr_kernel_store_object_new( VrrKernelStore * store, struct o * o );
/** Emits a GTK signal that the some gos inside group parent changed their positions. */
static void vrr_kernel_store_gos_reordered( VrrKernelStore * store, struct go_group * parent, struct go * go, struct go * old_prev );

static inline int go_has_child( struct go * go );
/** A hook callback, handles selection events and property changes - refreshes the object in the view. */
static void ub_obj_changed( struct o * o, struct obj_hook * hook UNUSED );
/** A hook callback, removes the object from the view. */
static void ub_obj_after_delete( struct o * o, struct o * father, struct o * prev, struct obj_hook * hook UNUSED );
/** A hook callback, adds the object to the view. */
static void ub_obj_added( struct o * o, struct obj_hook * hook UNUSED );
/** A hook callback, handles selection events and property changes - refreshes the object in the view. */
static void ub_go_changed( struct o * o, struct go_hook * hook UNUSED );
/** A hook callback, removes the object from the view. */
static void ub_go_after_delete( struct o * o, struct o * owner, struct o * prev, struct go_hook * hook UNUSED );
/** A hook callback, adds the object to the view. */
static void ub_go_added( struct o * o, struct go_hook * hook UNUSED );
/** A hook callback, removes the object and adds in a new place. */ 
static void ub_go_relink( struct o * o, struct go_group *old_group, struct go * prev, struct go_hook* h UNUSED );

static struct trans_hook uth;
/** A callback for undo history item changes */
static void ub_uh_item_delete( struct obj_tlo *tlo, struct undo_gui* ug, struct undo_gui* prev_item, struct trans_hook* h );
/** A callback for undo history item changes */
static void ub_uh_item_new_change( struct obj_tlo *tlo, struct undo_gui* ug, struct trans_hook* h );
/** A callback for undo history item changes */
static void ub_uh_item_undo_redo( struct obj_tlo* tlo, struct trans_hook* h );
/** A callback for changed of the saved state */
static void ub_uh_saved( struct obj * obj, struct trans_hook* h );

static gboolean on_button_press( GtkTreeView * tree_view, GdkEventButton *event, gpointer user_data );

static gboolean on_destroy_ub( GtkWindow * w UNUSED, gpointer data UNUSED )
{
	DBG("Destroying universe browser\n");

	obj_hook_remove_by_data(&wnd_univ_browser);
	trans_hook_remove(tlo_universum, &uth);

	struct obj* ooo;
	struct go_hook * h;
	struct trans_hook * t;
	WALK_OBJ_RECURSIVE( obj_universum, ooo )
		if ( ooo->type == OT_TLO )
		{
			DBG("UB: Removing a GO hook from %p, type %d\n", ooo, ooo->type );
			if ( h = go_hook_remove_by_data( obj2tlo(ooo), &wnd_univ_browser )) xfree(h); else BUG();
			if ( t = trans_hook_remove_by_data( obj2tlo(ooo), &wnd_univ_browser )) xfree(t); else BUG();
		}

	wnd_univ_browser.visible = 0;
	win_o_finish( (struct window*)&wnd_univ_browser );
	wnd_univ_browser.window = NULL;
	wnd_univ_browser.tree_store = NULL;
	refresh_gui();
	return TRUE;
}

/* Create the browser and fill it with experimental data. */
void create_universe_browser(void){

	GtkWidget *wnd;
	wnd_univ_browser.window = wnd = gtk_window_new( GTK_WINDOW_TOPLEVEL );
	win_o_init( (struct window*)&wnd_univ_browser, WT_UNIVERSE_BROWSER );

	gtk_window_set_title( GTK_WINDOW( wnd ), "Universe browser" );
	gtk_window_set_default_size( GTK_WINDOW( wnd ), 400, 300 );
	gtk_window_add_accel_group( GTK_WINDOW(wnd), accel_group);

	GtkWidget * vbox = gtk_vbox_new( FALSE, 0 );
	GtkWidget * sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start( GTK_BOX( vbox ), sw, TRUE, TRUE, 0 );

	wnd_univ_browser.tree_store = vrr_kernel_store_new();
	
	wnd_univ_browser.tree_view = gtk_tree_view_new_with_model( GTK_TREE_MODEL( wnd_univ_browser.tree_store) );
	g_object_set( G_OBJECT( wnd_univ_browser.tree_view ), "expander-column", VSC_TITLE, "enable-search", FALSE, NULL );

	/* Set alternating row colors */
	gtk_tree_view_set_rules_hint( GTK_TREE_VIEW( wnd_univ_browser.tree_view ), TRUE );

	/* NAME column */
	GtkTreeViewColumn * column = gtk_tree_view_column_new();
	g_object_set( G_OBJECT(column), "title", "Name", "max-width", 250, NULL );

	GtkCellRenderer * renderer = gtk_cell_renderer_pixbuf_new(); /* The context icon -- an indicator if the object is in context */
	g_object_set( G_OBJECT( renderer ), "pixbuf", icon_pencil_on, NULL );
	gtk_tree_view_column_pack_start( column, renderer, FALSE );
	gtk_tree_view_column_set_attributes( column, renderer, "visible", VSC_IS_CONTEXT, "cell-background", VSC_BG_COLOR, NULL );

	renderer = gtk_cell_renderer_pixbuf_new(); /* The modified icon */
	g_object_set( G_OBJECT( renderer ), "pixbuf", icon_modified, NULL );
	gtk_tree_view_column_pack_start( column, renderer, FALSE );
	gtk_tree_view_column_set_attributes( column, renderer, "visible", VSC_MODIFIED, "cell-background", VSC_BG_COLOR, NULL );

	renderer = gtk_cell_renderer_text_new(); /* The text renderer */
	g_object_set( G_OBJECT( renderer ), "weight", 1000, NULL );
	gtk_tree_view_column_pack_start( column, renderer, TRUE );
	gtk_tree_view_column_set_attributes( column, renderer, "text", VSC_TITLE, "cell-background", VSC_BG_COLOR, NULL );
	
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_univ_browser.tree_view), column );

	/* TYPE column */
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes( "Type", renderer, "text", VSC_TYPE, "cell-background", VSC_BG_COLOR, NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_univ_browser.tree_view), column );

#ifdef GUI_DEBUG
	/* ADDRESS COLUMN */
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
			"Address", renderer, "text", VSC_ADDRESS, "cell-background", VSC_BG_COLOR, NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_univ_browser.tree_view), column );
#endif
	/* FILENAME column */
	renderer = gtk_cell_renderer_text_new();
	g_object_set( G_OBJECT(renderer), "ellipsize", PANGO_ELLIPSIZE_START, NULL );
	column = gtk_tree_view_column_new_with_attributes(
			"Filename", renderer, "text", VSC_FILENAME, "cell-background", VSC_BG_COLOR, NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_univ_browser.tree_view), column );

	/* Selection handling */
	gtk_tree_selection_set_mode( gtk_tree_view_get_selection(GTK_TREE_VIEW( wnd_univ_browser.tree_view )), GTK_SELECTION_NONE );

	gtk_container_add( GTK_CONTAINER( sw ), wnd_univ_browser.tree_view );

	gtk_box_pack_start( GTK_BOX( vbox ), wnd_univ_browser.statusbar, FALSE, TRUE, 0 );

	g_signal_connect( G_OBJECT( wnd_univ_browser.tree_view ), "button-press-event", G_CALLBACK( on_button_press ), NULL );

	g_signal_connect( G_OBJECT( wnd_univ_browser.window ), "delete_event", G_CALLBACK( gtk_widget_destroy ), NULL );
	g_signal_connect( G_OBJECT( wnd_univ_browser.window ), "destroy", G_CALLBACK( on_destroy_ub ), NULL );

	gtk_container_add( GTK_CONTAINER( wnd ), vbox );
	gtk_widget_show_all( GTK_WIDGET(wnd) );

	wnd_univ_browser.visible = 1;

	DBG("UB: adding an OBJ hook\n" );
	obj_hook_add( obj_hook_fill( xmalloc_zero(sizeof(struct obj_hook)), &wnd_univ_browser,
		ub_obj_added, ub_obj_after_delete, ub_obj_changed, ub_obj_changed, NULL, ub_obj_changed, NULL) );

	trans_hook_add( tlo_universum, trans_hook_fill(&uth, &wnd_univ_browser, ub_uh_item_new_change, ub_uh_item_delete,
			ub_uh_item_new_change, ub_uh_item_undo_redo, ub_uh_item_undo_redo, ub_uh_saved ));

	struct obj* ooo;
	WALK_OBJ_RECURSIVE( obj_universum, ooo )
		if ( ooo->type == OT_TLO )
		{
			DBG("UB: adding a GO hook to %p\n", ooo );
			go_hook_add( obj2tlo(ooo), go_hook_fill( xmalloc_zero(sizeof(struct go_hook)),
				&wnd_univ_browser, ub_go_added, ub_go_after_delete, ub_go_relink, ub_go_changed, ub_go_changed,
				NULL, ub_go_changed, NULL, NULL));
			trans_hook_add( obj2tlo(ooo), trans_hook_fill( xmalloc_zero(sizeof(struct trans_hook)),
				&wnd_univ_browser, ub_uh_item_new_change, ub_uh_item_delete, ub_uh_item_new_change,
				ub_uh_item_undo_redo, ub_uh_item_undo_redo, ub_uh_saved ));
		}
}

void ub_refresh_menu(void)
{
	if ( !(wnd_univ_browser.menu) || !sync_menu( wnd_univ_browser.menu, CL_UB_MENU ) ){
		struct cmddata cd = { (struct window*)&wnd_univ_browser, NULL };
                DBG("UB menu needs rebuilding!\n");
                if (wnd_univ_browser.menu) gtk_object_sink( GTK_OBJECT(wnd_univ_browser.menu) );
                wnd_univ_browser.menu = create_menu( CL_UB_MENU, 1, cd );
                DBG("Rebuild OK\n");
        }
}

void ub_obj_changed( struct o * o, struct obj_hook * hook UNUSED )
{
	DBG("UB: object %p changed\n", o );
	ASSERT ( o_is_linked(o) );
	vrr_kernel_store_object_changed( wnd_univ_browser.tree_store, o );
}

void ub_go_changed( struct o * o, struct go_hook * hook UNUSED )
{
	DBG("UB: GO %p changed\n", o );
	ASSERT( o_is_linked(o) );
	vrr_kernel_store_object_changed( wnd_univ_browser.tree_store, o );
}

void ub_obj_after_delete( struct o * o, struct o * father, struct o * prev, struct obj_hook * hook UNUSED )
{
	struct obj * obj = o2obj(o), * ooo;
	struct go_hook * h;
	struct trans_hook * t;
	DBG("UB: object %p after delete\n", o );

	WALK_OBJ_RECURSIVE( obj, ooo )
		if ( ooo->type == OT_TLO )
		{
			DBG("UB: remove GO hook from %p\n", ooo );
			if ( h = go_hook_remove_by_data( obj2tlo(ooo), &wnd_univ_browser )) xfree(h); else BUG();
			if ( t = trans_hook_remove_by_data( obj2tlo(ooo), &wnd_univ_browser )) xfree(t); else BUG();
		}

	vrr_kernel_store_object_deleted( wnd_univ_browser.tree_store, father, obj_get_rank_son((struct obj*)prev) + 1 );
}

void ub_go_after_delete( struct o * o UNUSED, struct o * owner, struct o * prev, struct go_hook * hook UNUSED )
{
	struct go_group * gr = (struct go_group*)owner;
	struct obj_tlo * tlo = go_get_tlo(&gr->go);
	DBG("UB: go deleted from group %p, prev = %p, rank=%u\n", owner, prev, group_get_rank_go(gr, (struct go*)prev) + 1 );
	// if the owner group of the go is the toplevel group of the tlo, then pass the tlo as father;
	// otherwise, pass the owner group
	vrr_kernel_store_object_deleted( wnd_univ_browser.tree_store, ( tlo_get_tlg(tlo) == gr ) ? obj2o(&tlo->obj) : go2o(&gr->go),
		group_get_rank_go(gr, (struct go*)prev) + 1 );
}

void ub_obj_added( struct o * o, struct obj_hook * hook UNUSED )
{
	DBG("UB: obj %p added, type %d\n", o, o->type );
	ASSERT( o_is_linked(o) );

	struct obj* ooo, * obj = o2obj(o);
	WALK_OBJ_RECURSIVE( obj, ooo )
	{
		vrr_kernel_store_object_new( wnd_univ_browser.tree_store, obj2o(ooo) );
		if ( ooo->type == OT_TLO )
		{
			DBG("UB: adding a GO hook to %p\n", ooo );
			go_hook_add( obj2tlo(ooo), go_hook_fill( xmalloc_zero(sizeof(struct go_hook)),
				&wnd_univ_browser, ub_go_added, ub_go_after_delete, ub_go_relink, ub_go_changed, ub_go_changed,
				NULL, ub_go_changed, NULL, NULL));
			trans_hook_add( obj2tlo(ooo), trans_hook_fill( xmalloc_zero(sizeof(struct trans_hook)),
				&wnd_univ_browser, ub_uh_item_new_change, ub_uh_item_delete, ub_uh_item_new_change,
				ub_uh_item_undo_redo, ub_uh_item_undo_redo, ub_uh_saved ));
		}
	}
}

void ub_go_added( struct o * o, struct go_hook * hook UNUSED )
{
	struct go * go = o2go(o), * go2;
	DBG("UB: go %p added\n", o );
	ASSERT( o_is_linked(o) );
	WALK_GO_RECURSIVE( go, go2 ) vrr_kernel_store_object_new( wnd_univ_browser.tree_store, go2o(go2) );
}

void ub_go_relink( struct o * o, struct go_group *old_group, struct go * prev, struct go_hook* h UNUSED )
{
	struct obj_tlo * tlo = go_get_tlo(&old_group->go);
	DBG("UB: go %p relinked\n", o );
	// if the go was relinked into the same group, the rank of 'prev' might have increased
	if ( old_group == go_get_group(o2go(o)) )
	{
		DBG("UB: gos in group %p reordered\n", old_group );
		vrr_kernel_store_gos_reordered( wnd_univ_browser.tree_store, old_group, o2go(o), prev );
	}
	else
	{
		uns rank = group_get_rank_go(old_group, prev) + 1;
		DBG("UB: go relinked from group %p, prev = %p, rank=%u\n", old_group, prev, rank );
		// if the owner group of the go is the toplevel group of the tlo, then pass the tlo as father;
		// otherwise, pass the owner group
		vrr_kernel_store_object_deleted( wnd_univ_browser.tree_store,
			( tlo_get_tlg(tlo) == old_group ) ? obj2o(&tlo->obj) : go2o(&old_group->go), rank );
		vrr_kernel_store_object_new( wnd_univ_browser.tree_store, o );
	}
}

#define COMPARE_AND_UPDATE_CONTEXT(_name) \
	if (context._name != old_context->_name) { \
		o = (struct o*)context._name; \
		if (o && o_is_linked(o)) vrr_kernel_store_object_changed(wnd_univ_browser.tree_store, o ); \
		o = (struct o*)old_context->_name; \
		if (o && o_is_linked(o)) vrr_kernel_store_object_changed(wnd_univ_browser.tree_store, o ); \
	}

void ub_context_changed(struct context * old_context)
{
	if ( wnd_univ_browser.tree_store )
	{
		ASSERT( VRR_IS_KERNEL_STORE(wnd_univ_browser.tree_store) );
		struct o * o;
		COMPARE_AND_UPDATE_CONTEXT(o)
		COMPARE_AND_UPDATE_CONTEXT(tlo)
		COMPARE_AND_UPDATE_CONTEXT(doc)
		if (context.group != old_context->group)
		{
			// the context group might be a top level group which is not shown in the structure
			// (there is the parent tlo instead)
			struct go * go;
			if (context.group && go_get_group(go = &context.group->go) && go_is_linked(go))
				vrr_kernel_store_object_changed(wnd_univ_browser.tree_store, go2o(go) );
			if (old_context->group && go_get_group(go = &old_context->group->go) && go_is_linked(go))
				vrr_kernel_store_object_changed(wnd_univ_browser.tree_store, go2o(go) );
		}
	}
}

void modify_meta_selection( struct obj * obj, GdkEventButton * event )
{
	ASSERT( obj && event );
	if ( event->button != 1 ) return;
	char err[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Change selection", err )

		if ( !(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK)) && !(obj->flags & OF_SELECTED) )
			obj_unselect_all();

		if ( !(event->state & GDK_CONTROL_MASK) )
			obj_select(obj);

		if ( event->state & GDK_CONTROL_MASK && !(event->state & GDK_SHIFT_MASK) )
			obj_unselect(obj);
	TRANS_FAILED
		statusbar_err(NULL, "Cannot change selection: %s", err);
	TRANS_END
}

#define WALK_AND_DO_ENCLOSED(_xxx,_what) \
	if ( rect ) \
	{ \
		ASSERT(group); \
		TLO_RECT_ENCLOSE_QUERY_BEGIN( _xxx, tlo, rect, gg ) \
			if ((grp = go_get_owner_and_member(gg, group))) \
			{ \
				if(grp->type != GOT_GROUP) \
					_what(grp); \
				else \
					if(geom_rectangle_is_in(group_get_rtree_bbox(go2group(grp)), rect)) \
						_what(grp); \
			} \
		TLO_RECT_ENCLOSE_QUERY_END(_xxx); \
	} else { \
		ASSERT(go); \
		if (group && (go = go_get_owner_and_member(go, group))) \
			_what(go); \
	}

void modify_selection( struct go * volatile go, struct obj_tlo * tlo, GdkEventButton * event, struct geom_rectangle * rect, struct go_group * group )
{
	ASSERT( tlo && (rect || go) && event );
	if ( event->button != 1 ) return;
	char err[TRANS_ERR_SIZE];
	struct go * grp;

	TRANS_BEGIN_MAIN( tlo, "Change selection", err )

		if ( !(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK)) && (!go || !go_is_selected(go)) )
			if ( group ) group_unselect_all_go(group);

		if ( !(event->state & GDK_CONTROL_MASK) )
			WALK_AND_DO_ENCLOSED(xxx,go_select);

		if ( event->state & GDK_CONTROL_MASK && !(event->state & GDK_SHIFT_MASK) )
			WALK_AND_DO_ENCLOSED(yyy,go_unselect);

	TRANS_FAILED
		statusbar_err(NULL, "Cannot change selection: %s", err);
	TRANS_END
}

static gboolean on_button_press( GtkTreeView * tree_view, GdkEventButton *event, gpointer user_data UNUSED )
{
	GtkTreePath * tp;
	GtkTreeViewColumn * twc;
	GtkTreeIter ti;
	GtkTreeModel * tm = GTK_TREE_MODEL(wnd_univ_browser.tree_store);

	factory_op_break();

	/* Let the TreeView process the event first */
	/* Do not edit this unless you know what you are doing! */
	GTK_WIDGET_GET_CLASS(tree_view)->button_press_event( GTK_WIDGET(tree_view), event );

	gtk_tree_view_get_cursor( tree_view, &tp, &twc );

	if (likely( tp && gtk_tree_model_get_iter( tm, &ti, tp ))) {
		//the user clicked a row
		struct o * o = (struct o*)ti.user_data;
		if ( o->kind == T_GO )
		{
			struct go * go = o2go(o);
			modify_selection( go, go_get_tlo(go), event, NULL, go_get_group(go) );
			change_context( (struct window*)&wnd_univ_browser, go_get_group(go), o );
		}
		else
		{
			modify_meta_selection( o2obj(o), event );
			change_context( (struct window*)&wnd_univ_browser,
				( o->type == OT_TLO ) ? tlo_get_tlg(obj2tlo(o2obj(o))) : NULL, o );
		}
	} else {
		change_context( (struct window*)&wnd_univ_browser, NULL, NULL );
	}
	refresh_gui();

	if ( event->button == 3 ){
		gtk_menu_popup( GTK_MENU(wnd_univ_browser.menu), NULL, NULL, NULL, NULL, event->button,
				gtk_get_current_event_time() );
		return TRUE;
	}
	return TRUE;
}


/* *** THE IMPLEMENTATION OF A GTK WIDGET TO SHOW OUR KERNEL STRUCTURES ****/

#define VALID_ITER(iter, vrr_kernel_store) (iter && iter->user_data)

static GObjectClass *parent_class = NULL;
/* initializations and finalization */
/** Initialization in the GTK object system. */
static void vrr_kernel_store_init( VrrKernelStore *vrr_kernel_store );
/** Initialization in the GTK object system. */
static void vrr_kernel_store_class_init( VrrKernelStoreClass *vrr_kernel_store_class );
/** Initialization in the GTK object system. */
static void vrr_kernel_store_tree_model_init( GtkTreeModelIface *iface );
/** Finalization in the GTK object system. */
static void vrr_kernel_store_finalize( GObject *object );

/* implementation of GtkTreeModelIface */
static GtkTreeModelFlags vrr_kernel_store_get_flags( GtkTreeModel *tree_model );
/** Returns the number of columns. */
static gint vrr_kernel_store_get_n_columns( GtkTreeModel *tree_model);
static GType vrr_kernel_store_get_column_type( GtkTreeModel *tree_model, gint index);
static gboolean	vrr_kernel_store_get_iter( GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreePath *path);
static GtkTreePath *vrr_kernel_store_get_path( GtkTreeModel *tree_model, GtkTreeIter *iter);
static void vrr_kernel_store_get_value( GtkTreeModel *tree_model, GtkTreeIter *iter, gint column, GValue *value);
static gboolean	vrr_kernel_store_iter_next( GtkTreeModel *tree_model, GtkTreeIter *iter);
static gboolean	vrr_kernel_store_iter_children( GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent);
static gboolean vrr_kernel_store_iter_has_child( GtkTreeModel *tree_model, GtkTreeIter *iter);
static gint vrr_kernel_store_iter_n_children( GtkTreeModel *tree_model, GtkTreeIter *iter);
static gboolean vrr_kernel_store_iter_nth_child( GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent, gint n);
/** Fills the iter data with the pointer to the parent object.
 * If the parent object is a top level group, sets the pointer to point to parent tlo instead. */
static gboolean	vrr_kernel_store_iter_parent( GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *child);

GType vrr_kernel_store_get_type(void)
{
	static GType vrr_kernel_store_type = 0;

	if (!vrr_kernel_store_type){
		static const GTypeInfo vrr_kernel_store_info =
		{ sizeof (VrrKernelStoreClass), NULL, NULL, (GClassInitFunc) vrr_kernel_store_class_init,
		NULL, NULL, sizeof (VrrKernelStore), 0, (GInstanceInitFunc) vrr_kernel_store_init, NULL };

		static const GInterfaceInfo tree_model_info =
		{ (GInterfaceInitFunc) vrr_kernel_store_tree_model_init, NULL, NULL };

		vrr_kernel_store_type = g_type_register_static (G_TYPE_OBJECT, "VrrKernelStore", &vrr_kernel_store_info, 0);

		g_type_add_interface_static (vrr_kernel_store_type, GTK_TYPE_TREE_MODEL, &tree_model_info);
	}
	return vrr_kernel_store_type;
}

static void vrr_kernel_store_class_init (VrrKernelStoreClass *class)
{
	GObjectClass *object_class;
	parent_class = g_type_class_peek_parent (class);
	object_class = (GObjectClass *) class;
	object_class->finalize = vrr_kernel_store_finalize;
}

static void vrr_kernel_store_tree_model_init (GtkTreeModelIface *iface)
{
	iface->get_flags = vrr_kernel_store_get_flags;
	iface->get_n_columns = vrr_kernel_store_get_n_columns;
	iface->get_column_type = vrr_kernel_store_get_column_type;
	iface->get_iter = vrr_kernel_store_get_iter;
	iface->get_path = vrr_kernel_store_get_path;
	iface->get_value = vrr_kernel_store_get_value;
	iface->iter_next = vrr_kernel_store_iter_next;
	iface->iter_children = vrr_kernel_store_iter_children;
	iface->iter_has_child = vrr_kernel_store_iter_has_child;
	iface->iter_n_children = vrr_kernel_store_iter_n_children;
	iface->iter_nth_child = vrr_kernel_store_iter_nth_child;
	iface->iter_parent = vrr_kernel_store_iter_parent;
}

static void vrr_kernel_store_init (VrrKernelStore *vrr_kernel_store)
{
	do vrr_kernel_store->stamp = g_random_int ();
	while (vrr_kernel_store->stamp == 0);
}

VrrKernelStore * vrr_kernel_store_new(void)
{ return VRR_KERNEL_STORE (g_object_new (GTK_TYPE_VRR_STORE, NULL)); }

static void vrr_kernel_store_finalize (GObject *object)
{ (* parent_class->finalize) (object); }

static GtkTreeModelFlags vrr_kernel_store_get_flags (GtkTreeModel *tree_model)
{
	DBG("vrr_kernel_store_get_flags called\n");
	g_return_val_if_fail (VRR_IS_KERNEL_STORE (tree_model), 0);
	return GTK_TREE_MODEL_ITERS_PERSIST; /* it means that the structure is branching */
}

static gint vrr_kernel_store_get_n_columns (GtkTreeModel *tree_model)
{
	//DBG("vrr_kernel_store_get_n_columns called\n");
	g_return_val_if_fail (VRR_IS_KERNEL_STORE (tree_model), 0);
	return VSC_COUNT;
}

static GType vrr_kernel_store_get_column_type (GtkTreeModel *tree_model, gint index)
{
	DBG("vrr_kernel_store_get_column type with index %d called\n", index);
	g_return_val_if_fail (VRR_IS_KERNEL_STORE (tree_model), G_TYPE_INVALID);
	g_return_val_if_fail (index < VSC_COUNT && index >= 0, G_TYPE_INVALID);
	/* determine what type the column is */
	return column_types[index];
}

static gboolean vrr_kernel_store_get_iter(GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreePath *path)
{
	DBG("vrr_kernel_store_get_iter called ");
	VrrKernelStore *vrr_kernel_store = (VrrKernelStore *) tree_model;
	g_return_val_if_fail (VRR_IS_KERNEL_STORE (vrr_kernel_store), FALSE);

	struct obj * obj;
	gint *indices;
	int depth, i = 0;

#ifdef LOCAL_DEBUG
	char * p = gtk_tree_path_to_string(path);
	DBG("for path = %s\n", p );
	g_free(p);
#endif
	indices = gtk_tree_path_get_indices (path);
	depth = gtk_tree_path_get_depth (path);
	g_return_val_if_fail (depth > 0, FALSE);

	iter->stamp = vrr_kernel_store->stamp;

	for ( obj = obj_universum; obj->type != OT_TLO && i < depth; )
		if ( !(obj = obj_get_nth_first_son(obj, indices[i++]) )) return FALSE;

	if ( i == depth ) iter->user_data = (gpointer)obj;
	else
	{
		struct go_group * group = tlo_get_tlg(obj2tlo(obj));
		struct go * go;
		for ( ; ; )
		{
			go = group_get_first_go(group);
			for ( int j = indices[i++]; j && go; j-- )
				if ( !(go = group_get_next_go(group, go)) ) return FALSE;

			if ( i == depth )
			{ iter->user_data = (gpointer)go; break; }

			if ( go->type == GOT_GROUP ) group = (struct go_group*)go;
			else return FALSE;
		}
	}
	DBG("found an object %p for that path\n", iter->user_data);
	return TRUE;
}

static GtkTreePath * vrr_kernel_store_get_path(GtkTreeModel *tree_model, GtkTreeIter  *iter)
{
	g_return_val_if_fail (VRR_IS_KERNEL_STORE (tree_model), NULL);
	g_return_val_if_fail (iter, NULL);
	g_return_val_if_fail (iter->user_data, NULL);
	g_return_val_if_fail (iter->stamp == VRR_KERNEL_STORE (tree_model)->stamp, NULL);
	DBG("vrr_kernel_store_get_path called for o = %p\n", iter->user_data);

	GtkTreePath *path = gtk_tree_path_new();

	struct o * o = void2o(iter->user_data);
	struct obj * obj;
	if ( o->kind == T_GO )
	{
		struct go * go = o2go(o);
		struct go_group * group = go_get_group(go);
		for ( ; group; go = &group->go, group = go_get_group(go))
			gtk_tree_path_prepend_index( path, group_get_rank_go(group, go) );
		obj = &go_get_tlo(go)->obj;
	}
	else obj = o2obj(o);

	for ( ; obj && obj->father && obj != obj_universum; obj = obj->father )
		gtk_tree_path_prepend_index( path, obj_get_rank_son(obj) );
#ifdef LOCAL_DEBUG
	char * p = gtk_tree_path_to_string(path);
	DBG("resulting path = %s\n", p );
	g_free(p);
#endif
	return path;
}

/* Get the value at given position */
static void vrr_kernel_store_get_value(GtkTreeModel *tree_model, GtkTreeIter *iter, gint column, GValue *value)
{
	DBG("vrr_kernel_store_get_value called\n");
	g_return_if_fail (VRR_IS_KERNEL_STORE (tree_model));
	g_return_if_fail (iter && iter->user_data);
	g_return_if_fail (iter->stamp == VRR_KERNEL_STORE (tree_model)->stamp);
	g_return_if_fail (column < VSC_COUNT);

	struct o * o = void2o(iter->user_data);
	g_value_init( value, column_types[column] );

	switch ( column ){
		case VSC_TITLE:
		{
			string n = o_get_name(o);
			g_value_set_string( value, n->text ); break;
		}
		case VSC_TYPE:
			if ( o->kind == T_OBJ ) g_value_set_string( value, string_obj_type[o->type]->text );
			else g_value_set_string( value, string_go_type[o->type]->text );
			break;
		case VSC_ADDRESS:
		{
			char buf[20];
			sprintf( buf, "%p", (void *)o );
			g_value_set_string( value, buf );
			break;
		}
		case VSC_BG_COLOR:
			if ( o->kind == T_GO ){
				if ( go_is_selected(o2go(o)) ) g_value_set_string( value, "lightblue" );
				else g_value_reset( value );
			} else {
				if ( obj_is_selected(o2obj(o)) ) g_value_set_string( value, "lightpink" );
				else g_value_reset( value );
			}
			break;
		case VSC_MODIFIED:
			g_value_set_boolean( value,
				(o->kind == T_OBJ &&
					(o->type == OT_DOCUMENT && !doc_is_all_saved(obj2doc(o2obj(o))) ||
					 o->type == OT_TLO && !tlo_is_saved(obj2tlo(o2obj(o))))
				));
			break;
		case VSC_IS_CONTEXT:
			g_value_set_boolean( value, (void *)o == context.doc || (void *)o == context.tlo ||
				(void *)o == context.group || o == context.o );
			break;
		case VSC_FILENAME:;
			const char * fn;
			if ( o->kind == T_OBJ && o->type == OT_DOCUMENT && (fn = prop_lookup_text(o, string_filename)))
				g_value_set_string( value, fn );
			else
				g_value_set_string( value, "" );
			break;
		default: BUG();
	}
}

/* Move the iter to next node */
static gboolean vrr_kernel_store_iter_next (GtkTreeModel *tree_model, GtkTreeIter *iter)
{
	g_return_val_if_fail (iter, FALSE);
	g_return_val_if_fail (iter->user_data, FALSE);
	g_return_val_if_fail (iter->stamp == VRR_KERNEL_STORE (tree_model)->stamp, FALSE);
	DBG("vrr_kernel_store_iter_next called for o = %p\n", iter->user_data);

	struct o * o = void2o(iter->user_data);

	if ( o->kind == T_OBJ )
		iter->user_data = (gpointer)obj_get_next_son(o2obj(o));

	else
	{
		struct go * go = o2go(o);
		iter->user_data = (gpointer)group_get_next_go(go_get_group(go), go);
	}
	DBG("Has next: %s\n", iter->user_data ? "yes" : "no" );
	return iter->user_data != NULL;
}

/* Move the iter to child */
static gboolean vrr_kernel_store_iter_children (GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent)
{
	g_return_val_if_fail (!parent || parent->user_data, FALSE);
	g_return_val_if_fail (!parent || parent->stamp == VRR_KERNEL_STORE (tree_model)->stamp, FALSE);
	DBG("vrr_kernel_store_iter_children called for o %p\n", parent->user_data);

	iter->stamp = parent->stamp;
	struct o * o = void2o(parent->user_data);
	if ( o->kind == T_OBJ )
	{
		if ( o->type == OT_TLO ) iter->user_data = (gpointer)group_get_first_go(tlo_get_tlg(obj2tlo(o2obj(o))));
		else iter->user_data = (gpointer)obj_get_first_son( o2obj(o) );
	} else {
		if ( o->type == GOT_GROUP ) iter->user_data = (gpointer)group_get_first_go((struct go_group*)o);
		else iter->user_data = NULL;
	}

	DBG("any children: %s\n", iter->user_data ? "yes" : "no" );
	return iter->user_data != NULL;
}
static inline int go_has_child( struct go * go )
{ return (go->type == GOT_GROUP && group_get_first_go((struct go_group*)go)); }

/* If the iter has a child */
static gboolean vrr_kernel_store_iter_has_child (GtkTreeModel *tree_model, GtkTreeIter  *iter)
{
	DBG("vrr_kernel_store_iter_has_child called\n");
	g_return_val_if_fail (VRR_IS_KERNEL_STORE (tree_model), FALSE);
	g_return_val_if_fail (iter->stamp == VRR_KERNEL_STORE (tree_model)->stamp, FALSE);
	g_return_val_if_fail (iter->user_data, FALSE);

	struct o * o = void2o(iter->user_data);

	if ( o->kind == T_OBJ )
	{
		if ( o->type == OT_TLO ) return group_get_first_go(tlo_get_tlg(obj2tlo(o2obj(o)))) != NULL;
		else return obj_get_first_son(o2obj(o)) != NULL;
	}
	else return go_has_child(o2go(o));
}
/* Probably the number of childen of given node */
static gint vrr_kernel_store_iter_n_children (GtkTreeModel *tree_model, GtkTreeIter  *iter)
{
	DBG("vrr_kernel_store_iter_n_children called\n");
	g_return_val_if_fail (VRR_IS_KERNEL_STORE (tree_model), 0);
	g_return_val_if_fail (!iter || iter->user_data, FALSE);
	struct o  * o = void2o(iter->user_data);
	struct go_group * group = NULL;
	int cnt = 0;

	if ( o->kind == T_OBJ && o->type == OT_TLO ) group = tlo_get_tlg(obj2tlo(o2obj((o))));
	if ( o->kind == T_GO && o->type == GOT_GROUP ) group = (struct go_group*)group;

	if ( group )
		for ( struct go * g = group_get_first_go(group); g; g = group_get_next_go(group, g) ) cnt++;

	else if ( o->kind  == T_OBJ )
		for ( struct obj * ooo = obj_get_first_son( o2obj(o) ); ooo; ooo = obj_get_next_son(ooo) ) cnt++;

	DBG("the count of children: %d\n", cnt );
	return cnt;
}

/* Move to the nth child */
static gboolean vrr_kernel_store_iter_nth_child (GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent, gint n)
{
	DBG("vrr_kernel_store_iter_nth_child called\n");
	g_return_val_if_fail (VRR_IS_KERNEL_STORE (tree_model), FALSE);

	struct o * o;
	if ( !parent || !(parent->user_data) ) o = obj2o(obj_universum);
	else o = void2o(parent->user_data);

	struct go_group * group = NULL;
	iter->stamp = VRR_KERNEL_STORE(tree_model)->stamp;
	iter->user_data = NULL;

	if ( o->kind == T_OBJ && o->type == OT_TLO ) group = tlo_get_tlg(obj2tlo(o2obj(o)));
	if ( o->kind == T_GO && o->type == GOT_GROUP ) group = (struct go_group*)group;

	if ( group )
	{
		struct go * g;
		for ( g = group_get_first_go(group); g && n; group_get_next_go(group, g) ) n--;
		iter->user_data = (gpointer)g;
	}
	else if ( o->kind  == T_OBJ )
	{
		struct obj * ooo;
		for ( ooo = obj_get_first_son( o2obj(o) ); ooo && n; obj_get_next_son(ooo) ) n--;
		iter->user_data = (gpointer)ooo;
	}
	return iter->user_data != NULL && n == 0;
}

/* Move to parent */
static gboolean vrr_kernel_store_iter_parent (GtkTreeModel *tree_model, GtkTreeIter  *iter,
			    GtkTreeIter  *child)
{
	DBG("vrr_kernel_store_iter_parent called\n");
	g_return_val_if_fail (iter, FALSE);
	g_return_val_if_fail (child, FALSE);
	g_return_val_if_fail (child->user_data, FALSE);
	g_return_val_if_fail (child->stamp == (VRR_KERNEL_STORE(tree_model))->stamp, FALSE);

	struct o * o = void2o(child->user_data);
	iter->stamp = VRR_KERNEL_STORE(tree_model)->stamp;
	if ( o->kind == T_GO )
	{
		struct obj_tlo * tlo = go_get_tlo(o2go(o));
		if ( tlo_get_tlg(tlo) == go_get_group(o2go(o)) )
			iter->user_data = (gpointer)tlo; //we must not return a toplevel group, return the tlo instead
		else
			iter->user_data = o_get_ancestor(o);

	} else iter->user_data = o_get_ancestor(o);

	return iter->user_data != NULL;
}

void vrr_kernel_store_object_changed( VrrKernelStore * store, struct o * o )
{
	GtkTreeIter iter;
	iter.stamp = store->stamp;
	iter.user_data = (gpointer)o;
	DBG("object changed\n");
	g_signal_emit_by_name( (gpointer)store, "row-changed",  NULL, &iter, NULL );
}

void vrr_kernel_store_object_deleted( VrrKernelStore * store, struct o * father, int son_number )
{
	GtkTreeIter iter;
	GtkTreePath * path;
	iter.stamp = store->stamp;
	iter.user_data = (gpointer)father;
	DBG("object deleted from father %p, rank %d\n", father, son_number );
	path = vrr_kernel_store_get_path( GTK_TREE_MODEL(store), &iter ); //get the father's path
	gtk_tree_path_append_index( path, son_number );
	g_signal_emit_by_name( (gpointer)store, "row-deleted", path, NULL );

	if ( father->kind == T_OBJ && !obj_get_first_son(o2obj(father)) ||
		father->kind == T_GO && father->type == GOT_GROUP && !group_get_first_go(go2group(o2go(father))))
		// the last son will be deleted
		g_signal_emit_by_name( (gpointer)store, "row-has-child-toggled", NULL, &iter, NULL );

	gtk_tree_path_free( path );
}

void vrr_kernel_store_object_new( VrrKernelStore * store, struct o * o )
{
	GtkTreeIter iter;
	iter.stamp = store->stamp;
	iter.user_data = (gpointer)o;
	DBG("object added\n");
	g_signal_emit_by_name( (gpointer)store, "row-inserted",  NULL, &iter, NULL );

/*	if ( !obj_get_younger_brother(obj_get_oldest_son(obj->father))){
		iter.user_data = (gpointer)(obj->father); //his father has one child only, so it must have been added now
		g_signal_emit_by_name( (gpointer)store, "row-has-child-toggled", NULL, &iter, NULL );
	}*/
}
#define SIGNAL_TOGGLE_CHILD(_index) \
	{ \
		gtk_tree_path_append_index( path, (gint)_index ); \
		iter.user_data = (gpointer)go3; \
		gtk_tree_model_row_has_child_toggled( GTK_TREE_MODEL(store), path, &iter ); \
		gtk_tree_path_up( path ); \
	}

void vrr_kernel_store_gos_reordered( VrrKernelStore * store, struct go_group * parent, struct go * go, struct go * prev )
{
	uns cnt = group_get_count_go(parent);
	gint order[cnt];
	for (uns u = 0; u < cnt; u++) order[u] = u;
	uns old_rank = group_get_rank_go(parent, prev) + 1;
	uns new_rank = group_get_rank_go(parent, go);
	if (new_rank < old_rank)
		for (uns u = --old_rank; u > new_rank; u--) order[u] = u - 1;
	else
		for (uns u = old_rank; u < new_rank; u++) order[u] = u + 1;
	order[new_rank] = old_rank;

	GtkTreeIter iter;
	iter.stamp = store->stamp;
	iter.user_data = (gpointer)parent;
	GtkTreePath * path = gtk_tree_model_get_path( GTK_TREE_MODEL(store), &iter );
	DBG("Objects in %p reordered (within %u and %u)\n", parent, old_rank, new_rank );
	gtk_tree_model_rows_reordered( GTK_TREE_MODEL(store), path, &iter, order );

	struct go * go2, * go3 = go;
	if (new_rank < old_rank)
	{
		for (uns u = new_rank; u < old_rank; u++, go3 = go2)
			if ( go_has_child(go3) ^ go_has_child(go2 = group_get_next_go(parent, go3)) )
				SIGNAL_TOGGLE_CHILD(u);
	} else {
		for (uns u = new_rank; u > old_rank; u--, go3 = go2)
			if ( go_has_child(go3) ^ go_has_child(go2 = group_get_prev_go(parent, go3)) )
				SIGNAL_TOGGLE_CHILD(u);
	}
	if ( go_has_child(go3) ^ go_has_child(go) )
		SIGNAL_TOGGLE_CHILD(old_rank);
	
	gtk_tree_path_free(path);
}

/** Returns the same ancestor of the go as the kernel function o_get_ancestor except for the
 case that the ancestor is a top level group; in that case returns the parent tlo. */
struct o* o_get_gui_ancestor(struct o* o)
{
	ASSERT(o);
	if(o->kind == T_OBJ)
		return (struct o*)o2obj(o)->father;
	struct go *go = &go_get_group(o2go(o))->go;
	if (!go)
		return void2o(go_get_tlo(o2go(o)));
	return go_get_group(go) ? go2o(go) : void2o(go_get_tlo(o2go(o)));
}

static void ub_refresh_modified_icon( struct obj_tlo * tlo )
{
	GtkTreeIter iter;
	iter.stamp = wnd_univ_browser.tree_store->stamp;
	struct obj * obj;
	if( tlo == tlo_universum )
	{
		WALK_OBJ_RECURSIVE(obj_universum, obj)
		if ( obj->type == OT_DOCUMENT || obj->type == OT_TLO )
			{
				iter.user_data = (gpointer)obj;
				g_signal_emit_by_name( wnd_univ_browser.tree_store, "row-changed",  NULL, &iter, NULL );
			}
	} else {
		struct obj * father = (struct obj*)o_get_ancestor(void2o(tlo));
		iter.user_data = (gpointer)tlo;
		g_signal_emit_by_name( wnd_univ_browser.tree_store, "row-changed",  NULL, &iter, NULL );
		if (father && father != obj_universum)
		{
			iter.user_data = (gpointer)father;
			g_signal_emit_by_name( wnd_univ_browser.tree_store, "row-changed",  NULL, &iter, NULL );
		}
	}

}
static void ub_uh_item_delete( struct obj_tlo *tlo, struct undo_gui* ug UNUSED, struct undo_gui* prev_item UNUSED, struct trans_hook* h UNUSED )
{ ub_refresh_modified_icon(tlo);}
static void ub_uh_item_new_change( struct obj_tlo *tlo, struct undo_gui* ug UNUSED, struct trans_hook* h UNUSED )
{ ub_refresh_modified_icon(tlo);}
static void ub_uh_item_undo_redo( struct obj_tlo* tlo, struct trans_hook* h UNUSED )
{ ub_refresh_modified_icon(tlo);}
/** A callback for changed of the saved state */
static void ub_uh_saved( struct obj * obj, struct trans_hook* h UNUSED )
{
	GtkTreeIter iter;
        iter.stamp = wnd_univ_browser.tree_store->stamp;
	iter.user_data = (gpointer)obj;
	g_signal_emit_by_name( wnd_univ_browser.tree_store, "row-changed",  NULL, &iter, NULL );
}
