#ifndef _KERNEL_GEOM_H_
#define _KERNEL_GEOM_H_

/** @file
 *
 * Functions which directly call geomlib.
 * See geomlib/item.h for description.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

/*
 * PARAMETRIZATIONS
 */

SCHEME static inline SREAL go_tlength(SO_go go)
	{ return geom_item_tlength(go2geom(go)); }

SCHEME static inline SREAL go_alength(SO_go go)
	{ return geom_item_alength(go2geom(go)); }

SCHEME static inline SREAL go_blength(SO_go go)
	{ return geom_item_blength(go2geom(go)); }

	
/* TIME <-> ATIME conversions */

SCHEME static inline SREAL go_time_to_atime(SO_go go, SREAL time)
	{ return geom_item_time_to_atime(go2geom(go), time); }

static inline int go_times_to_atimes(struct go *go, uns count, real *times, real *atimes)
	{ return geom_item_times_to_atimes(go2geom(go), count, times, atimes); }

SCHEME static inline SREAL go_atime_to_time(SO_go go, SREAL time)
	{ return geom_item_atime_to_time(go2geom(go), time); }

static inline int go_atimes_to_times(struct go *go, uns count, real *times, real *atimes)
	{ return geom_item_atimes_to_times(go2geom(go), count, times, atimes); }

	
/* TIME <-> RATIME conversions */

SCHEME static inline SREAL go_time_to_ratime(SO_go go, SREAL time)
	{ return geom_item_time_to_ratime(go2geom(go), time); }

static inline int go_times_to_ratimes(struct go *go, uns count, real *times, real *atimes)
	{ return geom_item_times_to_ratimes(go2geom(go), count, times, atimes); }

SCHEME static inline SREAL go_ratime_to_time(SO_go go, SREAL time)
	{ return geom_item_ratime_to_time(go2geom(go), time); }

static inline int go_ratimes_to_times(struct go *go, uns count, real *times, real *atimes)
	{ return geom_item_ratimes_to_times(go2geom(go), count, times, atimes); }

	
/* TIME <-> BTIME conversions */
	
SCHEME static inline SINT go_time_dif_btime(SO_go go)
	{ return geom_item_time_dif_btime(go2geom(go)); }

static inline real go_time_to_btime(struct go *go, struct geom_expansion *expansion, real time)
	{ return geom_item_time_to_btime(go2geom(go), expansion, time); }

static inline int go_times_to_btimes(struct go *go, struct geom_expansion *expansion, uns count, real *times, real *btimes)
	{ return geom_item_times_to_btimes(go2geom(go), expansion, count, times, btimes); }
	
static inline real go_btime_to_time(struct go *go, struct geom_expansion *expansion, real btime)
	{ return geom_item_btime_to_time(go2geom(go), expansion, btime); }

static inline int go_btimes_to_times(struct go *go, struct geom_expansion *expansion, uns count, real *btimes, real *times)
	{ return geom_item_btimes_to_times(go2geom(go), expansion, count, btimes, times); }

	
	
/*
 * CONVERSION TO OTHER TYPES OF CURVES
 */

static inline int go_expansion_append(struct go *go, struct geom_fpath *expansion)
	{ return geom_item_expansion_append(go2geom(go), expansion); }

static inline struct geom_expansion* go_expand(struct go *go)
	{ return geom_item_expand(go2geom(go)); }

static inline void go_expansion_unref(struct geom_expansion *ref)
	 { geom_expansion_unref(ref); }

static inline int go_to_bezier(struct go *go, struct geom_bezier *bezier)
	{ return geom_item_to_bezier(go2geom(go), bezier); }


	
/*
 * GEOMETRIC PROPERTIES
 */
	
/* Point at curve */
	
static inline int go_point_start(struct go *go, struct geom_point *result)
	{ return geom_item_point_start(go2geom(go), result); }

static inline int go_point_end(struct go *go, struct geom_point *result)
	{ return geom_item_point_end(go2geom(go), result); }

static inline int go_point_at_time(struct go *go, real time, struct geom_point *result)
	{ return geom_item_point_at_time(go2geom(go), time, result); }


/* Derivation of curve */
	
static inline int go_derivation_start(struct go *go, struct geom_vector *result)
	{ return geom_item_derivation_start(go2geom(go), result); }

static inline int go_derivation_end(struct go *go, struct geom_vector *result)
	{ return geom_item_derivation_end(go2geom(go), result); }

static inline int go_derivation_at_time(struct go *go, real time, struct geom_vector *result)
	{ return geom_item_derivation_at_time(go2geom(go), time, result); }


/* Miscellaneous */
	
SCHEME static inline SINT go_is_closed(SO_go go)
	{ return geom_item_is_closed(go2geom(go)); }

static inline int go_bbox(struct go *go, struct geom_rectangle *bbox)
	{ return geom_item_bbox(go2geom(go), bbox); }

static inline int go_nearest_to_point(struct go *go, struct geom_point *point, uns flags, struct geom_nearest *nearest)
	{  return geom_item_nearest_to_point(go2geom(go), point, flags, nearest); }
	
static inline int go_horizontal_times(struct go *go, struct garr *result)
	{ return geom_item_horizontal_times(go2geom(go), result); }

static inline int go_intersections(struct go *go1, struct go *go2, uns flags, struct garr *result)
	{ return geom_items_intersections(go2geom(go1), go2geom(go2), flags, result); }


// TODO:
int go_approx(struct geom_approx *approx, struct go *go, uns flags);

#endif
