/*
 * Kernel - Decorators
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "kernel/internal.h"

#define GO_ID DECPOINT
#define GO_SUBID DECPOINT

PROP_STYLES
PROP_ALL_REHANG(coordinate, PROP_STYLES_LAST, x, 0, "Hanger of point must be a mouse click to change position.", 0)
PROP_ALL(radius, PSTRUCT(coordinate_x), PT_REAL, PTR_LENGTH, radius, r,
		value.r >= 0, (("Invalid value of radius."), ("Radius cannot be %f.", value.r)), CHANGED, LOW, PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(rotation, PSTRUCT(radius), PT_REAL, PTR_ANGLE_2PI, rotation, r,
		1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, PTF_RECYCLABLE, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(n, PSTRUCT(rotation), PT_UNS, PTU_UNSPECIFIED, n, u, 1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, PTF_RECYCLABLE, UINT_MAX)

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(coordinate_x), string_coordinate_x->text, &PSTRUCT(coordinate_y), string_coordinate_y->text,
		&PSTRUCT(radius), "radius", &PSTRUCT(rotation), "rotation", &PSTRUCT(n), "vertices")

INFO_ANCHOR_1("decorator-point", QT_POINT, LT_AFFINITY)
INFO_HANGER_0
INFO_GO_1("decorator-point",
		DECPOINT, "decorator-point", 0, -1, -1, -1, -1, 1, -1, -1, n)

void go_decorator_point_recompute(struct go_decorator_point *d, struct geom_rectangle *bbox)
{
	d->x = hanger_get_position(d->a->h);
	geom_item_after_change(go2geom(&d->go));
	if(!GEOM_ETEST(go_bbox(&d->go, bbox)))
		TRANS_FAIL(("Numeric error."), ("go_bbox fail."));
	// where 1 is 1 / (2 * sin (30 degrees)) and 60 degrees is inner angle in triangle
	if(!GEOM_ETEST(geom_rectangle_enlarge(bbox, d->stroke_width * 1 + 1, bbox))) 
		TRANS_FAIL(("Numeric error."), ("geom_rectangle_enlarge fail."));
}

void go_decorator_point_transform(struct go_decorator_point *d, struct geom_transform *t UNUSED, struct geom_rectangle *bbox)
{
	go_decorator_point_recompute(d, bbox);
}

static int go_dec_pt(struct go_decorator_point *go, struct geom_fpath *path, real d1, real a1, real d2, real a2)
{
	struct geom_bezier *ref;
	a1 = a1 * M_PI / 180;
	a2 = a2 * M_PI / 180;
	struct geom_point a = {go->x.x + d1 * go->radius * cos(a1 + go->rotation), go->x.y + d1 * go->radius * sin(a1 + go->rotation)};
	struct geom_point b = {go->x.x + d2 * go->radius * cos(a2 + go->rotation), go->x.y + d2 * go->radius * sin(a2 + go->rotation)};
	if (!geom_point_valid(&a) || !geom_point_valid(&b)) return 0;
	GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &a, &b);
	return 1;
}

static int go_decorator_point_callback(struct geom_callback_item *geom, struct geom_fpath *path)
{
	struct go_decorator_point *go = SKIP_BACK(struct go_decorator_point, geom, geom);
	struct geom_bezier *ref;

	/* Hidden feature :-) */
	if (go->n == 12345) {
		if (go_dec_pt(go, path, 1, -135, 1, 135) &&
			go_dec_pt(go, path, 1, 135, 1, -45) &&
			go_dec_pt(go, path, 1, -45, 1, 45) &&
			go_dec_pt(go, path, 1, 45, 1, 135) &&
			go_dec_pt(go, path, 1, 135, 1.5, 90) &&
			go_dec_pt(go, path, 1.5, 90, 1, 45) &&
			go_dec_pt(go, path, 1, 45, 1, -135) &&
			go_dec_pt(go, path, 1, -135, 1, -45))
		return 0;
		else return GEOM_ERR_NUMERIC;
	}

	if (go->n <= GO_DECORATOR_POINT_MAX_VERTICES)
		switch (go->n) {
		case 0:
			break;
		case 1:
			/* Single point */
			GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &go->x, &go->x);
			return 0;
		case 2: {
			/* Segment */
			struct geom_point p, q;
			p.x = go->radius * cos(go->rotation);
			p.y = go->radius * sin(go->rotation);
			q.x = go->x.x - p.x;
			q.y = go->x.y - p.y;
			p.x += go->x.x;
			p.y += go->x.y;
			if (!geom_point_valid(&p) || !geom_point_valid(&q)) return GEOM_ERR_NUMERIC;
			GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &p, &q);
			return 0; }
		default: {
			/* Polygon */
			uns i;
			struct geom_point p, q, s;
			real a = go->rotation;
			real a_inc = 2 * M_PI / go->n;
			s.x = go->radius * cos(a) + go->x.x;
			s.y = go->radius * sin(a) + go->x.y;
			if (!geom_point_valid(&s)) return GEOM_ERR_NUMERIC;
			for (i = 0, p = s; i < go->n; i++, p = q) {
				a += a_inc;
				q.x = go->radius * cos(a) + go->x.x;
				q.y = go->radius * sin(a) + go->x.y;
				if (!geom_point_valid(&q)) return GEOM_ERR_NUMERIC;
				GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &p, &q);
			}
			GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &p, &s);
			return 0; }
		}
	/* Circle */
	struct geom_elliptic_arc arc;
	int err = 0;
	GEOM_ITEM_INIT(&arc, elliptic_arc);
	geom_elliptic_arc_set_closed(&arc, &go->x, 0, go->radius, go->radius);
	err = geom_elliptic_arc_expansion_append(&arc, path);
	geom_elliptic_arc_destroy(&arc);
	return err;
}

struct go_decorator_point *go_decorator_point_new(struct hanger *point, uns n, real radius, real rotation)
{
	if(radius < 0)
		trans_fail("Invalid radius.");
	struct go_decorator_point *d = xmalloc_zero(sizeof(struct go_decorator_point));
	go_init(&d->go, GOT_DECORATOR_POINT, GOST_DECORATOR_POINT, point);
	geom_callback_item_create_set(&d->geom, &go_decorator_point_callback);
	d->radius = radius;
	d->rotation = rotation;
	d->n = n;

	PROP_STYLES_SET_DEFAULT_VALUE(d);
	d->unit_coordinate_x = d->unit_coordinate_y = d->unit_radius = d->unit_rotation = PROP_UNIT_MAX;
	return d;
}

#undef GO_ID
#undef GO_SUBID



#define GO_ID ARROW
#define GO_SUBID ARROW

PROP_STYLES
PROP_ALL(coordinate_y, PROP_STYLES_LAST, PT_REAL, PTR_COORDINATE, x.y, r,
		0, (("Read only property."), ("Bug, fix me.")), CHANGED, LOW, PTF_READ_ONLY, UINT_MAX)
PROP_ALL(coordinate_x, PSTRUCT(coordinate_y), PT_REAL, PTR_COORDINATE, x.x, r,
		0, (("Read only property."), ("Bug, fix me.")), CHANGED, LOW, PTF_READ_ONLY, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(reversed, PSTRUCT(coordinate_x), PT_UNS, PTU_BOOLEAN, reversed, u, value.u <= 1,
		(("Invalid value of reversed."), ("Reversed cannot be %u.", value.u)), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(arrow_alignment, PSTRUCT(reversed), PT_UNS, PTU_ARROW_ALIGN, arrow_alignment,
		u, value.u < PTAA_MAX, (("Invalid value of arrow alignment."), ("Arrow alignment cannot be %u.", value.u)),
		CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(arrow_front, PSTRUCT(arrow_alignment), PT_UNS, PTU_ARROW_FRONT, arrow_front,
		u, value.u < PTAF_MAX, (("Invalid value of arrow-front."), ("Arrow front cannot be %u.", value.u)), CHANGED, LOW,
		PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL_WITHOUT_UNIT(arrow_back, PSTRUCT(arrow_front), PT_UNS, PTU_ARROW_BACK, arrow_back,
		u, value.u < PTAB_MAX, (("Invalid value of arrow-back."), ("Arrow back cannot be %u.", value.u)), CHANGED, LOW,
		PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(arrow_length, PSTRUCT(arrow_back), PT_REAL, PTR_LENGTH, arrow_length, r,
		value.r >= 0, (("Invalid value of arrow-length."), ("Arrow length cannot be %f.", value.r)), CHANGED, LOW,
		PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(rotation, PSTRUCT(arrow_length), PT_REAL, PTR_ANGLE_2PI, rotation, r,
		1, (("Cannot fail."), ("Bug, fix me.")), CHANGED, LOW, PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(arrow_angle, PSTRUCT(rotation), PT_REAL, PTR_ANGLE_PI, arrow_angle, r,
		value.r >= 0 && value.r <= R_PI, (("Invalid value of arrow-angle."), ("Arrow angle cannot be %f.", value.r)),
		CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(parameter, PSTRUCT(arrow_angle), PT_REAL, PTR_REFERENCE, parameter, r,
		is_parameter_valid(value.r), (("Invalid value of parameter."), ("Parameter cannot be %f.", value.r)), CHANGED, LOW, 0, UINT_MAX)
PROP_ALL(front_curvature, PSTRUCT(parameter), PT_REAL, PTR_REFERENCE, front_curvature, r,
		is_parameter_valid(value.r), (("Invalid value of front-curvature."), ("Front curvature cannot be %f.", value.r)),
		CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
PROP_ALL(back_distance, PSTRUCT(front_curvature), PT_REAL, PTR_REFERENCE, back_distance, r,
		is_parameter_valid(value.r), (("Invalid value of back-distance."), ("Back distance cannot be %f.", value.r)),
		CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)

PROP_VIRTUAL(PROP_STYLES_INIT,
		&PSTRUCT(coordinate_x), string_coordinate_x->text, &PSTRUCT(coordinate_y), string_coordinate_y->text,
		&PSTRUCT(reversed), "reversed", &PSTRUCT(arrow_alignment), "arrow-alignment",
		&PSTRUCT(arrow_front), "arrow-front", &PSTRUCT(arrow_back), "arrow-back",
		&PSTRUCT(arrow_length), "arrow-length", &PSTRUCT(rotation), "rotation", 
		&PSTRUCT(arrow_angle), "arrow-angle", &PSTRUCT(parameter), "parameter",
		&PSTRUCT(front_curvature), "front-curvature", &PSTRUCT(back_distance), "back-distance")

INFO_ANCHOR_1("decorator-arrow", QT_CURVE, LT_AFFINITY)
INFO_HANGER_0
INFO_GO_1("decorator-arrow",
		ARROW, "decorator-arrow", 0, -1, -1, -1, -1, 1, -1, -1, back_distance)

void go_arrow_recompute(struct go_arrow *d, struct geom_rectangle *bbox)
{
	struct go *curve = hanger_get_go(d->a->h);
	if(curve->type != GOT_GROUP)
	{
		if(!GEOM_ETEST(go_point_at_time(curve, d->parameter, &d->x)))
			TRANS_FAIL(("Numeric error."), ("go_point_at_time fail."));
	}
	geom_item_after_change(go2geom(&d->go));
	if(!GEOM_ETEST(go_bbox(&d->go, bbox)))
		TRANS_FAIL(("Numeric error."), ("go_bbox fail."));
	// FIXME instead of 3 there can be exact value based on arrow angle
	if(!GEOM_ETEST(geom_rectangle_enlarge(bbox, d->stroke_width * 3 + 1, bbox))) 
		TRANS_FAIL(("Numeric error."), ("geom_rectangle_enlarge fail."));

}

void go_arrow_transform(struct go_arrow *d, struct geom_transform *t UNUSED, struct geom_rectangle *bbox)
{
	go_arrow_recompute(d, bbox);
}

static int go_arrow_callback(struct geom_callback_item *geom, struct geom_fpath *path)
{
	struct go_arrow *go = SKIP_BACK(struct go_arrow, geom, geom);
	void *curve = go2geom(hanger_get_go(go->a->h));

	enum prop_arrow_alignment arrow_alignment = go->arrow_alignment; // ... zpusob automatickeho zjistovani rotace krivky
	/*arrow_alignment = 1 - arrow_alignment;*/
	enum prop_arrow_front arrow_front = go->arrow_front; // ... tvar predni casti krivky
	enum prop_arrow_back arrow_back = go->arrow_back; // ... tvar zadni casti krivky
	int reverse = go->reversed; // boolean ... smer sipky vzhledem k parametrizaci krivky
	real rotation_correction = go->rotation; // [-neco, +neco] ... oprava rotace sipky, kdyby kuli hodne zatocene krivce nevypadala hezky
	
	struct garr garr;
	real rotation;
	uns n, i;
	struct geom_point front_point, back_point, left_point, right_point, left_mid_point, right_mid_point;
	real front_time, back_dist, back_time, left_rotation, right_rotation, left_mid_rotation, right_mid_rotation, t;
	struct geom_bezier *ref;
	struct geom_vector v;

	if(hanger_get_go(go->a->h)->type == GOT_GROUP)
	{
		GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &go->x, &go->x); // FIXME: Pavle, nemohla by se vratit prazdna cesta?
		return 0;
	}

	// initialize
	if(!GEOM_ETEST(geom_item_derivation_at_time(curve, go->parameter, &v)))
		return GEOM_ERR_NUMERIC;
	front_point = go->x;
	front_time = go->parameter;
	rotation = atan2(v.dy, v.dx);
	if (!GEOM_RTEST(rotation))
		rotation = 0;

	// compute back point
	if (arrow_alignment == PTAA_BACK || arrow_back == PTAB_POLY || arrow_back == PTAB_PARABOLIC) {
		back_dist = go->back_distance * go->arrow_length * cos(0.5 * go->arrow_angle);
		garr_init(&garr, 10, sizeof(real));
		n = geom_item_distance_times(curve, &front_point, back_dist, &garr);
		if (!GEOM_ETEST(n)) {
			garr_cleanup(&garr);
			return n;
		}
		back_time = NAN;
		if (reverse == 0) {
			for (i = n - 1; (int)i >= 0; i--)
				if ((t = *(real *)garr_get(&garr, i)) <= front_time) {
					back_time = t;
					break;
				}
		}
		else {
			for (i = 0; i < n; i++)
				if ((t = *(real *)garr_get(&garr, i)) >= front_time) {
					back_time = t;
					break;
				}
		}
		garr_cleanup(&garr);
		if (isfinite(back_time)) {
			if (!GEOM_ETEST(geom_errno = geom_item_point_at_time(curve, back_time, &back_point)))
				return geom_errno;
		}
		else {
			if (reverse == 0) {
				if (!GEOM_ETEST(geom_errno = geom_item_point_start(curve, &back_point)))
					return geom_errno;
			}
			else {
				if (!GEOM_ETEST(geom_errno = geom_item_point_end(curve, &back_point)))
					return geom_errno;
			}
			t = atan2(back_point.y - front_point.y, back_point.x - front_point.x);
			if (!GEOM_RTEST(t))
				t = reverse ? rotation : rotation + R_PI;
			back_point.x = front_point.x + back_dist * cos(t);
			back_point.y = front_point.y + back_dist * sin(t);
			if (!geom_point_valid(&back_point))
				return GEOM_ERR_NUMERIC;
		}
	}

	// compute arrow rotation
	switch (arrow_alignment) {
	case PTAA_BACK:
		rotation = atan2(back_point.y - front_point.y, back_point.x - front_point.x);
		if (GEOM_RTEST(rotation))
			break;
	case PTAA_FRONT:
		rotation = reverse ? rotation : rotation + R_PI;
		break;
	default:
		BUG();
	}
	if (!GEOM_RTEST(rotation += rotation_correction))
		return GEOM_ERR_NUMERIC;

	// compute arrow endpoints
	left_rotation = rotation - 0.5 * go->arrow_angle;
	right_rotation = rotation + 0.5 * go->arrow_angle;
	left_point.x = front_point.x + go->arrow_length * cos(left_rotation);
	left_point.y = front_point.y + go->arrow_length * sin(left_rotation);
	right_point.x = front_point.x + go->arrow_length * cos(right_rotation);
	right_point.y = front_point.y + go->arrow_length * sin(right_rotation);
	if (!(geom_point_valid(&left_point) && geom_point_valid(&right_point)))
		return GEOM_ERR_NUMERIC;

	// compute arrow midpoints
	if (arrow_front == PTAF_PARABOLIC) {
		t = 0.5 * go->arrow_angle * (1 - go->front_curvature);
		left_mid_rotation = rotation - t;
		right_mid_rotation = rotation + t;
		t = 0.5 * go->arrow_length * cos(0.5 * go->arrow_angle * go->front_curvature);
		left_mid_point.x = front_point.x + t * cos(left_mid_rotation);
		left_mid_point.y = front_point.y + t * sin(left_mid_rotation);
		right_mid_point.x = front_point.x + t * cos(right_mid_rotation);
		right_mid_point.y = front_point.y + t * sin(right_mid_rotation);
		if (!(geom_point_valid(&left_mid_point) && geom_point_valid(&right_mid_point)))
			return GEOM_ERR_NUMERIC;
	}
	
	// expand front
	switch (arrow_front) {
	case PTAF_STRAIGHT:
		GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &left_point, &front_point);
		GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &front_point, &right_point);
		break;
	case PTAF_PARABOLIC:
		GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_quadratic_nonrational, &left_point, &left_mid_point, &front_point);
		GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_quadratic_nonrational, &front_point, &right_mid_point, &right_point);
		break;
	default:
		BUG();
	}
	
	// expand back
	switch (arrow_back) {
	case PTAB_NONE:
		break;
	case PTAB_STRAIGHT:
		GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &right_point, &left_point);
		break;
	case PTAB_POLY:
		GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &right_point, &back_point);
		GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_linear_nonrational, &back_point, &left_point);
		break;
	case PTAB_PARABOLIC:
		GEOM_FPATH_APPEND_NEW_CREATE2(ref, path, bezier, geom_bezier_set_quadratic_nonrational, &right_point, &back_point, &left_point);
		break;
	default:
		BUG();
	}
	
	return 0;
}

struct go_arrow *go_arrow_new_on_hanger(struct hanger *point, real parameter, real arrow_length, real rotation)
{
	struct go_arrow *d = xmalloc_zero(sizeof(struct go_arrow));
	go_init(&d->go, GOT_ARROW, GOST_ARROW, point);
	geom_callback_item_create_set(&d->geom, &go_arrow_callback);
	d->parameter = parameter;
	d->arrow_length = arrow_length;
	d->rotation = rotation;

	d->arrow_angle = 1;
	d->front_curvature = 0.5;
	d->front_curvature = 0.5;
	d->back_distance = 0.7;
	d->arrow_alignment = PTAA_BACK;
	d->arrow_front = PTAF_PARABOLIC;
	d->arrow_back = PTAB_POLY;
	d->reversed = 0;

	PROP_STYLES_SET_DEFAULT_VALUE(d);
	d->unit_coordinate_x = d->unit_coordinate_y = d->unit_arrow_length = d->unit_rotation = d->unit_arrow_angle = d->unit_parameter = d->unit_back_distance = d->unit_front_curvature = PROP_UNIT_MAX;
	return d;
}

struct go_arrow *go_arrow_new(struct go *go, real parameter, real arrow_length, real rotation)
{
	struct hanger *h = go_get_hanger_curve(go);
	if(!h)
		trans_fail("Cannot create an arrow on a %s.", go_get_info(go)->name->text);

	return go_arrow_new_on_hanger(h, parameter, arrow_length, rotation);
}

#undef GO_ID
#undef GO_SUBID

