/*
 * Kernel - Main module
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "kernel/internal.h"
#include <getopt.h>
#include <guile/gh.h>
#include "font/font.h"
#include "lib/cache.h"
#include "scheme/glt_kernel.h"

string NAMES_STRINGS,
	string_stroke_cap_type[PSC_MAX],
	string_stroke_join_type[PSJ_MAX],
	string_go_type[GOT_MAX], string_obj_type[OT_MAX],
	string_prop_type[PT_MAX],
	string_quantity_type[QT_MAX],
	string_conic_type[GEOM_CONIC_ARC_MAX],
	string_alignment_x_type[PTAX_MAX], string_alignment_y_type[PTAY_MAX],
	string_arrow_front_type[PTAF_MAX], string_arrow_back_type[PTAB_MAX],
	string_arrow_align_type[PTAA_MAX];

static struct obj_tlo tlo_univ;
struct obj *obj_universum = &tlo_univ.obj;
struct obj_tlo *tlo_universum = &tlo_univ;

static struct obj_tlo tlo_zomb;
struct obj *obj_zombie = &tlo_zomb.obj;
struct obj_tlo *tlo_zombie = &tlo_zomb;

static struct obj obj_tem;
struct obj *obj_temp = &obj_tem;
struct obj_tlo *tlo_temp;

struct obj_tlo *tlo_clipboard;

struct trans *trans_present;
struct obj_tlo *trans_tlo;
struct trans* trans_present;

clist obj_lsel;
uns obj_sel_count[OT_MAX];

struct obj_info obj_info[OT_MAX];

#define INFO(_id) [GO_TYPE(_id)] = GLUEE(GO_NAME(_id), _info),
struct go_info *go_info[GOT_MAX] = { GO_FOR_ALL_ID(INFO) };

uns vrr_with_gui;


static void parse_command_option(int argc, char **argv)
{
	char buf[TRANS_ERR_SIZE];
	int opt;

	while((opt = getopt(argc, argv, "s:c:hv")) != -1)
		switch(opt)
		{
			case 's':
				if(!gh_eval_file(optarg))
					exit(1);
				break;

			case 'c':
				if(!gh_eval_str(optarg))
					exit(1);
				break;

			case '?':
				err("usage: %s [-v] [-s script] [-c expresion] [<vrr-file> ...]\n", argv[0]);
				exit(1);

			case 'v':
				msg(
					"VRR " PACKAGE_VERSION "\n"
					"The authors (a.k.a. The VRR Team):\n"
					"  Martin \"MJ\" Mares (advisor)\n  Pavel Charvat\n  Jirka Fink\n  Eva Ondrackova\n"
					"  Tomas Valla\n  Zuzka Vlckova\n  Ondra \"Santiago\" Zajicek\n\n"
					"Copyright (c) 2004-2005 The VRR Team  <vrr@ucw.cz>\n"
					"This program is distributed under GNU General Public License.\n"
				);
				exit(0);

			case 'h':
				err("usage: %s [-v] [-s script] [-e expresion] [<vrr-file> ...]\n", argv[0]);
				err("  -s SCRIPT      load Scheme source code from SCRIPT\n"
					 "  -c EXPR        evalute Scheme expression EXPR\n"
					 "  -h             print usage information\n"
					 "  -v             print version and copyright information\n");
				exit(1);

			default:;
				BUG();
		}

	for(; optind < argc; optind++)
	{
		TRANS_BEGIN_ANONYMOUS(buf)
			doc_load(argv[optind]);
		TRANS_FAILED
			err("Couldn\'t open file \'%s\'\n: %s", argv[optind], buf);
			DBG("%s\n", buf);
		TRANS_END
	}
}

void kernel_init(int argc, char **argv, uns gui)
{
	vrr_with_gui = gui;

	smalloc_init();
	cache_init(&global_cache);	
	geomlib_init();

	string_init();
	kernel_cache_init();
	property_init();
	object_init();
	hook_init();

#define INIT(_id) GLUEE(GO_NAME(_id), _init) ();
	GO_FOR_ALL_ID(INIT)

	clipboard_init();

	font_init(argv[0]);
	plugin_init();

	guilelink_first_init();
	guilelink_kernel_init();

	settings_init();
	parse_command_option(argc, argv);
}

void kernel_finish(void)
{
	settings_finish();
	
	glt_kernel_finish();
	plugin_finish();
	font_finish();

	clipboard_finish();

#define FINISH(_id) GLUEE(GO_NAME(_id), _finish) ();
	GO_FOR_ALL_ID(FINISH)

	hook_finish();
	object_finish();
	property_finish();
	kernel_cache_finish();
	string_finish();

	geomlib_finish();
	cache_destroy(&global_cache);
	
	smalloc_finish();
}

