#ifndef _KERNEL_MACRO_H_
#define _KERNEL_MACRO_H_

/*
 * Kernel - Useful macros
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#ifdef DEBUG_KERNEL
#define ASSERT_KERNEL ASSERT
#else
#define ASSERT_KERNEL(...) do { } while(0);
#endif

#ifdef DEBUG_KERNEL_SLOW
#define ASSERT_KERNEL_SLOW ASSERT
#else
#define ASSERT_KERNEL_SLOW(...) do { } while(0);
#endif

#ifdef DEBUG
#define DANGEROUS(_cond, ...) do { if(!(_cond)) msg("! "__VA_ARGS__); } while(0)
#else
#define DANGEROUS(...) do { } while(0)
#endif

#ifdef DEBUG_REF_O
#define DBG_O(...) msg(__VA_ARGS__)
#else
#define DBG_O(...) do { } while(0)
#endif

#ifdef DEBUG_TSORT
#define DBG_TSORT(...) msg(__VA_ARGS__)
#else
#define DBG_TSORT(...) do { } while(0)
#endif

#ifdef DEBUG_TRANS
#define DBG_TRANS(...) msg(__VA_ARGS__)
#else
#define DBG_TRANS(...) do { } while(0)
#endif

#define TRANS_FAIL1(...) TRANS_FAIL((__VA_ARGS__), (__VA_ARGS__))
#ifdef DEBUG_KERNEL
#define TRANS_FAIL(_release, _debug) do { msg("Trans fail in %s: ", __func__); msg _debug; msg("\n"); trans_fail _release ; } while(0)
#else
#define TRANS_FAIL(_release, _debug) trans_fail _release;
#endif


#define GO_NAME(_id) GLUEE(GO_NAME_, _id)
#define GO_TYPE(_id) GLUEE(GO_TYPE_, _id)
#define GO_STRUCT(_id) struct GLUEE(GO_STRUCT_, _id)
#define GO_GEOM(_id) GLUEE(GO_GEOM_, _id)
#define GO_SUBNAME(_id, _sid) GLUEE4(GO_SUBNAME_, _id, _, _sid)
#define GO_SUBTYPE(_id, _sid) GLUEE4(GO_SUBTYPE_, _id, _, _sid)
#define GO_SUBSTRUCT(_id, _sid) struct GLUEE4(GO_SUBSTRUCT_, _id, _, _sid)

#define GO_NAME_ID GO_NAME(GO_ID)
#define GO_TYPE_ID GO_TYPE(GO_ID)
#define GO_STRUCT_ID GO_STRUCT(GO_ID)
#define GO_SUBNAME_ID GO_SUBNAME(GO_ID, GO_SUBID)
#define GO_SUBTYPE_ID GO_SUBTYPE(GO_ID, GO_SUBID)
#define GO_SUBSTRUCT_ID GO_SUBSTRUCT(GO_ID, GO_SUBID)



#define ASSERT_META_TRANS ASSERT(trans_present && trans_tlo == tlo_universum)
#define ASSERT_GO_TRANS(_go) ASSERT(trans_present && void2tlo(trans_tlo) == ((_go) ? go_get_tlo(_go) : trans_tlo))
#define ASSERT_O_TRANS(_o) ASSERT(trans_present && void2o(_o)->kind == T_OBJ ? trans_tlo == tlo_universum : void2tlo(trans_tlo) == go_get_tlo(o2go(_o)))


#define ASSERT_GO_TYPE(_o, _id) ASSERT((_o) && ((struct o*)(_o))->kind == T_GO && ((struct o*)(_o))->type == GO_TYPE(_id))
#define ASSERT_GO_SUBTYPE(_o, _id, _sid) ASSERT((_o) && ((struct o*)(_o))->kind == T_GO && \
		((struct o*)(_o))->type == GO_TYPE(_id) && ((struct o*)(_o))->subtype == GO_SUBTYPE(_id, _sid))
#define ASSERT_HANGER(_h) ASSERT((_h) && ((struct hanger*)(_h))->info && hanger_get_go((struct hanger*)(_h))->kind == T_GO)
#define ASSERT_ANCHOR(_a) ASSERT((_a) && ((struct anchor*)(_a))->info && anchor_get_go((struct anchor*)(_a))->kind == T_GO)



#define GO_TYPE_SWITCH(...) \
	switch(void2go(go)->type) \
	{	__VA_ARGS__; \
		default: die("Function %s hasn't label for the go type %d - %s.\n", __func__, go->type, string_go_type[go->type]->text); \
	}

#define GO_SUBTYPE_SWITCH(_id, ...) \
	case GO_TYPE(_id): { \
		UNUSED GO_STRUCT(_id) *spec = (GO_STRUCT(_id)*)go; \
		switch(go->subtype) { \
			__VA_ARGS__ \
			default: die("Function %s hasn't label for the go type[%d] %s, subtype[%d] %s.\n", \
			__func__, go->type, string_go_type[go->type]->text, go->subtype, go_get_info(go)->name->text); } break; }



/*
 * Following macros must be defined before using this generator.
 * GO_NAME Must be same as name of structure. (For example go_elarc)
 * GO_SUBNAME Name of subtype. (go_elarc_xyr)
 * GO_TYPE Type of go. (GOT_ELARC)
 * GO_SUBTYPE Subtype of go. (GOST_ELARC_XYR)
 *
 * Arguments of macros.
 * _name Name of property. If this property has unit then structure GO_NAME must contain variable unit##_name of type uns. (dif)
 * _next Next property in link list. Last property is o_prop_name. (PSTRUCT_SID(ecc))
 * _prop_type Type of property. (PT_REAL)
 * _prop_subtype Subtype of property. (PTR_ANGLE_4PI)
 * _var Property is fixed to this variable of structure GO_NAME. (geom.dif)
 * _union Item of union prop_value. (r)
 * _anchor Index of anchor.
 * _cond Condition which must be satisfied to change value in function PSET. New value is saved in variable prop_value value. (is_dif_valid(value.r))
 * _msg If condition isn't satisfied trans_fail is called with this error message.
 * _recomp Which function is called in PSET. Possibilities are TLO (for recomputing by tsort) and ALTERED (for calling go_altered).
 * _low Possibilities are LOW and LOW_DIE. LOW_DIE is used iff property isn't undo.
 * _flags Flags of property. Flag PTF_VIRTUAL is automatically added and PTF_READ_ONLY is added if PROP_ALL_READ is used.
 */

static inline uns prop_unit_get_bug(struct o UNUSED *o) { BUG(); }
static inline void prop_unit_set_bug(struct o UNUSED *o, uns unit UNUSED) { BUG(); }

void prop_virtual_init(uns count, ...);
void prop_virtual_finish(uns count, ...);

#define PSTRUCT_SID(_name, _sid) GLUEE3(GO_SUBNAME(GO_ID, _sid), _prop_, _name)
#define PSTRUCT(_name) PSTRUCT_SID(_name, GO_SUBID)
#define PGET(_name) GLUEE4(GO_SUBNAME_ID, _prop_, _name, _get)
#define PSET(_name) GLUEE4(GO_SUBNAME_ID, _prop_, _name, _set)
#define PLOW(_name) GLUEE4(GO_SUBNAME_ID, _prop_, _name, _set_low)
#define PGETU(_name) GLUEE4(GO_SUBNAME_ID, _prop_, _name, _get_unit)
#define PSETU(_name) GLUEE4(GO_SUBNAME_ID, _prop_, _name, _set_unit)
#define PINIT(_sid) GLUEE(GO_SUBNAME(GO_ID, _sid), _virtual_prop_init)
#define PFINISH(_sid) GLUEE(GO_SUBNAME(GO_ID, _sid), _virtual_prop_finish)

#define RECOMPUTE_CHANGED tsort_dirty_recompute(&g->go)
#define RECOMPUTE_ALTERED go_altered(&g->go)

#define ASSERT_GO_TYPE_ID(_o) ASSERT_GO_TYPE(_o, GO_ID)
#define ASSERT_GO_SUBTYPE_ID(_o) ASSERT_GO_SUBTYPE(_o, GO_ID, GO_SUBID)

#define PROP_INIT(_name, _key) PSTRUCT(_name).p.key = string_lookup_ref(_key)
#define PROP_FINISH(_name) string_unref(PSTRUCT(_name).p.key)

#define PROP_VIRTUAL(...) \
static inline void PINIT(GO_SUBID) (void) { prop_virtual_init(0, __VA_ARGS__, NULL); } \
static inline void PFINISH(GO_SUBID) (void) { prop_virtual_finish(0, __VA_ARGS__, NULL); }

#define PROP_GET(_name, _var, _union)	\
static prop_value PGET(_name) (struct o *o)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	return (prop_value){._union = ((GO_STRUCT_ID *)o)->_var};	\
}

#define PROP_SET(_name, _var, _union, _cond, _msg, _recomp)	\
static void PSET(_name) (struct o *o, prop_value value)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	ASSERT(#_union[0] != 'r' || isfinite(value.r));	\
	GO_STRUCT_ID *g = (GO_STRUCT_ID *)o;	\
	if(!(_cond))	\
		TRANS_FAIL _msg;	\
	trans_new_undo_prop_virtual(o, &PSTRUCT(_name), (prop_value){._union = g->_var}); \
	g->_var = value._union; \
	RECOMPUTE_##_recomp;	\
}

#define PROP_SET_READ(_name)	\
static void PSET(_name) (struct o *o UNUSED, prop_value value UNUSED)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	trans_fail("Property %s is read only.", #_name); \
}

#define PROP_SET_REHANG_X(_name, _var, _anchor, _msg)	\
static void PSET(_name##_x) (struct o *o, prop_value value)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	ASSERT(isfinite(value.r));	\
	GO_STRUCT_ID *g = (GO_STRUCT_ID *)o;	\
	struct anchor *a = go_get_anchor(o2go(o), _anchor); \
	if(!hanger_is_mc(a->h)) \
		trans_fail(_msg);	\
	struct geom_point pos = {value.r, g->_var.y}; \
	anchor_rehang(a, hanger_new_mc(&pos)); \
}

#define PROP_SET_REHANG_Y(_name, _var, _anchor, _msg)	\
static void PSET(_name##_y) (struct o *o, prop_value value)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	ASSERT(isfinite(value.r));	\
	GO_STRUCT_ID *g = (GO_STRUCT_ID *)o;	\
	struct anchor *a = go_get_anchor(o2go(o), _anchor); \
	if(!hanger_is_mc(a->h)) \
		trans_fail(_msg);	\
	struct geom_point pos = {g->_var.x, value.r}; \
	anchor_rehang(a, hanger_new_mc(&pos)); \
}

#define PROP_SET_BOOLEAN(_name, _var, _mask, _recomp)	\
static void PSET(_name) (struct o *o, prop_value value)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	GO_STRUCT_ID *g = (GO_STRUCT_ID *)o;	\
	if(((g->_var & _mask) == 0) ^ (value.u != 0)) \
		return; \
	trans_new_undo_prop_virtual(o, &PSTRUCT(_name), (prop_value){.u = g->_var}); \
	g->_var ^= _mask; \
	RECOMPUTE_##_recomp;	\
}

#define PROP_SET_LOW(_name, _var, _union)	\
static void PLOW(_name) (struct o *o, prop_value value)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	((GO_STRUCT_ID *)o)->_var = value._union;	\
}

#define PROP_SET_LOW_BOOLEAN(_name, _var, _mask)	\
static void PLOW(_name) (struct o *o, prop_value value)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	GO_STRUCT_ID *g = (GO_STRUCT_ID *)o;	\
	ASSERT(((g->_var & _mask) == 0) ^ (value.u == 0)); \
	g->_var ^= _mask;	\
}

#define PROP_SET_LOW_DIE(_name, ...)	\
static void PLOW(_name) (struct o *o UNUSED, prop_value value UNUSED)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	die("Value of the property %s cannot set by low_set function because this is not a basic property of this subtype.", #_name); \
}

#define PROP_UNIT_GET(_name, _var) \
static uns PGETU(_name) (struct o *o) \
{ \
	ASSERT_GO_SUBTYPE_ID(o);	\
	return ((GO_STRUCT_ID *)o)-> _var; \
}

#define PROP_UNIT_GET_DIE(_name)	\
static uns PGETU(_name) (struct o *o UNUSED)	\
{	\
	ASSERT_GO_SUBTYPE_ID(o);	\
	die("Property %s hasn't unit.", #_name); \
}

#define PROP_UNIT_SET(_name, _var) \
static void PSETU(_name) (struct o *o, uns unit) \
{ \
	ASSERT_GO_SUBTYPE_ID(o);	\
	((GO_STRUCT_ID *)o)-> _var = unit; \
}

#define PROP_UNIT_SET_DIE(_name) \
static void PSETU(_name) (struct o *o UNUSED, uns unit UNUSED) \
{ \
	ASSERT_GO_SUBTYPE_ID(o);	\
	die("Property %s hasn't unit.", #_name); \
}

#define PROP_STRUCT(_name, _next, _prop_type, _prop_subtype, _flags, _undo_mask)	\
static struct prop_virtual PSTRUCT(_name) = {	\
	{ { &_next.p.n }, NULL, {0}, _prop_type, _prop_subtype, _flags | PTF_VIRTUAL, 0 },	\
	_undo_mask, PGET(_name), PSET(_name), PLOW(_name),	PGETU(_name), PSETU(_name)	\
};

#define PROP_ALL_WITHOUT_SET(_name, _next, _prop_type, _prop_subtype, _var, _union, _low, _flags, _undo_mask)	\
static struct prop_virtual PSTRUCT(_name);	\
PROP_GET(_name, _var, _union)	\
static void PSET(_name) (struct o *o, prop_value value); \
PROP_SET_##_low(_name, _var, _union) \
PROP_UNIT_GET(_name, unit_##_name)	\
PROP_UNIT_SET(_name, unit_##_name)	\
PROP_STRUCT(_name, _next, _prop_type, _prop_subtype, _flags, _undo_mask)

#define PROP_ALL(_name, _next, _prop_type, _prop_subtype, _var, _union, _cond, _msg, _recomp, _low, _flags, _undo_mask)	\
PROP_ALL_WITHOUT_SET(_name, _next, _prop_type, _prop_subtype, _var, _union, _low, _flags, _undo_mask)	\
PROP_SET(_name, _var, _union, _cond, _msg, _recomp)

#define PROP_ALL_REHANG(_name, _next, _var, _anchor, _msg, _flags)	\
PROP_ALL_WITHOUT_SET(_name##_y, _next, PT_REAL, PTR_COORDINATE, _var.y, r, LOW_DIE, _flags, UINT_MAX)	\
PROP_ALL_WITHOUT_SET(_name##_x, PSTRUCT(_name##_y), PT_REAL, PTR_COORDINATE, _var.x, r, LOW_DIE, _flags, UINT_MAX)	\
PROP_SET_REHANG_Y(_name, _var, _anchor, _msg) \
PROP_SET_REHANG_X(_name, _var, _anchor, _msg)

#define PROP_ALL_READ(_name, _next, _prop_type, _prop_subtype, _var, _union, _flags, _undo_mask)	\
PROP_ALL_WITHOUT_SET(_name, _next, _prop_type, _prop_subtype, _var, _union, LOW_DIE, _flags | PTF_READ_ONLY, _undo_mask)	\
PROP_SET_READ(_name)

#define PROP_ALL_WITHOUT_UNIT_AND_SET(_name, _next, _prop_type, _prop_subtype, _var, _union, _low, _flags, _undo_mask)	\
static struct prop_virtual PSTRUCT(_name);	\
PROP_GET(_name, _var, _union)	\
static void PSET(_name) (struct o *o, prop_value value); \
PROP_SET_##_low(_name, _var, _union) \
PROP_UNIT_GET_DIE(_name)	\
PROP_UNIT_SET_DIE(_name)	\
PROP_STRUCT(_name, _next, _prop_type, _prop_subtype, _flags, _undo_mask)

#define PROP_ALL_WITHOUT_UNIT(_name, _next, _prop_type, _prop_subtype, _var, _union, _cond, _msg, _recomp, _low, _flags, _undo_mask)	\
PROP_ALL_WITHOUT_UNIT_AND_SET(_name, _next, _prop_type, _prop_subtype, _var, _union, _low, _flags, _undo_mask)	\
PROP_SET(_name, _var, _union, _cond, _msg, _recomp) \

#define PROP_ALL_BOOLEAN(_name, _next, _var, _mask, _recomp, _flags, _undo_mask)	\
static struct prop_virtual PSTRUCT(_name);	\
PROP_GET(_name, _var & _mask ? 1 : 0, u)	\
PROP_SET_BOOLEAN(_name, _var, _mask, _recomp) \
PROP_SET_LOW_BOOLEAN(_name, _var, _mask) \
PROP_UNIT_GET_DIE(_name)	\
PROP_UNIT_SET_DIE(_name)	\
PROP_STRUCT(_name, _next, PT_UNS, PTU_BOOLEAN, _flags, _undo_mask)


#define PROP_INVISIBLE PROP_ALL_BOOLEAN(invisible, o_prop_name, go.flags, GOF_INVISIBLE, ALTERED, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
#define PROP_INVISIBLE_INIT &PSTRUCT(invisible), string_invisible->text

#define PROP_FILL_COLOR PROP_ALL_WITHOUT_UNIT(fill_color, PSTRUCT(invisible), PT_UNS, PTU_COLOR, fill_color, u, \
		1, (("Cannot fail."), ("Bug, fix me.")), ALTERED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
#define PROP_FILL_COLOR_INIT &PSTRUCT(fill_color), string_fill_color->text

#define PROP_STROKE_COLOR PROP_ALL_WITHOUT_UNIT(stroke_color, PSTRUCT(fill_color), PT_UNS, PTU_COLOR, stroke_color, u, \
		1, (("Cannot fail."), ("Bug, fix me.")), ALTERED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
#define PROP_STROKE_COLOR_INIT &PSTRUCT(stroke_color), string_stroke_color->text

#define PROP_STROKE_WIDTH PROP_ALL(stroke_width, PSTRUCT(stroke_color), PT_REAL, PTR_LENGTH, stroke_width, r, \
		value.r >= 0, (("Stroke width cannot be negative number."), ("Stroke width cannot be %f.", value.r)), CHANGED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
#define PROP_STROKE_WIDTH_INIT &PSTRUCT(stroke_width), string_stroke_width->text

#define PROP_STROKE_JOIN PROP_ALL_WITHOUT_UNIT(stroke_join, PSTRUCT(stroke_width), PT_UNS, PTU_JOIN_STYLE, stroke_join, u, \
		value.u < PSJ_MAX, (("Invalid value."), ("Stroke join cannot be %u.", value.u)), ALTERED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
#define PROP_STROKE_JOIN_INIT &PSTRUCT(stroke_join), string_stroke_join->text

#define PROP_STROKE_CAP PROP_ALL_WITHOUT_UNIT(stroke_cap, PSTRUCT(stroke_join), PT_UNS, PTU_CAP_STYLE, stroke_cap, u, \
		value.u < PSC_MAX, (("Invalid value."), ("Stroke cap cannot be %u.", value.u)), ALTERED, LOW, PTF_SAVE | PTF_RECYCLABLE, UINT_MAX)
#define PROP_STROKE_CAP_INIT &PSTRUCT(stroke_cap), string_stroke_cap->text

#define PROP_STYLES PROP_INVISIBLE PROP_FILL_COLOR PROP_STROKE_COLOR PROP_STROKE_WIDTH PROP_STROKE_JOIN PROP_STROKE_CAP
#define PROP_STYLES_INIT PROP_INVISIBLE_INIT, PROP_STROKE_COLOR_INIT, PROP_STROKE_WIDTH_INIT, PROP_STROKE_CAP_INIT, PROP_STROKE_JOIN_INIT, PROP_FILL_COLOR_INIT
#define PROP_STYLES_LAST PSTRUCT(stroke_cap)
#define PROP_STYLES_SET_DEFAULT_VALUE(_go) \
	(_go)->fill_color = PROP_INIT_FILL_COLOR; \
	(_go)->stroke_color = PROP_INIT_STROKE_COLOR; \
	(_go)->stroke_width = PROP_INIT_STROKE_WIDTH; \
	(_go)->stroke_cap = PROP_INIT_STROKE_CAP; \
	(_go)->stroke_join = PROP_INIT_STROKE_JOIN; \
	(_go)->unit_stroke_width = PROP_UNIT_MAX; \

#define IG_STRUCT GLUEE(GO_NAME_ID, _info)
#define ALL_INIT GLUEE(GO_NAME_ID, _init)
#define ALL_FINISH GLUEE(GO_NAME_ID, _finish)
#define IH_STRUCT(_sid) GLUEE3(hanger_, GO_SUBNAME(GO_ID, _sid), _info)
#define IH_STRUCT_ID IH_STRUCT(GO_SUBID)
#define IH_INIT(_sid) GLUEE(IH_STRUCT(_sid), _init)
#define IH_INIT_ID IH_INIT(GO_SUBID)
#define IH_GET(_hanger_name) GLUEE4(hanger_, GO_SUBNAME_ID, _get_, _hanger_name)
#define IA_STRUCT(_sid) GLUEE3(anchor_, GO_SUBNAME(GO_ID, _sid), _info)
#define IA_STRUCT_ID IA_STRUCT(GO_SUBID)
#define IA_INIT(_sid) GLUEE(IA_STRUCT(_sid), _init)
#define IA_INIT_ID IA_INIT(GO_SUBID)


#define IG_STRUCT_FCE(_name, _struct, _init, _finish, _subgo_count) \
	struct go_info IG_STRUCT[] = { _struct }; \
	void ALL_INIT(void) { string_go_type[GO_TYPE_ID] = string_lookup_ref(_name); _init } \
	void ALL_FINISH(void) { go_info_finish(IG_STRUCT, _subgo_count); _finish }

#define IG_STRUCT_ONE(_sid, _hc, _curve, _hstart_, _hend, _center, _ac, _astart, _aend, _vp) [GO_SUBTYPE(GO_ID, _sid)] = \
	{ NULL, _hc, OFFSETOF(GO_STRUCT_ID, h), 0, IH_STRUCT(_sid), _curve, _hstart_, _hend, _center, _ac, \
	OFFSETOF(GO_STRUCT_ID, a), IA_STRUCT(_sid), _astart, _aend, &PSTRUCT_SID(_vp, _sid) },
#define IG_STRUCT_1(_sid, _name, _hc, _curve, _hstart_, _hend, _center, _ac, _astart, _aend, _vp) \
	IG_STRUCT_ONE(_sid, _hc, _curve, _hstart_, _hend, _center, _ac, _astart, _aend, _vp)
#define IG_STRUCT_2(_sid, _name, _hc, _curve, _hstart_, _hend, _center, _ac, _astart, _aend, _vp, ...) \
	IG_STRUCT_ONE(_sid, _hc, _curve, _hstart_, _hend, _center, _ac, _astart, _aend, _vp) IG_STRUCT_1(__VA_ARGS__)
#define IG_STRUCT_3(_sid, _name, _hc, _curve, _hstart_, _hend, _center, _ac, _astart, _aend, _vp, ...) \
	IG_STRUCT_ONE(_sid, _hc, _curve, _hstart_, _hend, _center, _ac, _astart, _aend, _vp) IG_STRUCT_2(__VA_ARGS__)
#define IG_STRUCT_4(_sid, _name, _hc, _curve, _hstart_, _hend, _center, _ac, _astart, _aend, _vp, ...) \
	IG_STRUCT_ONE(_sid, _hc, _curve, _hstart_, _hend, _center, _ac, _astart, _aend, _vp) IG_STRUCT_3(__VA_ARGS__)
#define IG_STRUCT_5(_sid, _name, _hc, _curve, _hstart_, _hend, _center, _ac, _astart, _aend, _vp, ...) \
	IG_STRUCT_ONE(_sid, _hc, _curve, _hstart_, _hend, _center, _ac, _astart, _aend, _vp) IG_STRUCT_4(__VA_ARGS__)

#define IG_INIT_ONE(_sid, _name) IG_STRUCT[GO_SUBTYPE(GO_ID, _sid)].name = string_lookup_ref(_name); IH_INIT(_sid)(); IA_INIT(_sid)(); PINIT(_sid)();
#define IG_INIT_1(_sid, _name, _hc, _curve, _hstart, _hend, _center, _ac, _astart, _aend, _vp) IG_INIT_ONE(_sid, _name)
#define IG_INIT_2(_sid, _name, _hc, _curve, _hstart, _hend, _center, _ac, _astart, _aend, _vp, ...) IG_INIT_ONE(_sid, _name) IG_INIT_1(__VA_ARGS__)
#define IG_INIT_3(_sid, _name, _hc, _curve, _hstart, _hend, _center, _ac, _astart, _aend, _vp, ...) IG_INIT_ONE(_sid, _name) IG_INIT_2(__VA_ARGS__)
#define IG_INIT_4(_sid, _name, _hc, _curve, _hstart, _hend, _center, _ac, _astart, _aend, _vp, ...) IG_INIT_ONE(_sid, _name) IG_INIT_3(__VA_ARGS__)
#define IG_INIT_5(_sid, _name, _hc, _curve, _hstart, _hend, _center, _ac, _astart, _aend, _vp, ...) IG_INIT_ONE(_sid, _name) IG_INIT_4(__VA_ARGS__)

#define IG_FINISH_ONE(_sid) PFINISH(_sid) ();
#define IG_FINISH_1(_sid, _name, _hc, _curve, _hstart, _hend, _center, _ac, _astart, _aend, _vp) IG_FINISH_ONE(_sid)
#define IG_FINISH_2(_sid, _name, _hc, _curve, _hstart, _hend, _center, _ac, _astart, _aend, _vp, ...) IG_FINISH_ONE(_sid) IG_FINISH_1(__VA_ARGS__)
#define IG_FINISH_3(_sid, _name, _hc, _curve, _hstart, _hend, _center, _ac, _astart, _aend, _vp, ...) IG_FINISH_ONE(_sid) IG_FINISH_2(__VA_ARGS__)
#define IG_FINISH_4(_sid, _name, _hc, _curve, _hstart, _hend, _center, _ac, _astart, _aend, _vp, ...) IG_FINISH_ONE(_sid) IG_FINISH_3(__VA_ARGS__)
#define IG_FINISH_5(_sid, _name, _hc, _curve, _hstart, _hend, _center, _ac, _astart, _aend, _vp, ...) IG_FINISH_ONE(_sid) IG_FINISH_4(__VA_ARGS__)

#define INFO_GO_1(_name, ...) IG_STRUCT_FCE(_name, IG_STRUCT_1(__VA_ARGS__), IG_INIT_1(__VA_ARGS__), IG_FINISH_1(__VA_ARGS__), 1)
#define INFO_GO_2(_name, ...) IG_STRUCT_FCE(_name, IG_STRUCT_2(__VA_ARGS__), IG_INIT_2(__VA_ARGS__), IG_FINISH_2(__VA_ARGS__), 2)
#define INFO_GO_3(_name, ...) IG_STRUCT_FCE(_name, IG_STRUCT_3(__VA_ARGS__), IG_INIT_3(__VA_ARGS__), IG_FINISH_3(__VA_ARGS__), 3)
#define INFO_GO_4(_name, ...) IG_STRUCT_FCE(_name, IG_STRUCT_4(__VA_ARGS__), IG_INIT_4(__VA_ARGS__), IG_FINISH_4(__VA_ARGS__), 4)
#define INFO_GO_5(_name, ...) IG_STRUCT_FCE(_name, IG_STRUCT_5(__VA_ARGS__), IG_INIT_5(__VA_ARGS__), IG_FINISH_5(__VA_ARGS__), 5)


#define IH_STRUCT_FCE(_struct, _fce) static struct hanger_info IH_STRUCT_ID[] = { _struct }; static void IH_INIT_ID(void) { _fce }

#define IH_STRUCT_ONE(_i, _q, _var, _get, _lin) { NULL, _q, OFFSETOF(GO_STRUCT_ID, h[_i]), IG_STRUCT + GO_SUBTYPE_ID, \
	OFFSETOF(GO_STRUCT_ID, _var) - OFFSETOF(GO_STRUCT_ID, h[_i]), _get, _lin },
#define IH_STRUCT_1(_i, _n, _q, _var, _get, _lin) IH_STRUCT_ONE(_i, _q, _var, _get, _lin)
#define IH_STRUCT_2(_i, _n, _q, _var, _get, _lin, ...) IH_STRUCT_ONE(_i, _q, _var, _get, _lin) IH_STRUCT_1(_i + 1, __VA_ARGS__)
#define IH_STRUCT_3(_i, _n, _q, _var, _get, _lin, ...) IH_STRUCT_ONE(_i, _q, _var, _get, _lin) IH_STRUCT_2(_i + 1, __VA_ARGS__)
#define IH_STRUCT_4(_i, _n, _q, _var, _get, _lin, ...) IH_STRUCT_ONE(_i, _q, _var, _get, _lin) IH_STRUCT_3(_i + 1, __VA_ARGS__)

#define IH_FCE_ONE(_i, _n) IH_STRUCT_ID[_i].name = string_lookup_ref(_n); glt_kernel_init_hanger(&IH_STRUCT_ID[_i]);
#define IH_FCE_1(_i, _n, _q, _var, _get, _lin) IH_FCE_ONE(_i, _n)
#define IH_FCE_2(_i, _n, _q, _var, _get, _lin, ...) IH_FCE_ONE(_i, _n) IH_FCE_1(_i + 1, __VA_ARGS__)
#define IH_FCE_3(_i, _n, _q, _var, _get, _lin, ...) IH_FCE_ONE(_i, _n) IH_FCE_2(_i + 1, __VA_ARGS__)
#define IH_FCE_4(_i, _n, _q, _var, _get, _lin, ...) IH_FCE_ONE(_i, _n) IH_FCE_3(_i + 1, __VA_ARGS__)

#define INFO_HANGER_0 static struct hanger_info IH_STRUCT_ID[1]; static void IH_INIT_ID(void) { }
#define INFO_HANGER_1(...) IH_STRUCT_FCE(IH_STRUCT_1(0, __VA_ARGS__), IH_FCE_1(0, __VA_ARGS__))
#define INFO_HANGER_2(...) IH_STRUCT_FCE(IH_STRUCT_2(0, __VA_ARGS__), IH_FCE_2(0, __VA_ARGS__))
#define INFO_HANGER_3(...) IH_STRUCT_FCE(IH_STRUCT_3(0, __VA_ARGS__), IH_FCE_3(0, __VA_ARGS__))
#define INFO_HANGER_4(...) IH_STRUCT_FCE(IH_STRUCT_4(0, __VA_ARGS__), IH_FCE_4(0, __VA_ARGS__))


#define IA_STRUCT_FCE(_struct, _fce) static struct anchor_info IA_STRUCT_ID[] = { _struct }; static void IA_INIT_ID (void) { _fce }

#define IA_STRUCT_ONE(_i, _q, _lin) { NULL, _q, OFFSETOF(GO_STRUCT_ID, a[_i]), IG_STRUCT + GO_SUBTYPE_ID, _lin },
#define IA_STRUCT_1(_i, _n, _q, _lin) IA_STRUCT_ONE(_i, _q, _lin)
#define IA_STRUCT_2(_i, _n, _q, _lin, ...) IA_STRUCT_ONE(_i, _q, _lin) IA_STRUCT_1(_i + 1, __VA_ARGS__)
#define IA_STRUCT_3(_i, _n, _q, _lin, ...) IA_STRUCT_ONE(_i, _q, _lin) IA_STRUCT_2(_i + 1, __VA_ARGS__)
#define IA_STRUCT_4(_i, _n, _q, _lin, ...) IA_STRUCT_ONE(_i, _q, _lin) IA_STRUCT_3(_i + 1, __VA_ARGS__)

#define IA_FCE_ONE(_i, _n) IA_STRUCT_ID[_i].name = string_lookup_ref(_n); glt_kernel_init_anchor(&IA_STRUCT_ID[_i]);
#define IA_FCE_1(_i, _n, _q, _lin) IA_FCE_ONE(_i, _n)
#define IA_FCE_2(_i, _n, _q, _lin, ...) IA_FCE_ONE(_i, _n) IA_FCE_1(_i + 1, __VA_ARGS__)
#define IA_FCE_3(_i, _n, _q, _lin, ...) IA_FCE_ONE(_i, _n) IA_FCE_2(_i + 1, __VA_ARGS__)
#define IA_FCE_4(_i, _n, _q, _lin, ...) IA_FCE_ONE(_i, _n) IA_FCE_3(_i + 1, __VA_ARGS__)

#define INFO_ANCHOR_0(...) IA_STRUCT_FCE(,)
#define INFO_ANCHOR_1(...) IA_STRUCT_FCE(IA_STRUCT_1(0, __VA_ARGS__), IA_FCE_1(0, __VA_ARGS__))
#define INFO_ANCHOR_2(...) IA_STRUCT_FCE(IA_STRUCT_2(0, __VA_ARGS__), IA_FCE_2(0, __VA_ARGS__))
#define INFO_ANCHOR_3(...) IA_STRUCT_FCE(IA_STRUCT_3(0, __VA_ARGS__), IA_FCE_3(0, __VA_ARGS__))
#define INFO_ANCHOR_4(...) IA_STRUCT_FCE(IA_STRUCT_4(0, __VA_ARGS__), IA_FCE_4(0, __VA_ARGS__))

#endif
