/*
 * Kernel - Testing-only functions
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define LOCAL_DEBUG

#include "lib/lib.h"
#include "geomlib/geomlib.h"
#include "kernel/kernel.h"

void pattern_test(void);
void pattern_test(void)
{
	struct obj_tlo *tlo;
	char err_msg_main[TRANS_ERR_SIZE];
	struct geom_pattern pattern;
	struct geom_bezier bezier;
	struct geom_point pt0, pt1, pt2, pt3;
	struct geom_path path;
	struct geom_bezier *item;

	// FIXME: Osklive, funkce by mela dostavat tlo jako argument. Nebo by mela dostavat i grupu.
	tlo = obj2tlo(obj_get_first_son(obj_get_first_son(obj_universum)));

	TRANS_BEGIN_MAIN(tlo, "Tsunami", err_msg_main)

		DBG("Starting pattern-test.\n");

		GEOM_ITEM_INIT(&bezier, bezier);
		GEOM_ITEM_CREATE(&path, path);
		pt0 = (struct geom_point){-50, -50};
		pt1 = (struct geom_point){50, 20};
		pt2 = (struct geom_point){-70, 10};
		pt3 = (struct geom_point){10, 50};
		geom_bezier_set_cubic_nonrational(&bezier, &pt0, &pt1, &pt2, &pt3);
		go_link_head(&go_decorator_point_new(hanger_new_mc(&pt0), 0, 3, 0)->go, tlo_get_tlg(tlo));
		go_link_head(&go_decorator_point_new(hanger_new_mc(&pt3), 0, 3, 0)->go, tlo_get_tlg(tlo));
	
		real len = 3;
		real space = 1;
		pattern.start_count = 0;
		pattern.period_count = 4;
		pattern.period_items = xmalloc(pattern.period_count * sizeof(struct geom_pattern_item));
		pattern.period_items[0].alength = len;
		pattern.period_items[0].stretch = 1;
		pattern.period_items[0].flags = GEOM_PATTERN_LINE;
		pattern.period_items[1].alength = space;
		pattern.period_items[1].stretch = 1;
		pattern.period_items[1].flags = GEOM_PATTERN_SPACE;
		pattern.period_items[2].alength = 0.2;
		pattern.period_items[2].stretch = 1;
		pattern.period_items[2].flags = GEOM_PATTERN_LINE;
		pattern.period_items[3].alength = space;
		pattern.period_items[3].stretch = 1;
		pattern.period_items[3].flags = GEOM_PATTERN_SPACE;
		pattern.end_count = 1;
		pattern.end_items = xmalloc(pattern.end_count * sizeof(struct geom_pattern_item));
		pattern.end_items[0].alength = len;
		pattern.end_items[0].stretch = 1;
		pattern.end_items[0].flags = GEOM_PATTERN_LINE;

		if (!GEOM_ETEST(geom_item_pattern_append(&bezier, &pattern, &path))) trans_fail("geom_item_pattern_append failed");

		GEOM_GROUP_WALK(&path, item) {
			DBG("Inserting segment\n");
			go_link_head(&go_segment_new(hanger_new_mc((struct geom_point *)&item->pt[0]), hanger_new_mc((struct geom_point *)&item->pt[1]))->go,
				tlo_get_tlg(tlo));
		}

		geom_path_destroy(&path);
		geom_bezier_destroy(&bezier);
		xfree(pattern.period_items);
		xfree(pattern.end_items);
		
		DBG("Ending pattern-test.\n");

	TRANS_FAILED

		DBG("Failed :-(\n");
		
	TRANS_END
}

void trans_test(void);
void trans_test(void)
{
	struct obj_doc* volatile doc;
	struct obj_tlo* volatile tlo;
	char err[TRANS_ERR_SIZE], err2[TRANS_ERR_SIZE], err3[TRANS_ERR_SIZE], err4[TRANS_ERR_SIZE], err5[TRANS_ERR_SIZE], err6[TRANS_ERR_SIZE];
	struct geom_point gp;
	gp.x = gp.y = 0;
	struct hanger* h1, *h2;
	struct go* g1,* g2;

	TRANS_BEGIN_MAIN(tlo_universum, "trans test 1", err)
		doc = doc_new_named_link("test doc");
	TRANS_FAILED
		msg("trans test fail: %s", err);
	TRANS_END

	TRANS_BEGIN_ANONYMOUS(err)
		TRANS_BEGIN_MAIN(tlo_universum, "trans test 2", err2)
			TRANS_BEGIN(err3)
				tlo = tlo_new_named_link(doc, "trans test 4");
				TRANS_BEGIN_MAIN(tlo, "trans test 4", err4)
					TRANS_BEGIN_MAIN(tlo_universum, "trans test 5", err5)
						TRANS_BEGIN_ANONYMOUS(err6)
							trans_fail("fail test 6");
						TRANS_FAILED
							msg("trans test 6 fail: %s\n", err6);
						TRANS_END
						trans_fail("fail test 5");
					TRANS_FAILED
						msg("trans test 5 fail: %s\n", err5);
						trans_fail("fail test 4");
					TRANS_END
						trans_fail("fail test 4");
				TRANS_FAILED
					msg("trans test 4 fail: %s\n", err4);
				TRANS_END
			TRANS_FAILED
				msg("trans test 3 fail: %s\n", err3);
			TRANS_END
		TRANS_FAILED
			msg("trans test 2 fail: %s\n", err2);
		TRANS_END
	TRANS_FAILED
		msg("trans test 1 fail: %s", err);
	TRANS_END

	TRANS_BEGIN_MAIN(tlo, "trans test", err)
		h1 = hanger_new_mc(&gp);
		h2 = hanger_new_mc(&gp);
		g1 = &go_segment_new(h1, h2)->go;
		h1 = go_get_hanger(g1, 1);
		h2 = go_get_hanger(g1, 2);
		g2 = &go_segment_new(h1, h2)->go;
		go_link_head(g2, tlo_get_tlg(tlo));
		go_link_head(g1, tlo_get_tlg(tlo));
		trans_fail("test fail");
	TRANS_FAILED
		msg("trans test fail: %s\n", err);
	TRANS_END
}
