#ifndef _KERNEL_TRANS_H_
#define _KERNEL_TRANS_H_

/*
 * Kernel - Transaction subsystem
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define TRANS_ERR_SIZE 256

#define TRANS_BEGIN_MAIN(_tlo, _name, _es)			\
do {								\
		jmp_buf _es##jb;				\
		trans_begin_main(_tlo, _name, _es, &_es##jb);	\
		if(!setjmp(_es##jb))				\
		{	do	{

#define TRANS_BEGIN(_es)					\
do {								\
		jmp_buf _es##jb;				\
		trans_begin_sub(_es, &_es##jb);			\
		if(!setjmp(_es##jb))				\
		{	do	{

#define TRANS_BEGIN_ANONYMOUS(_es)				\
do {								\
		jmp_buf _es##jb;				\
		trans_begin_anonymous(_es, &_es##jb);		\
		if(!setjmp(_es##jb))				\
		{	do	{

#define TRANS_FAILED						\
				}	while(0);		\
				trans_end();			\
		} 						\
		else do 	{

#define TRANS_END						\
					} while(0);		\
	} while(0);


void trans_fail(const char* format, ...) FORMAT(1,2) NONRET;

/*
 * Register function func to be called when current transaction fails.
 */
void trans_log(void (*func)(uintptr_t arg), uintptr_t arg);

/* Undo top-level gui command. Do not call inside transaction. */
uns trans_undo(struct obj_tlo* tlo, char *error_string);

/* Redo top-level gui command. Do not call inside transaction. */
uns trans_redo(struct obj_tlo* tlo, char *error_string);

/* Clear the oldest item of undo history. Undo will has max_item items (if it
 * has more now). Do not call inside transaction. */
SCHEME SVOID trans_clear_undo(SO_obj_tlo tlo);

void trans_clear_one_redo(struct obj_tlo* tlo);
void trans_clear_all_redo(struct obj_tlo* tlo);
struct undo_gui* trans_merge_undo(struct obj_tlo *tlo, struct undo_gui *first, string name);
void trans_rename_undo(struct obj_tlo *tlo, struct undo_gui *ug, string name);

/* Undo history movement */
struct undo_gui* trans_get_first_gui(struct obj_tlo* t);
struct undo_gui* trans_get_next_gui(struct obj_tlo* t, struct undo_gui *ug);
struct undo_gui* trans_get_last_gui(struct obj_tlo* t);
struct undo_gui* trans_get_prev_gui(struct obj_tlo* t, struct undo_gui *ug);
uns trans_get_gui_rank(struct obj_tlo* t, struct undo_gui *ug);


/*
 * Transaction hooks, data structure, global variable, filling, adding & removing
 */

static inline struct trans_hook* trans_hook_fill(
		struct trans_hook *h,
		void* data,
		void (*undo_new) (struct obj_tlo *tlo, struct undo_gui* ug, struct trans_hook* h),
		void (*undo_delete) (struct obj_tlo *tlo, struct undo_gui* ug, struct undo_gui *prev_item, struct trans_hook* h),
		void (*undo_update) (struct obj_tlo *tlo, struct undo_gui* ug, struct trans_hook* h),
		void (*undo) (struct obj_tlo* tlo, struct trans_hook* h),
		void (*redo) (struct obj_tlo *tlo, struct trans_hook* h),
		void (*saved) (struct obj * obj, struct trans_hook* h))
	{ h->data = data; h->undo_new = undo_new; h->undo_delete = undo_delete; h->undo_update = undo_update; h->undo = undo; h->redo = redo; h->saved = saved; return h; }

struct trans_hook* trans_hook_add(struct obj_tlo *tlo, struct trans_hook *hook);
struct trans_hook* trans_hook_remove(struct obj_tlo *tlo, struct trans_hook *hook);
struct trans_hook* trans_hook_remove_by_data(struct obj_tlo *tlo, void * data);


/* Transaction dump function */
void trans_dump_undo_kernel(struct undo_gui * ug);
SCHEME SVOID trans_dump_undo_gui(SO_obj_tlo tlo);
SCHEME SVOID trans_dump_undo(SO_obj_tlo tlo);


/* Internal function which are called in macros TRANS_BEGIN* */
void trans_begin_main(struct obj_tlo* t, const char *name, char* error_string, jmp_buf* jb);
void trans_begin_sub(char* error_string, jmp_buf* jb);
void trans_begin_anonymous(char* error_string, jmp_buf* jb);
void trans_end(void);

#endif
