/*
 * VRRLIB - AVL trees
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/lib.h"
#include "lib/avltree.h"


/* avltree_node_rebalance */

void avltree_node_rebalance(struct avltree *tree, struct avlnode *node)
{
	struct avlnode *x, *y, *z, *p;
	char h, ah, bh, ch, dh;

	ASSERT(tree && node);

	for (; node; node = p) {

		h = node->height;
		x = node->left;
		z = node->right;
		ah = x ? x->height : 0;
		dh = z ? z->height : 0;
		p = node->parent;
		if ((int)(dh) - (int)ah == -2) {
			z = node;
			ah = x->left ? x->left->height : 0;
			ch = x->right ? x->right->height : 0;
			if (ah >= ch) { /* LL rotation */
				y = x;
				x = y->left;
			}
			else { /* LR rotation */
				y = x->right;
				x->right = y->left;
				bh = y->left ? y->left->height : 0;
				ch = y->right ? y->right->height : 0;
				x->height = MAX(ah, bh) + 1;
			}
			z->left = y->right;
			z->height = MAX(ch, dh) + 1;
		}
		else if (dh - ah == 2) {
			x = node;
			bh = z->left ? z->left->height : 0;
			dh = z->right ? z->right->height : 0;
			if (bh <= dh) { /* RR rotation */
				y = z;
				z = y->right;
			}
			else { /* RL rotation */
				y = z->left;
				z->left = y->right;
				bh = y->left ? y->left->height : 0;
				ch = y->right ? y->right->height : 0;
				z->height = MAX(ch, dh) + 1;
			}
			x->right = y->left;
			x->height = MAX(ah, bh) + 1;
		}
		else {
			if ((node->height = MAX(ah, dh) + 1) == h)
				return;
			continue;
		}
		if (!p)
			tree->root = y;
		else if (p->left == node)
			p->left = y;
		else
			p->right = y;
		y->parent = p;
		y->left = x;
		y->right = z;
		x->parent = y;
		z->parent = y;
		if (x->right)
			x->right->parent = x;
		if (z->left)
			z->left->parent = z;
		if ((y->height = MAX(x->height, z->height) + 1) == h)
			return;
	}
}


/* avltree_add_head */

void avltree_add_head(struct avltree *tree, struct avlnode *node)
{
	struct avlnode *parent;

	ASSERT(tree && node);

	node->left = node->right = NULL; node->height = 1;
	if (tree->root) {
		parent = avltree_head(tree);
		parent->left = node;
		node->parent = parent;
		avltree_node_rebalance(tree, parent);
	}
	else {
		node->parent = NULL;
		tree->root = node;
	}
}


/* avltree_add_tail */

void avltree_add_tail(struct avltree *tree, struct avlnode *node)
{
	struct avlnode *parent;

	ASSERT(tree && node);

	node->left = node->right = NULL; node->height = 1;
	if (tree->root) {
		parent = avltree_tail(tree);
		parent->right = node;
		node->parent = parent;
		avltree_node_rebalance(tree, parent);
	}
	else {
		node->parent = NULL;
		tree->root = node;
	}
}


/* avltree_insert_before */

void avltree_insert_before(struct avltree *tree, struct avlnode *node, struct avlnode *before)
{
	struct avlnode *parent;

	ASSERT(tree && node && before);

	node->left = node->right = NULL; node->height = 1;
	if (parent = before->left) {
		while (parent->right) parent = parent->right;
		parent->right = node;
	}
	else
		(parent = before)->left = node;
	node->parent = parent;
	avltree_node_rebalance(tree, parent);
}


/* avltree_insert_after */

void avltree_insert_after(struct avltree *tree, struct avlnode *node, struct avlnode *after)
{
	struct avlnode *parent;

	ASSERT(tree && node && after);

	node->left = node->right = NULL; node->height = 1;
	if (parent = after->right) {
		while (parent->left) parent = parent->left;
		parent->left = node;
	}
	else
		(parent = after)->right = node;
	node->parent = parent;
	avltree_node_rebalance(tree, parent);
}


/* avltree_remove */

void avltree_remove(struct avltree *tree, struct avlnode *node)
{
	struct avlnode *p, *q, **r, **s, *t;

	ASSERT(tree && node);

	if (p = node->parent)
		s = (node == p->left) ? &p->left : &p->right;
	else
		s = &tree->root;
	if (!node->left) {
		if (*s = node->right)
			node->right->parent = p;
	}
	else if (!node->right)
		(*s = node->left)->parent = p;
	else {
		t = node->left;
		for (r = &node->left; (q = *r)->right; r = &q->right)
			t = q;
		if (*r = q->left)
			(*r)->parent = t;
		*s = q;
		q->height = node->height;
		q->parent = p;
		if (p = q->left = node->left)
			p->parent = q;
		(q->right = node->right)->parent = q;
		p = t;
	}
	if (p)
		avltree_node_rebalance(tree, p);
}


/* avltree_remove_head */

void avltree_remove_head(struct avltree *tree)
{
	// FIXME: faster code
	avltree_remove(tree, avltree_head(tree));
}


/* avltree_remove_tail */

void avltree_remove_tail(struct avltree *tree)
{
	// FIXME: faster code
	 avltree_remove(tree, avltree_tail(tree));
}


/* avltree_common_parent */

/*struct avlnode *avltree_common_parent(struct avlnode *node1, struct avlnode *node2)
{
	ASSERT(node1 && node2);

	while (1) {
		if (node1->height < node2->height)
			node1 = node1->parent;
		else if (node1->height > node2->height)
			node2 = node2->parent;
		else if (node1 == node2)
			return node1;
		else {
			node1 = node1->parent;
			node2 = node2->parent;
		}
	}
}*/


/* avltree_order_compare */

/*int avltree_order_compare(struct avlnode *node1, struct avlnode *node2)
{
	struct avlnode *prev1, *prev2;

	ASSERT(node1 && node2);

	if (node1 == node2)
		return 0;
	prev1 = node1;
	while (1) {
		if (node1->height < node2->height)
			node1 = (prev1 = node1)->parent;
		else if (node1->height > node2->height)
			node2 = (prev2 = node2)->parent;
		else if (node1 == node2) {
			if (prev1 != node1)
				if (prev1 == node1->left)
					return -1;
				else
					return 1;
			else
				if (prev2 == node1->left)
					return 1;
				else
					return -1;
		}
		else {
			node1 = (prev1 = node1)->parent;
			node2 = node2->parent;
		}
	}
}*/


/* avltree_order_sort */

/*#define ASORT_PREFIX(x) avltree_order_sort_##x
#define ASORT_EXTRA_ARGS , struct avlnode **nodes
#define ASORT_KEY_TYPE struct avlnode *
#define ASORT_ELT(x) nodes[x]
#define ASORT_LT(x,y) (avltree_order_compare(x, y) < 0)
#include "arraysort.h"

void avltree_order_sort(uns n, struct avlnode **nodes)
{
	// FIXME: faster code
	avltree_order_sort_sort(n, nodes);
}*/


