/*
 * VRRLIB - Memory allocation
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/lib.h"

#include <stdlib.h>
#include <string.h>

#ifndef DEBUG_MALLOC

void *xmalloc(uns size)
{
	void *p = malloc(size);
	if (!p)
		die("Can't allocate %d bytes of memory", size);
	return p;
}

void *xmalloc_nonfail(uns size)
{
	return malloc(size);
}

void *xmalloc_zero(uns size)
{
	void *p = xmalloc(size);
	memset(p, 0, size);
	return p;
}

void xfree(void *p)
{
	free(p);
}

void *xrealloc(void *old, uns size)
{
	void *p = realloc(old, size);
	if (!p)
		die("Can't reallocate %d bytes od memory", size);
	return p;
}

char *xstrdup(const char* string)
{
	uns len = strlen(string) + 1;
	return memcpy(xmalloc(len), string, len);
}

char *xstrndup(const char *string, uns n)
{
	uns i = 0;
	char *tmp;

	while (i <= n)
		if (!string[i++])
			break;
	tmp = memcpy (xmalloc(i), string, i);
	tmp[i-1] = 0;
	return tmp;
}

#endif

