
(define autolink-state (make-fluid))
(fluid-set! autolink-state #t)

(define user-autolink-state #t)

(define (set-autolink-state! value)
  (if (boolean? value)
      (begin
	(set! user-autolink-state value)
	(fluid-set! autolink-state value))
      (typearg-error "set-autolink-state!" 1 value)))

(define (get-autolink-state)
  (fluid-ref autolink-state))

(define (call-with-autolink-state value thunk)
  (with-fluids ((autolink-state value))
	       (thunk)))

(define (autolink go)
  (if (get-autolink-state)
      (go-link-head go (tlo-get-tlg (get-trans-tlo)))
      go))

(define (make-idtable . items)
  (let* ((l (length items))
	 (t1 (make-vector l))
	 (t2 (make-hash-table (* l 2))))
    (for-each (lambda (key value)
		(vector-set! t1 value key)
		(hashq-set! t2 key value))
	      items (iota l))
    (cons t1 t2)))

(define (make-idtable-lookup-1 idt)
  (lambda (n) (vector-ref (car idt) n)))

(define (make-idtable-lookup-2 idt)
  (lambda (k) (hashq-ref (cdr idt) k)))

(define conic-table (make-idtable
		     'start-entire 'start-dif 'point-entire 'point-dif
		     'ccw 'cw 'smaller 'bigger 'middle 'opposite 'entire))

(define code-conic (make-idtable-lookup-2 conic-table))
(define decode-conic (make-idtable-lookup-1 conic-table))

(define cap-style-table (make-idtable 'butt 'round 'projecting))

(define code-cap-style (make-idtable-lookup-2 cap-style-table))
(define decode-cap-style (make-idtable-lookup-1 cap-style-table))

(define join-style-table (make-idtable 'miter 'round 'bevel))

(define code-join-style (make-idtable-lookup-2 join-style-table))
(define decode-join-style (make-idtable-lookup-1 join-style-table))

(define alignment-x-table (make-idtable 'ref-left 'ref-center 'ref-right 'ref-relative
					'bbox-left 'bbox-center 'bbox-right 'bbox-relative))

(define code-alignment-x (make-idtable-lookup-2 alignment-x-table))
(define decode-alignment-x (make-idtable-lookup-1 alignment-x-table))

(define alignment-y-table (make-idtable 'baseline 'bbox-left 'bbox-center 'bbox-right 'bbox-relative))

(define code-alignment-y (make-idtable-lookup-2 alignment-y-table))
(define decode-alignment-y (make-idtable-lookup-1 alignment-y-table))



(define (prepare fn fargs largs)
  (lambda args
    (autolink (apply fn (append fargs args largs)))))

(define a-vtables '(#t))
(define h-vtables '(#t))

(define rh-table (make-hash-table 31))
(define src-table (make-hash-table 31))
(define cons-table (make-hash-table 31))

(define (get-vtable vtlist op)
  (or (assq-ref (cdr vtlist) op)
      (begin
	(set-cdr! vtlist (assq-set! (cdr vtlist) op (make-hash-table 31)))
	(assq-ref (cdr vtlist) op))))


(define (define-source subtype srclist)
  (hashq-set! src-table subtype srclist))

(define (define-constructor subtype make-fn)
  (hashq-set! cons-table subtype make-fn))

(define (make-go subtype args)
  (apply (hashq-ref cons-table subtype) args))

;;; Common user functions


(define (property-get-subtype obj name)
  (let ((x (property-get-subtype-i obj name)))
    (and x (property-type-symbol (car x) (cdr x)))))

(define (property-get obj name . missing)
  (let ((st (property-get-int obj name)))
    (if st
	(prop-decode (cadr st) (caddr st) (car st))
	(if (null? missing) #f (car missing)))))

(define (property-set! obj name value)
  (cond
   ((real? value)
    (property-real-set obj name value))
   ((string? value)
    (property-string-set obj name value))
   ((boolean? value)
    (property-boolean-set obj name value))
   ((pair? value)
    (property-specific-set obj name (car value) (cdr value)))
   (else
    (typearg-error "property-set!" 3 value))))

(define (skip-head f)
  (if (pair? f)
      (cdr f)
      f))

(define (go-get-sources go)
  (let ((v (hashq-ref src-table (proxy-get-subtype go))))
    (and v (flatten (map (lambda (x) ; in v should be only numbers, getters or property names
			   (cond
			    ((number? x)
			     (if (negative? x)
				 (go-get-anchor go (- -1 x))
				 (go-get-anchor go x)))
			    ((symbol? x) (skip-head (property-get go x)))
			    ((procedure? x) (x go))
			    (else (error "BUG in go-get-sources"))))
			 (if (number? v) (iota v) v))))))

(define (go-get-source-dsc subtype)
  (let ((v (hashq-ref src-table subtype)))
    (and v (map (lambda (x) ; in v should be only numbers, getters or property names
		  (if (number? x)
		      (if (negative? x) 'hanger-c 'hanger)
		      'data))
		(if (number? v) (iota v) v)))))

(define coords mc-new)
(define (coords-c p) (coords (real-part p) (imag-part p)))

(define make-point (prepare go-point-new '() '() ))
(define make-segment (prepare go-segment-new '() '()))

(define make-ellipse-by-center-2-radii-rotation
  (lambda (c r1 r2 rot) (autolink (go-elarc-new-xyr c rot r1 r2 0 7 (code-conic 'start-entire)))))

(define make-ellipse-by-2-foci-point (prepare go-elarc-new-foci '() (list (code-conic 'point-entire))))
(define make-ellipse-by-3-points-smallest (prepare go-elarc-new-3small '() (list (code-conic 'point-entire))))
(define make-ellipse-by-3-points-rotation-eccentricity (prepare go-elarc-new-3ecc '() (list (code-conic 'point-entire))))
(define make-ellipse-by-center-point-rotation-eccentricity (prepare go-elarc-new-xy1ecc '() (list (code-conic 'point-entire))))

(define make-ellipse/c2rr make-ellipse-by-center-2-radii-rotation)
(define make-ellipse/2fp make-ellipse-by-2-foci-point)
(define make-ellipse/3ps make-ellipse-by-3-points-smallest)
(define make-ellipse/3pre make-ellipse-by-3-points-rotation-eccentricity)
(define make-ellipse/cpre make-ellipse-by-center-point-rotation-eccentricity)
(define make-ellipse make-ellipse-by-center-2-radii-rotation)

(define make-elarc-by-center-2-radii-rotation-2-angles
  (lambda (c r1 r2 rot a1 a2 int) (autolink (go-elarc-new-xyr c rot r1 r2 a1 a2 (code-conic int)))))

(define make-elarc-by-3-points-smallest
  (lambda (p1 p2 p3 int) (autolink (go-elarc-new-foci p1 p2 p3 (code-conic int)))))

(define make-elarc/c2rr2a make-elarc-by-center-2-radii-rotation-2-angles)
(define make-elarc/3ps make-elarc-by-3-points-smallest)

(define make-quadratic-bezier
  (lambda (p1 p2 p3) (autolink (go-bezier-new-quadratic p1 1 p2 1 p3 1))))
(define make-quadratic-rational-bezier (prepare go-bezier-new-quadratic '() '()))
(define make-cubic-bezier
  (lambda (p1 p2 p3 p4) (autolink (go-bezier-new-cubic p1 1 p2 1 p3 1 p4 1))))
(define make-cubic-rational-bezier (prepare go-bezier-new-cubic '() '()))
(define make-bezier/2 make-quadratic-bezier)
(define make-bezier/2r make-quadratic-rational-bezier)
(define make-bezier/3 make-cubic-bezier)
(define make-bezier/3r make-cubic-rational-bezier)

(define (make-tex-text h str)
  (autolink (go-tex-text-new h 0 0 0 0 str)))

(define (make-text h str family style size)
  (autolink (go-text-new h 0 0 0 0 str (font-load-by-family-style family style) size)))

					;  (lambda (p rx1 ry mx my str) (autolink (go-text-new 0 p1 1 p2 1 p3 1)))

(define make-parametric-point (prepare go-parametric-point-new '() '()))
(define make-parametric-point-on-hanger (prepare go-parametric-point-new-on-hanger '() '(#f 0 0)))

(define make-intersection (prepare go-intersection-new '() '()))
(define make-intersection-on-hanger (prepare go-intersection-new-on-hanger '() '(#f 0 0)))

(define make-decorator-point (prepare go-decorator-point-new '() '()))
(define make-decorator-arrow (prepare go-arrow-new '() '()))
(define make-decorator-arrow-on-hanger (prepare go-arrow-new-on-hanger '() '()))

(define make-group (prepare go-group-new '() '()))
(define make-path  (prepare go-path-new '() '()))

(define-constructor 'point make-point)
(define-constructor 'segment make-segment)
					;(define-constructor 'circle-by-center-radius make-circle-by-center-radius)
					;(define-constructor 'circle-by-center-point make-circle-by-center-point)
					;(define-constructor 'circle-by-3-points make-circle-by-3-points)

					;(define-constructor 'arc-by-center-radius-2-angles make-arc-by-center-radius-2-angles)
					;(define-constructor 'arc-by-center-point-angle make-arc-by-center-point-angle)
					;(define-constructor 'arc-by-3-points make-arc-by-3-points)

(define-constructor 'ellipse-by-center-2-radii-rotation-2-angles 
  make-elarc-by-center-2-radii-rotation-2-angles) ;; OBSOLETE

(define-constructor 'ellipse-by-center-2-radii-rotation make-ellipse-by-center-2-radii-rotation)
(define-constructor 'ellipse-by-2-foci-point make-ellipse-by-2-foci-point)
(define-constructor 'ellipse-by-3-points-smallest make-ellipse-by-3-points-smallest)
(define-constructor 'ellipse-by-3-points-rotation-eccentricity make-ellipse-by-3-points-rotation-eccentricity)
(define-constructor 'ellipse-by-center-point-rotation-eccentricity make-ellipse-by-center-point-rotation-eccentricity)

(define-constructor 'quadratic-bezier make-quadratic-rational-bezier)
(define-constructor 'cubic-bezier make-cubic-rational-bezier)

(define-constructor 'text make-text)
(define-constructor 'tex-text make-tex-text)

(define-constructor 'parametric-point make-parametric-point-on-hanger)
(define-constructor 'intersection make-intersection-on-hanger)
(define-constructor 'decorator-point make-decorator-point)
(define-constructor 'decorator-arrow make-decorator-arrow-on-hanger)

(define-constructor 'group make-group)
(define-constructor 'path make-path)


(define-source 'point 1)
(define-source 'segment 2)

(define-source 'ellipse-by-center-2-radii-rotation-2-angles 
  (list 0 'a-radius 'b-radius 'rotation 'start 'dif 'conic)) ;; OBSOLETE

(define-source 'ellipse-by-center-2-radii-rotation (list 0 'a-radius 'b-radius 'rotation))
(define-source 'ellipse-by-2-foci-point 3)
(define-source 'ellipse-by-3-points-smallest 3)
(define-source 'ellipse-by-3-points-rotation-eccentricity (list 0 1 2 'rotation 'ecc))
(define-source 'ellipse-by-center-point-rotation-eccentricity (list 0 1 'rotation 'ecc))

(define-source 'quadratic-bezier (list 0 'weight-0 1 'weight-1 2 'weight-2))
(define-source 'cubic-bezier (list 0 'weight-0 1 'weight-1 2 'weight-2 3 'weight-3))

(define (text-get-font-family go) (font-family (car (property-get-int go 'font))))
(define (text-get-font-style go)  (font-style  (car (property-get-int go 'font))))

(define-source 'text (list 0 'text text-get-font-family text-get-font-style 'size))
(define-source 'tex-text (list 0 'text))

(define-source 'parametric-point (list -1 'parameter))
(define-source 'intersection (list -1 -2 'parameter))
(define-source 'decorator-point (list 0 'vertices 'radius 'rotation))
(define-source 'decorator-arrow (list -1 'parameter 'arrow-length 'rotation))

(define-source 'group '())
(define-source 'path '())


(define (lookup name) (o-lookup name))
(define (o-trans o) (o-get-trans-tlo o))

(define (set-conic! go conic . rest)
  (or (proxy? go 'go 'elliptic-arc) (typearg-error "set-conic!" 1 go))
  (or (code-conic conic) (typearg-error "set-conic!" 2 conic))
  (property-uns-set go 'conic (code-conic conic))
  (case conic
    ((start-entire)
     (or (real? (car rest)) (typearg-error "set-conic!" 3 (car rest)))
     (or (null? (cdr rest)) (typearg-error "set-conic!" 4 (cadr rest)))
     (property-real-set go 'start (car rest)))
    ((start-dif)
     (or (real? (car rest)) (typearg-error "set-conic!" 3 (car rest)))
     (or (real? (cadr rest)) (typearg-error "set-conic!" 4 (cadr rest)))
     (or (null? (cddr rest)) (typearg-error "set-conic!" 5 (caddr rest)))
     (property-real-set go 'start (car rest))
     (property-real-set go 'dif (cadr rest)))
    ((point-dif)
     (or (real? (car rest)) (typearg-error "set-conic!" 3 (car rest)))
     (or (null? (cdr rest)) (typearg-error "set-conic!" 4 (cadr rest)))
     (property-real-set go 'dif (car rest)))
    (else
     (or (null? rest) (typearg-error "set-conic!" 3 (car rest))))))

(define (get-conic go)
  (or (proxy? go 'go 'elliptic-arc) (typearg-error "get-conic" 1 go))
  (let ((conic (cadr (property-get go 'conic))))
    (case conic
      ((start-entire)
       (list conic (property-get go 'start)))
      ((start-dif)
       (list conic (property-get go 'start) (property-get go 'dif)))
      ((point-dif)
       (list conic (property-get go 'dif)))
      (else
       (list conic)))))

