/** @file
 * GUILE-link - interface functions for GUI data types.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/lib.h"
#include <stdio.h>
#include <string.h>
#include <guile/gh.h>
#include "kernel/kernel.h"
#include "gui/main.h"
#include "gui/guileaccess.h"
#include "build/path.h"

#include "scheme/gl_trans.h"
#include "scheme/glt_common.h"
#include "scheme/glt_kernel.h"
#include "scheme/glt_gui.h"

scm_t_bits window_tag;

uns is_window_proxy (SCM proxy)
{ return SCM_SMOB_PREDICATE(window_tag, proxy); }


static size_t window_proxy_free (SCM proxy)
{
    if (vptr (proxy) != NULL)
    {
	vptr (proxy) -> proxy = SCM_BOOL_F;
	SCM_SET_SMOB_DATA (proxy, NULL); // unneccessary
    }
    return (0);
}

SCM window2scm(struct window *p)
{
  if (p == NULL) return (SCM_BOOL_F);

  if (SCM_EQ_P (p -> proxy, SCM_BOOL_F))
    SCM_NEWSMOB (p -> proxy, window_tag, p);

  return (p -> proxy);
}

void window_proxy_clean (struct window *p)
{
    if (! SCM_EQ_P (p -> proxy, SCM_BOOL_F))
    {
	SCM_SET_SMOB_DATA (p -> proxy, NULL);
	p -> proxy = SCM_BOOL_F; // unneccessary
    }
}

static SCM
gl_window_p (SCM obj)
{ return (SCM_BOOL (is_window_proxy (obj))); }

static SCM
gl_view_p (SCM obj)
{ return (SCM_BOOL (is_specific_window_proxy (obj, WT_VIEW))); }


#include "obj/gui/generated_bindings.h"

static SCM
update_gui_module (UNUSED void *arg)
{
  scm_c_define_gsubr ("window?", 1, 0, 0, gl_window_p);
  scm_c_define_gsubr ("view?", 1, 0, 0, gl_view_p);
  scm_c_export ("window?", "view?", NULL);

  gui_generated_bindings_init ();

  return (SCM_UNSPECIFIED);
}


void
guilelink_gui_init (void)
{
  window_tag = scm_make_smob_type ("vrr-window", 0);
  scm_set_smob_free (window_tag, window_proxy_free);

  scheme_module_load (LIB_PATH "gui.scm");

  SCM gui = scm_c_resolve_module ("vrr gui");
  scm_c_call_with_current_module (gui, update_gui_module, NULL);
  console_init ();

  scm_c_use_module ("vrr gui");
}
