;;; testing suite
;;;
;;; Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
;;; This program is distributed under GNU General Public License, see COPYING.


(define (call-with-error-pack thunk)
  (catch #t 
	 (lambda () (cons #t (thunk)))
	 (lambda args (cons #f args))))

(define (output . items)
  (for-each display items)
  (newline))

(define (run-test-line . items)
  (let run-test-case ((fn (car items))
		      (name (cadr items))
		      (args '())
		      (rest (cddr items)))
    (let ((result (call-with-error-pack (lambda () (apply fn args)))))
      (if (car result)
	  (begin
	    (output "success: " name)
	    (if (not (null? rest))
		(run-test-case (car rest) (cadr rest) (cdr result) (cddr rest))))
	  (begin
	     (output "FAIL: " name " error: " (cdr result)))))))

(define (test-make-tlo)
  (trans-main universe
	      (lambda ()
		(let ((doc (make-document))
		      (tlo (make-page)))
		  (n-set-parent! tlo doc)
		  (n-set-parent! doc universe)
		  (list doc tlo)))))

(define (test-make-point doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-point (coords 12 12))
		(list doc tlo))))

(define (test-make-segment doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-segment (coords 10 10) (coords 20 30))
		(list doc tlo))))

(define (test-make-quadratic doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-bezier/2 (coords 10 10) (coords -5 40) (coords 20 30))
		(list doc tlo))))


(define (test-make-cubic doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-bezier/3 (coords 10 10) (coords -5 40) (coords 40 0) (coords 20 30))
		(list doc tlo))))

(define (test-make-e/c2rr doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-ellipse/c2rr (coords -10 10) 50 20 0.2)
		(list doc tlo))))

(define (test-make-e/2fp doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-ellipse/2fp (coords -15 10) (coords 15 15) (coords 5 0))
		(list doc tlo))))

(define (test-make-e/3ps doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-ellipse/3ps (coords -15 10) (coords 15 15) (coords 5 0))
		(list doc tlo))))

(define (test-make-e/3pre doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-ellipse/3pre (coords -15 10) (coords 15 15) (coords 5 0) 0.7 0.6)
		(list doc tlo))))

(define (test-make-e/cpre doc tlo)
  (trans-main tlo
	      (lambda () 
		(make-ellipse/cpre (coords -10 -5) (coords 15 20) 0.2 0.6)
		(list doc tlo))))

(define (test-make-text doc tlo)
  (trans-main tlo
	      (lambda ()
		(make-text (coords 25 25) "Hello Hello!" "Sans" "Medium" 10)
		(list doc tlo))))

(define (test-make-tex doc tlo)
  (trans-main tlo
	      (lambda ()
		(make-tex-text (coords 30 30) "$\\alpha + 2^\\beta$")
		(list doc tlo))))

(define (test-make-decorator-point doc tlo)
  (trans-main tlo
	      (lambda ()
		(make-decorator-point (coords 17 13) 5 10 0.2)
		(list doc tlo))))

(define (test-make-decorator-arrow doc tlo)
  (trans-main tlo
	      (lambda ()
		(let ((s (make-segment (coords 12 12) (coords 40 30))))
		  (make-decorator-arrow s 0.5 5 0))
		(list doc tlo))))

(define (test-make-deps-1 doc tlo)
  (trans-main tlo
	      (lambda ()
		(let ((s1 (make-segment (coords 5 5) (coords 15 15)))
		      (s2 (make-segment (coords 25 5) (coords 35 15))))
		  (make-segment (h-start s1) (h-end s2)))
		(list doc tlo))))

(define (test-make-deps-2 doc tlo)
  (trans-main tlo
	      (lambda ()
		(let ((s1 (make-segment (coords 5 25) (coords 15 35)))
		      (s2 (make-segment (coords 25 25) (coords 35 35))))
		  (make-segment (h-center (make-parametric-point s1 0.3))
				(h-center (make-parametric-point s2 0.6))))
		(list doc tlo))))

(define (test-make-deps-3 doc tlo)
  (trans-main tlo
	      (lambda ()
		(let ((s1 (make-segment (coords 0 0) (coords -50 -50)))
		      (s2 (make-segment (coords -50 0) (coords 0 -50))))
		  (make-segment (h-center (make-intersection s1 s2 0.5))
				(coords -25 0)))
		(list doc tlo))))

(define (test-save-doc doc tlo)
  (save-doc doc "file.vrr")
  (list doc tlo))

(define (test-save-ps doc tlo)
  (save-as-ps doc "file.ps" 100 100 #f #f)
  (list doc tlo))

(define (test-save-pdf doc tlo)
  (save-as-pdf doc "file.pdf" #f 5)
  (list doc tlo))

(define (test-save-eps doc tlo)
  (save-as-eps tlo "file.eps" #f)
  (list doc tlo))

(define (test-save-svg doc tlo)
  (save-as-svg tlo "file.svg")
  (list doc tlo))

(define (test-load-doc)
  (list (load-doc "file.vrr")))

(define (test-unlink-doc doc tlo)
  (trans-main universe
	      (lambda ()
		(n-set-parent! doc #f)))
  '())

(define (run-tests)  
(run-test-line test-make-tlo      "make-page"
	       test-make-point    "make-point"
	       test-save-doc      "save-point"
	       test-save-ps       "export-point-ps"
	       test-save-pdf      "export-point-pdf"
	       test-save-eps      "export-point-eps"
	       test-save-svg      "export-point-svg"
	       test-unlink-doc    "unlink-doc"
	       test-load-doc      "load-point")

(run-test-line test-make-tlo      "make-page"
	       test-make-segment  "make-segment"
	       test-save-doc      "save-segment"
	       test-save-ps       "export-segment-ps"
	       test-save-pdf      "export-segment-pdf"
	       test-save-eps      "export-segment-eps"
	       test-save-svg      "export-segment-svg"
	       test-unlink-doc    "unlink-doc"
	       test-load-doc      "load-segment")

(run-test-line test-make-tlo       "make-page"
	       test-make-quadratic "make-bezier-2"
	       test-save-doc       "save-bezier-2"
	       test-save-ps       "export-2-ps"
	       test-save-pdf      "export-2-pdf"
	       test-save-eps      "export-2-eps"
	       test-save-svg      "export-2-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-bezier-2")

(run-test-line test-make-tlo       "make-page"
	       test-make-cubic     "make-bezier-3"
	       test-save-doc       "save-bezier-3"
	       test-save-ps       "export-bezier-3-ps"
	       test-save-pdf      "export-bezier-3-pdf"
	       test-save-eps      "export-bezier-3-eps"
	       test-save-svg      "export-bezier-3-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-bezier-3")

(run-test-line test-make-tlo       "make-page"
	       test-make-e/c2rr    "make-ellipse-c2rr"
	       test-save-doc       "save-ellipse-c2rr"
	       test-save-ps       "export-ellipse-c2rr-ps"
	       test-save-pdf      "export-ellipse-c2rr-pdf"
	       test-save-eps      "export-ellipse-c2rr-eps"
	       test-save-svg      "export-ellipse-c2rr-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-ellipse-c2rr")

(run-test-line test-make-tlo       "make-page"
	       test-make-e/2fp     "make-ellipse-2fp"
	       test-save-doc       "save-ellipse-2fp"
	       test-save-ps       "export-ellipse-2fp-ps"
	       test-save-pdf      "export-ellipse-2fp-pdf"
	       test-save-eps      "export-ellipse-2fp-eps"
	       test-save-svg      "export-ellipse-2fp-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-ellipse-2fp")

(run-test-line test-make-tlo       "make-page"
	       test-make-e/3ps    "make-ellipse-3ps"
	       test-save-doc       "save-ellipse-3ps"
	       test-save-ps       "export-ellipse-3ps-ps"
	       test-save-pdf      "export-ellipse-3ps-pdf"
	       test-save-eps      "export-ellipse-3ps-eps"
	       test-save-svg      "export-ellipse-3ps-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-ellipse-3ps")

(run-test-line test-make-tlo       "make-page"
	       test-make-e/3pre    "make-ellipse-3pre"
	       test-save-doc       "save-ellipse-3pre"
	       test-save-ps       "export-ellipse-3pre-ps"
	       test-save-pdf      "export-ellipse-3pre-pdf"
	       test-save-eps      "export-ellipse-3pre-eps"
	       test-save-svg      "export-ellipse-3pre-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-ellipse-3pre")

(run-test-line test-make-tlo       "make-page"
	       test-make-e/cpre    "make-ellipse-cpre"
	       test-save-doc       "save-ellipse-cpre"
	       test-save-ps       "export-ellipse-cpre-ps"
	       test-save-pdf      "export-ellipse-cpre-pdf"
	       test-save-eps      "export-ellipse-cpre-eps"
	       test-save-svg      "export-ellipse-cpre-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-ellipse-cpre")

(run-test-line test-make-tlo       "make-page"
	       test-make-text      "make-text"
	       test-save-doc       "save-text"
	       test-save-ps       "export-text-ps"
	       test-save-pdf      "export-text-pdf"
	       test-save-eps      "export-text-eps"
	       test-save-svg      "export-text-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-text")


(run-test-line test-make-tlo       "make-page"
	       test-make-tex       "make-tex-text"
	       test-save-doc       "save-tex-text"
	       test-save-ps       "export-tex-text-ps"
	       test-save-pdf      "export-tex-text-pdf"
	       test-save-eps      "export-tex-text-eps"
	       test-save-svg      "export-tex-text-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-tex-text")

(run-test-line test-make-tlo       "make-page"
	       test-make-decorator-point "make-decorator-point"
	       test-save-doc       "save-decorator-point"
	       test-save-ps       "export-decorator-point-ps"
	       test-save-pdf      "export-decorator-point-pdf"
	       test-save-eps      "export-decorator-point-eps"
	       test-save-svg      "export-decorator-point-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-decorator-point")

(run-test-line test-make-tlo       "make-page"
	       test-make-decorator-arrow "make-decorator-arrow"
	       test-save-doc       "save-decorator-arrow"
	       test-save-ps       "export-decorator-arrow-ps"
	       test-save-pdf      "export-decorator-arrow-pdf"
	       test-save-eps      "export-decorator-arrow-eps"
	       test-save-svg      "export-decorator-arrow-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-decorator-arrow")

(run-test-line test-make-tlo       "make-page"
	       test-make-deps-1    "make-dependent-segments-endpoints"
	       test-save-doc       "save-dependent-segments-endpoints"
	       test-save-ps       "export-dependent-segments-endpoints-ps"
	       test-save-pdf      "export-dependent-segments-endpoints-pdf"
	       test-save-eps      "export-dependent-segments-endpoints-eps"
	       test-save-svg      "export-dependent-segments-endpoints-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-dependent-segments-endpoints")

(run-test-line test-make-tlo       "make-page"
	       test-make-deps-2    "make-dependent-segments-parametric-points"
	       test-save-doc       "save-dependent-segments-parametric-points"
	       test-save-ps       "export-dependent-segments-parametric-points-ps"
	       test-save-pdf      "export-dependent-segments-parametric-points-pdf"
	       test-save-eps      "export-dependent-segments-parametric-points-eps"
	       test-save-svg      "export-dependent-segments-parametric-points-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-dependent-segments-parametric-points")

(run-test-line test-make-tlo       "make-page"
	       test-make-deps-3    "make-dependent-segments-intersections"
	       test-save-doc       "save-dependent-segments-intersections"
	       test-save-ps       "export-dependent-segments-intersections-ps"
	       test-save-pdf      "export-dependent-segments-intersections-pdf"
	       test-save-eps      "export-dependent-segments-intersections-eps"
	       test-save-svg      "export-dependent-segments-intersections-svg"
	       test-unlink-doc     "unlink-doc"
	       test-load-doc       "load-dependent-segments-intersections"))
	       
(run-tests)	     

