#ifndef _VCL_CONTEXT_H_
#define _VCL_CONTEXT_H_

/**@file
 *
 * VCL module header
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */



enum vcl_context_property_id
  {
    VCL_PROP_FILL_COLOR,
    VCL_PROP_STROKE_COLOR,
    VCL_PROP_STROKE_WIDTH,
    VCL_PROP_STROKE_CAP_STYLE,
    VCL_PROP_STROKE_JOIN_STYLE,
    VCL_PROP_COUNT
  };

enum vcl_cap_style
  {
    VCL_CAP_BUTT,
    VCL_CAP_ROUND,
    VCL_CAP_PROJECTING,
    VCL_CAP_COUNT
  };

enum vcl_join_style
  {
    VCL_JOIN_MITER,
    VCL_JOIN_ROUND,
    VCL_JOIN_BEVEL,
    VCL_JOIN_COUNT
  };


#ifdef _VCL_INTERNAL_


extern int vcl_property_sizes [];
extern int vcl_property_offsets [];

struct vcl_context
{
  struct vcl_growing_array storage;
  struct geom_transform2 t;
  u32 fill_color;
  u32 stroke_color;
  double stroke_width;
  double dist_transform;
  u8 stroke_cap_style;
  u8 stroke_join_style;
};


extern struct vcl_context *
vcl_context_new (void);

extern void
vcl_context_destroy (struct vcl_context *ct);

extern void
vcl_context_set_property (struct vcl_context *ct, int prop_id, void *data);

extern void
vcl_context_unset_property (struct vcl_context *ct, int prop_id);

extern void
vcl_context_preply_transform (struct vcl_context *ct, void *trans);

extern void
vcl_context_unpreply_transform (struct vcl_context *ct);

extern void
vcl_context_update_dist_transform (struct vcl_context *ct);

static inline double
vcl_context_get_dist_transform (struct vcl_context *ct)
{
  if (ct -> dist_transform == -1) vcl_context_update_dist_transform (ct);
  return (ct -> dist_transform);
}

extern void
vcl_context_walk_node (struct vcl_context *ct, void *painter, void *node,
		       const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local);

void
vcl_context_walk_children (struct vcl_context *ct, void *painter, void *node,
			   const struct vcl_rectangle *bbox, const struct geom_rectangle *bbox_local);


#endif
#endif
