/*
 * VCL module
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vcl_vtable *vcl_vt_rect;
vcl_vtable *vcl_vt_rect_s;


// OBJECT interface

static void
vcl_rect_destroy (struct vcl_rect * obj)
{
  ASSERT (obj -> parent == NULL);
  xfree (obj);
}


// NODE interface

static void
vcl_rect_set_data (struct vcl_rect * obj, int data)
{ obj -> data = data; }

static int
vcl_rect_get_data (struct vcl_rect * obj)
{ return (obj -> data); }

static void
vcl_rect_set_parent (struct vcl_rect * obj, void * parent)
{ obj -> parent = parent; }

static void *
vcl_rect_get_parent (struct vcl_rect * obj)
{ return (obj -> parent); }

static void
vcl_rect_get_bbox (struct vcl_rect * obj, struct geom_rectangle * bbox)
{
  *bbox = obj -> r;
}


// SHAPE interface

static void *
vcl_rect_polygonize (void * a, void * it, UNUSED double delta,
			const struct geom_transform *t,
			UNUSED const struct geom_rectangle *bbox,
			u8 ** buff, uns * buff_size,
			uns *closed, uns *points, UNUSED uns hints)
{
  struct vcl_rect *obj = a;

  ASSERT (it == NULL);

  uns wanted = (5 * sizeof (struct geom_point));
  if (*buff_size < wanted)
    {
      *buff_size = wanted;
      *buff = xrealloc (*buff, *buff_size);
    }

  struct geom_point pi;
  struct geom_point *xbuff = (struct geom_point *) *buff;
  pi . x = obj -> r . ll . x;
  pi . y = obj -> r . ll . y;
  geom_point_transform (&pi, t, xbuff + 0);
  pi . x = obj -> r . ur . x;
  pi . y = obj -> r . ll . y;
  geom_point_transform (&pi, t, xbuff + 1);
  pi . x = obj -> r . ur . x;
  pi . y = obj -> r . ur . y;
  geom_point_transform (&pi, t, xbuff + 2);
  pi . x = obj -> r . ll . x;
  pi . y = obj -> r . ur . y;
  geom_point_transform (&pi, t, xbuff + 3);
  xbuff [4] = xbuff [0];
  *points = 5;
  *closed = 1;

  return (NULL);
}


// specific functions

void
vcl_rect_init (struct vcl_rect *obj, struct geom_rectangle r)
{
  vcl_object_init (obj, vcl_vt_rect_s);
  obj -> r = r;
}

struct vcl_rect *
vcl_rect_new (struct geom_rectangle r)
{
  struct vcl_rect *obj = xmalloc (sizeof (struct vcl_rect));
  vcl_object_init (obj, vcl_vt_rect);
  obj -> r = r;
  obj -> parent = NULL;
  obj -> data = 0;

  return (obj);
}

void
vcl_rect_main (void)
{
  vcl_vt_rect = vcl_class_new ("rect");

  vcl_class_add_interface (vcl_vt_rect, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_rect_destroy);

  vcl_class_add_interface (vcl_vt_rect, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_rect_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_rect_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_rect_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_rect_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_rect_get_bbox);

  vcl_class_add_interface (vcl_vt_rect, vcl_if_shape);
  vcl_class_add_method_li (vcl_mid_shape_polygonize, vcl_rect_polygonize);

  vcl_vt_rect_s = vcl_class_new ("rect_s");

  vcl_class_add_interface (vcl_vt_rect_s, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_rect_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_rect_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_rect_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_rect_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_rect_get_bbox);

  vcl_class_add_interface (vcl_vt_rect_s, vcl_if_shape);
  vcl_class_add_method_li (vcl_mid_shape_polygonize, vcl_rect_polygonize);
}
