#ifndef _VCL_RECTANGLE_H_
#define _VCL_RECTANGLE_H_

/**@file
 *
 * VCL module header
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */



struct vcl_rectangle
{
  int lx;
  int ly;
  int hx;
  int hy;
};

static inline void
vcl_rectangle_set_max (struct vcl_rectangle *r)
{
  r -> lx = r -> ly = INT_MIN;
  r -> hx = r -> hy = INT_MAX;
}

static inline void
vcl_rectangle_set_empty (struct vcl_rectangle *r)
{
  r -> lx = r -> ly = INT_MAX;
  r -> hx = r -> hy = INT_MIN;
}

static inline void
vcl_rectangle_enlarge (struct vcl_rectangle *r, int x, int y)
{
  if (r -> lx > x) r -> lx = x;
  if (r -> hx <= x) r -> hx = x + 1;
  if (r -> ly > y) r -> ly = y;
  if (r -> hy <= y) r -> hy = y + 1;
}

static inline void
vcl_rectangle_union (const struct vcl_rectangle *r1,
		     const struct vcl_rectangle *r2,
		     struct vcl_rectangle *result)
{
  result -> lx = MIN (r1 -> lx, r2 -> lx);
  result -> ly = MIN (r1 -> ly, r2 -> ly);
  result -> hx = MAX (r1 -> hx, r2 -> hx);
  result -> hy = MAX (r1 -> hy, r2 -> hy);
}

static inline void
vcl_rectangle_intersection (const struct vcl_rectangle *r1,
			    const struct vcl_rectangle *r2,
			    struct vcl_rectangle *result)
{
  result -> lx = MAX (r1 -> lx, r2 -> lx);
  result -> ly = MAX (r1 -> ly, r2 -> ly);
  result -> hx = MIN (r1 -> hx, r2 -> hx);
  result -> hy = MIN (r1 -> hy, r2 -> hy);
}

static inline void
vcl_rectangle_to_geom_rectangle (const struct vcl_rectangle *src,
				 struct geom_rectangle *dst)
{
  dst -> ll . x = src -> lx;
  dst -> ll . y = src -> ly;
  dst -> ur . x = src -> hx;
  dst -> ur . y = src -> hy;
}

static inline int
vcl_rectangle_line_in_p (struct vcl_rectangle *r, int line)
{
  return ((line >= r -> ly) && (line < r -> hy));
}

static inline int
vcl_rectangle_empty_p (struct vcl_rectangle *r)
{
  return ((r -> hy <= r -> ly) || (r -> hx <= r -> lx));
}


#endif
