/*
 * VCL module
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG

#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "import/import.h"
#include "kernel/kernel.h"
#include "font/font.h"

vcl_vtable *vcl_vt_text_layout;


// OBJECT interface

static void
vcl_text_layout_destroy (struct vcl_text_layout * obj)
{
  ASSERT (obj -> parent == NULL);
  if (obj -> copy) xfree (obj -> string . string);
  xfree (obj);
}


// NODE interface

static void
vcl_text_layout_set_data (struct vcl_text_layout * obj, int data)
{ obj -> data = data; }

static int
vcl_text_layout_get_data (struct vcl_text_layout * obj)
{ return (obj -> data); }

static void
vcl_text_layout_set_parent (struct vcl_text_layout * obj, void * parent)
{ obj -> parent = parent; }

static void *
vcl_text_layout_get_parent (struct vcl_text_layout * obj)
{ return (obj -> parent); }

static void
vcl_text_layout_get_bbox (struct vcl_text_layout * obj, struct geom_rectangle * bbox)
{ *bbox = obj -> string . bbox; }


// COMPOSITE interface

static void
vcl_text_layout_get_children (struct vcl_text_layout * obj, void (* cb)(void *, void * ), void * cb_data)
{ cb (&(obj -> string), cb_data); }

static void
vcl_text_layout_get_children_backwards (struct vcl_text_layout * obj, void (* cb)(void *, void * ), void * cb_data)
{ cb (&(obj -> string), cb_data); }

static void
vcl_text_layout_get_children_in_bbox (struct vcl_text_layout * obj, const struct geom_rectangle * wanted_bbox, void (* cb)(void *, void * ), void * cb_data)
{
  if (geom_rectangle_have_intersection (&(obj -> string . bbox), wanted_bbox))
    cb (&(obj -> string), cb_data);
}

static void
vcl_text_layout_get_children_in_bbox_backwards (struct vcl_text_layout * obj, const struct geom_rectangle * wanted_bbox, void (* cb)(void *, void * ), void * cb_data)
{
  if (geom_rectangle_have_intersection (&(obj -> string . bbox), wanted_bbox))
    cb (&(obj -> string), cb_data);
}


// specific functions

struct vcl_text_layout *
vcl_text_layout_new (uns copy, char *string, int font_id, real font_size, const struct geom_rectangle *bbox)
{
  struct vcl_text_layout *obj = xmalloc (sizeof (struct vcl_text_layout));
  vcl_object_init (obj, vcl_vt_text_layout);

  struct geom_rectangle tmp_bbox;
  if (bbox == NULL)
    {
      struct font_ctl ctl;

      font_ctl_init (&ctl);
      ctl . font_id = font_id;
      ctl . font_size = font_size;
      ctl . charmap = FONT_CHARMAP_STD;
      if (font_get_string_bbox (&ctl, string, &tmp_bbox) < 0)
	geom_rectangle_set_empty (&tmp_bbox);
      font_ctl_cleanup (&ctl);
      bbox = &tmp_bbox;
    }

  vcl_string_init (&(obj -> string), string, FONT_CHARMAP_STD, font_id, font_size, bbox);
  obj -> copy = copy;
  return (obj);
}

void
vcl_text_layout_main (void)
{
  vcl_vt_text_layout = vcl_class_new ("text-layout");

  vcl_class_add_interface (vcl_vt_text_layout, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_text_layout_destroy);

  vcl_class_add_interface (vcl_vt_text_layout, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_text_layout_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_text_layout_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_text_layout_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_text_layout_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_text_layout_get_bbox);

  vcl_class_add_interface (vcl_vt_text_layout, vcl_if_composite);
  vcl_class_add_method_li (vcl_mid_composite_get_children, vcl_text_layout_get_children);
  vcl_class_add_method_li (vcl_mid_composite_get_children_backwards, vcl_text_layout_get_children_backwards);
  vcl_class_add_method_li (vcl_mid_composite_get_children_in_bbox, vcl_text_layout_get_children_in_bbox);
  vcl_class_add_method_li (vcl_mid_composite_get_children_in_bbox_backwards, vcl_text_layout_get_children_in_bbox_backwards);
}
